/*
 * WCOL:
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include <sys/types.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif

#include <netinet/in.h>
#include <netdb.h>

#include <errno.h>

#include "wcol.h"
#include "args.h"
#include "http.h"
#include "sim_stat.h"
#include "sess.h"

/*
 * externals
 */

/* automatic */
extern                       char config_filename[];
extern                      char* log_filename;
extern                        int trace_flag;
extern                        int monitor_flag;
extern                        int prefetch_flag;
extern                        int service_portno;
extern                      char* localdomain_padding;
extern                      char* report_hook_hostname;
extern                        int require_sess_length;
extern                        int require_n_handler;
extern                        int require_chunk_size;
extern                      char* cache_dir_prefix;
extern                        int require_radix_h2;
extern                        int require_radix_h3;
extern                      off_t require_dir_size;
extern                        int require_swap_token_val;
extern                        int require_swap_token_interval;
extern                        int require_disk_sweep_interval;
extern                        int require_n_dsweeper;
extern                        int require_n_msweeper;
extern                      off_t require_total_memory_size;
extern                        int require_num_bucket;
extern                        int require_level_bucket_skiplist;
extern                        int require_size_frag;
extern                        int require_ratio_frag_length;
extern                        int num_prefetch_referedpages;
extern                        int num_prefetch_includedobjs;

/* static */
extern pid_t    wcol_pid;
extern char     *progname;

/* end of externals */

char config_file[STRING_SIZE];
char dummy_string[STRING_SIZE];


/*
 * SetGlobalDefaults -
 */
void
SetGlobalDefaults()
{
    strcpy(config_filename, DEFAULT_CONFIGFILE);
    log_filename = strdup(DEFAULT_LOGFILE);
    trace_flag = 1;
    monitor_flag = 1;
    prefetch_flag = 1;
    service_portno = (int) HTTP_PROXYPORTNO;
    localdomain_padding = strdup(".falcon.proxy.cache");
    report_hook_hostname = strdup("www.self-report.cache");
    require_sess_length = (int) 2048;
    require_n_handler = (int) 4;
    require_chunk_size = (int) 32;
    cache_dir_prefix = strdup(DEFAULT_CACHEDIR);
    require_radix_h2 = (int) 17;
    require_radix_h3 = (int) 19;
    require_dir_size = (off_t) 120*(1024);
    require_swap_token_val = (int) 320;
    require_swap_token_interval = (int) 2;
    require_disk_sweep_interval = (int) 500;
    require_n_dsweeper = (int) 1;
    require_n_msweeper = (int) 1;
    require_total_memory_size = (off_t) 32*(1048576);
    require_num_bucket = (int) 64;
    require_level_bucket_skiplist = (int) 8;
    require_size_frag = (int) 64;
    require_ratio_frag_length = (int) 5;
    num_prefetch_referedpages = (int) 8;
    num_prefetch_includedobjs = (int) 16;
}
/* end of SetGlobalDefaults() */





#define MAX_AV 10

/*
 * ReadConfigFile -
 */
void ReadConfigFile(char *filename)
{
    arg_t  *args, *pargs;
    int     num;
    int     tint;
    char   *av[MAX_AV];
    int     i, j;


    Trace("ReadConfigFile: filename='%s'\n", filename);
    if((num=ReadArgsFromFile(filename, &args))<=0) {
        Trace("ReadConfigFile: no arguments\n");
        return;
    }

    pargs = args;
    while(pargs) {
        /* project argument value to temporary pointor */
        for(i=0;pargs->vals[i] && i<MAX_AV;i++) {
            av[i] = pargs->vals[i];
        }
        for(j=i;j<MAX_AV;j++) {
            av[j] = NULL;
        }

        if(!pargs->name) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConfigFile")) {
            if(pargs->vals[0]) {
                strcpy(config_filename,av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LogFile")) {
            if(pargs->vals[0]) {
                log_filename = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "Trace")) {
            trace_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Trace")==0) {
            trace_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "Monitor")) {
            monitor_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Monitor")==0) {
            monitor_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "Prefetch")) {
            prefetch_flag = 1;
            goto nextone;
        }
        else
        if((pargs->name[0]=='N' || pargs->name[0]=='n')
            && (pargs->name[1]=='O' || pargs->name[1]=='o')
            && strcasecmp(&pargs->name[2], "Prefetch")==0) {
            prefetch_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "ServicePort")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    service_portno = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LocalDomain")) {
            if(pargs->vals[0]) {
                localdomain_padding = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "ReportHook")) {
            if(pargs->vals[0]) {
                report_hook_hostname = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumSession")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_sess_length = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumHandler")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_handler = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SessionChunkLength")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_chunk_size = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheDir")) {
            if(pargs->vals[0]) {
                cache_dir_prefix = strdup(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH2")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_radix_h2 = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH3")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_radix_h3 = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'D', "DirSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_dir_size = (off_t) tint * 1024;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SwapTokenVal")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_swap_token_val = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SwapTokenInterval")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_swap_token_interval = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'D', "DiskSweepInterval")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_disk_sweep_interval = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumDiskSweeper")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_dsweeper = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumMemorySweeper")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_n_msweeper = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "TotalMemorySize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_total_memory_size = (off_t) tint * 1048576;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumBucket")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_num_bucket = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LevelBucketSkipList")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_level_bucket_skiplist = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SizeFrag")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_size_frag = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RatioFragLength")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    require_ratio_frag_length = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrefetchReferedPages")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_prefetch_referedpages = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrefetchIncludedObjects")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_prefetch_includedobjs = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'A', "AcceptAddress")) {
            if(pargs->vals[0]) {
                if(AppendAcceptEntry(av[0], av[1])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RadixH1")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxLinePerWrite")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxLinePerObj")) {
            /* nothing */
            goto nextone;
        }
        else {
            Error("Unknow argument '%s'", pargs->name);
            goto nextone;
        }
fail:
        Error("Argment specification ('%s') is wrong", pargs->name);
nextone:
        pargs = pargs->next;
    }

    FreeArgs(args);
}



/*
 * FixGlobals - Fix global vars.
 */
void
FixGlobals()
{

    if(service_portno<1) {
        Error("FixGlobals: 'ServicePort' is too small (min %d)",1);
        service_portno = (int) 1;
    }
    if(service_portno>32000) {
        Error("FixGlobals: 'ServicePort' is too large (max %d)",32000);
        service_portno = (int) 32000;
    }
    if(require_sess_length<128) {
        Error("FixGlobals: 'NumSession' is too small (min %d)",128);
        require_sess_length = (int) 128;
    }
    if(require_sess_length>N_SESS_MAX) {
        Error("FixGlobals: 'NumSession' is too large (max %d)",N_SESS_MAX);
        require_sess_length = (int) N_SESS_MAX;
    }
    if(require_n_handler<1) {
        Error("FixGlobals: 'NumHandler' is too small (min %d)",1);
        require_n_handler = (int) 1;
    }
    if(require_n_handler>N_HANDLER_MAX) {
        Error("FixGlobals: 'NumHandler' is too large (max %d)",N_HANDLER_MAX);
        require_n_handler = (int) N_HANDLER_MAX;
    }
    if(require_chunk_size<1) {
        Error("FixGlobals: 'SessionChunkLength' is too small (min %d)",1);
        require_chunk_size = (int) 1;
    }
    if(require_chunk_size>SESS_CHUNK_SIZE) {
        Error("FixGlobals: 'SessionChunkLength' is too large (max %d)",SESS_CHUNK_SIZE);
        require_chunk_size = (int) SESS_CHUNK_SIZE;
    }
    if(require_radix_h2<3) {
        Error("FixGlobals: 'RadixH2' is too small (min %d)",3);
        require_radix_h2 = (int) 3;
    }
    if(require_radix_h2>293) {
        Error("FixGlobals: 'RadixH2' is too large (max %d)",293);
        require_radix_h2 = (int) 293;
    }
    if(require_radix_h3<3) {
        Error("FixGlobals: 'RadixH3' is too small (min %d)",3);
        require_radix_h3 = (int) 3;
    }
    if(require_radix_h3>293) {
        Error("FixGlobals: 'RadixH3' is too large (max %d)",293);
        require_radix_h3 = (int) 293;
    }
    if(require_swap_token_val<1) {
        Error("FixGlobals: 'SwapTokenVal' is too small (min %d)",1);
        require_swap_token_val = (int) 1;
    }
    if(require_swap_token_val>1073741824) {
        Error("FixGlobals: 'SwapTokenVal' is too large (max %d)",1073741824);
        require_swap_token_val = (int) 1073741824;
    }
    if(require_swap_token_interval<1) {
        Error("FixGlobals: 'SwapTokenInterval' is too small (min %d)",1);
        require_swap_token_interval = (int) 1;
    }
    if(require_swap_token_interval>7200) {
        Error("FixGlobals: 'SwapTokenInterval' is too large (max %d)",7200);
        require_swap_token_interval = (int) 7200;
    }
    if(require_disk_sweep_interval<1) {
        Error("FixGlobals: 'DiskSweepInterval' is too small (min %d)",1);
        require_disk_sweep_interval = (int) 1;
    }
    if(require_disk_sweep_interval>3600000) {
        Error("FixGlobals: 'DiskSweepInterval' is too large (max %d)",3600000);
        require_disk_sweep_interval = (int) 3600000;
    }
    if(require_n_dsweeper<1) {
        Error("FixGlobals: 'NumDiskSweeper' is too small (min %d)",1);
        require_n_dsweeper = (int) 1;
    }
    if(require_n_dsweeper>32) {
        Error("FixGlobals: 'NumDiskSweeper' is too large (max %d)",32);
        require_n_dsweeper = (int) 32;
    }
    if(require_n_msweeper<1) {
        Error("FixGlobals: 'NumMemorySweeper' is too small (min %d)",1);
        require_n_msweeper = (int) 1;
    }
    if(require_n_msweeper>32) {
        Error("FixGlobals: 'NumMemorySweeper' is too large (max %d)",32);
        require_n_msweeper = (int) 32;
    }
    if(require_num_bucket<4) {
        Error("FixGlobals: 'NumBucket' is too small (min %d)",4);
        require_num_bucket = (int) 4;
    }
    if(require_num_bucket>10240) {
        Error("FixGlobals: 'NumBucket' is too large (max %d)",10240);
        require_num_bucket = (int) 10240;
    }
    if(require_level_bucket_skiplist<2) {
        Error("FixGlobals: 'LevelBucketSkipList' is too small (min %d)",2);
        require_level_bucket_skiplist = (int) 2;
    }
    if(require_level_bucket_skiplist>64) {
        Error("FixGlobals: 'LevelBucketSkipList' is too large (max %d)",64);
        require_level_bucket_skiplist = (int) 64;
    }
    if(require_size_frag<4) {
        Error("FixGlobals: 'SizeFrag' is too small (min %d)",4);
        require_size_frag = (int) 4;
    }
    if(require_size_frag>1024000) {
        Error("FixGlobals: 'SizeFrag' is too large (max %d)",1024000);
        require_size_frag = (int) 1024000;
    }
    if(require_ratio_frag_length<2) {
        Error("FixGlobals: 'RatioFragLength' is too small (min %d)",2);
        require_ratio_frag_length = (int) 2;
    }
    if(require_ratio_frag_length>128) {
        Error("FixGlobals: 'RatioFragLength' is too large (max %d)",128);
        require_ratio_frag_length = (int) 128;
    }
    if(num_prefetch_referedpages<0) {
        Error("FixGlobals: 'NumPrefetchReferedPages' is too small (min %d)",0);
        num_prefetch_referedpages = (int) 0;
    }
    if(num_prefetch_referedpages>64) {
        Error("FixGlobals: 'NumPrefetchReferedPages' is too large (max %d)",64);
        num_prefetch_referedpages = (int) 64;
    }
    if(num_prefetch_includedobjs<0) {
        Error("FixGlobals: 'NumPrefetchIncludedObjects' is too small (min %d)",0);
        num_prefetch_includedobjs = (int) 0;
    }
    if(num_prefetch_includedobjs>128) {
        Error("FixGlobals: 'NumPrefetchIncludedObjects' is too large (max %d)",128);
        num_prefetch_includedobjs = (int) 128;
    }
}




void
ShowGlobals(FILE *s)
{
    fprintf(s,
"======================================================================\n\
Flag Name                            CurrentValue\n\
----------------------------------------------------------------------\n");

    fprintf(s, "  f  ConfigFile                      '%s'\n",
        config_filename);
    fprintf(s, "  l  LogFile                         '%s'\n",
        log_filename);
    fprintf(s, "  t  Trace                            %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  Monitor                          %s\n",
        monitor_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  P  Prefetch                         %s\n",
        prefetch_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  p  ServicePort                     %8d \n",
        service_portno);
    fprintf(s, "  -  LocalDomain                     '%s'\n",
        localdomain_padding);
    fprintf(s, "  R  ReportHook                      '%s'\n",
        report_hook_hostname);
    fprintf(s, "  n  NumSession                      %8d \n",
        require_sess_length);
    fprintf(s, "  H  NumHandler                      %8d \n",
        require_n_handler);
    fprintf(s, "  C  SessionChunkLength              %8d \n",
        require_chunk_size);
    fprintf(s, "  c  CacheDir                        '%s'\n",
        cache_dir_prefix);
    fprintf(s, "  -  RadixH2                         %8d \n",
        require_radix_h2);
    fprintf(s, "  -  RadixH3                         %8d \n",
        require_radix_h3);
    fprintf(s, "  -  DirSize                         %8d K (x 1024)\n",
        require_dir_size/(1024));
    fprintf(s, "  -  SwapTokenVal                    %8d \n",
        require_swap_token_val);
    fprintf(s, "  -  SwapTokenInterval               %8d \n",
        require_swap_token_interval);
    fprintf(s, "  -  DiskSweepInterval               %8d \n",
        require_disk_sweep_interval);
    fprintf(s, "  D  NumDiskSweeper                  %8d \n",
        require_n_dsweeper);
    fprintf(s, "  M  NumMemorySweeper                %8d \n",
        require_n_msweeper);
    fprintf(s, "  U  TotalMemorySize                 %8d M (x 1048576)\n",
        require_total_memory_size/(1048576));
    fprintf(s, "  -  NumBucket                       %8d \n",
        require_num_bucket);
    fprintf(s, "  -  LevelBucketSkipList             %8d \n",
        require_level_bucket_skiplist);
    fprintf(s, "  -  SizeFrag                        %8d \n",
        require_size_frag);
    fprintf(s, "  -  RatioFragLength                 %8d \n",
        require_ratio_frag_length);
    fprintf(s, "  r  NumPrefetchReferedPages         %8d \n",
        num_prefetch_referedpages);
    fprintf(s, "  i  NumPrefetchIncludedObjects      %8d \n",
        num_prefetch_includedobjs);
}

void
ShowGlobals_fd(int ofd)
{
    fdprintf(ofd,
"======================================================================\n\
Flag Name                        CurrentValue\n\
----------------------------------------------------------------------\n");

    fdprintf(ofd, "  f  ConfigFile                  '%s'\n",
        config_filename);
    fdprintf(ofd, "  l  LogFile                     '%s'\n",
        log_filename);
    fdprintf(ofd, "  t  Trace                        %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  -  Monitor                      %s\n",
        monitor_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  P  Prefetch                     %s\n",
        prefetch_flag==1 ? "*ON*" : "*OFF*" );
    fdprintf(ofd, "  p  ServicePort                 %8d \n",
        service_portno);
    fdprintf(ofd, "  -  LocalDomain                 '%s'\n",
        localdomain_padding);
    fdprintf(ofd, "  R  ReportHook                  '%s'\n",
        report_hook_hostname);
    fdprintf(ofd, "  n  NumSession                  %8d \n",
        require_sess_length);
    fdprintf(ofd, "  H  NumHandler                  %8d \n",
        require_n_handler);
    fdprintf(ofd, "  C  SessionChunkLength          %8d \n",
        require_chunk_size);
    fdprintf(ofd, "  c  CacheDir                    '%s'\n",
        cache_dir_prefix);
    fdprintf(ofd, "  -  RadixH2                     %8d \n",
        require_radix_h2);
    fdprintf(ofd, "  -  RadixH3                     %8d \n",
        require_radix_h3);
    fdprintf(ofd, "  -  DirSize                     %8d K (x 1024)\n",
        require_dir_size/(1024));
    fdprintf(ofd, "  -  SwapTokenVal                %8d \n",
        require_swap_token_val);
    fdprintf(ofd, "  -  SwapTokenInterval           %8d \n",
        require_swap_token_interval);
    fdprintf(ofd, "  -  DiskSweepInterval           %8d \n",
        require_disk_sweep_interval);
    fdprintf(ofd, "  D  NumDiskSweeper              %8d \n",
        require_n_dsweeper);
    fdprintf(ofd, "  M  NumMemorySweeper            %8d \n",
        require_n_msweeper);
    fdprintf(ofd, "  U  TotalMemorySize             %8d M (x 1048576)\n",
        require_total_memory_size/(1048576));
    fdprintf(ofd, "  -  NumBucket                   %8d \n",
        require_num_bucket);
    fdprintf(ofd, "  -  LevelBucketSkipList         %8d \n",
        require_level_bucket_skiplist);
    fdprintf(ofd, "  -  SizeFrag                    %8d \n",
        require_size_frag);
    fdprintf(ofd, "  -  RatioFragLength             %8d \n",
        require_ratio_frag_length);
    fdprintf(ofd, "  r  NumPrefetchReferedPages     %8d \n",
        num_prefetch_referedpages);
    fdprintf(ofd, "  i  NumPrefetchIncludedObjects  %8d \n",
        num_prefetch_includedobjs);
}

print_configration_FORM(int ofd)
{
    fdprintf(ofd,
"<form name=\"chk\" method=\"POST\" action=\"/config\">\n");

    fdprintf(ofd,
"<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n\
<tr bgcolor=\"#ffffcc\">\n\
    <th rowspan=\"2\"> name\n\
    <th colspan=\"2\"> current\n\
    <th rowspan=\"2\"> specified\n\
    <th rowspan=\"2\"> descriptions\n\
<tr bgcolor=\"#ffffcc\">\n\
    <th> value\n\
    <th> unit\n\
\n\
");

    fdprintf(ofd, "<tr><td>ConfigFile  <td colspan=\"2\"> %s <td><input name=\"ConfigFile\" value=\"%s\"> <td> path\n",
        config_filename, config_filename);
    fdprintf(ofd, "<tr><td>LogFile  <td colspan=\"2\"> %s <td><input name=\"LogFile\" value=\"%s\"> <td> path\n",
        log_filename, log_filename);
    fdprintf(ofd, "<tr><td>Trace  <td colspan=\"2\"> %s <td><input name=\"Trace\" value=\"%s\"> <td> on/off\n",
        trace_flag==1 ? "on" : "off",
        trace_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>Monitor  <td colspan=\"2\"> %s <td><input name=\"Monitor\" value=\"%s\"> <td> on/off\n",
        monitor_flag==1 ? "on" : "off",
        monitor_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>Prefetch  <td colspan=\"2\"> %s <td><input name=\"Prefetch\" value=\"%s\"> <td> on/off\n",
        prefetch_flag==1 ? "on" : "off",
        prefetch_flag==1 ? "on" : "off" );
    fdprintf(ofd, "<tr><td>ServicePort  <td align=\"right\"> %8d <td>  <td><input name=\"ServicePort\" value=\"%d\"> <td> max %d min %d\n",
        service_portno, service_portno, 32000, 1);
    fdprintf(ofd, "<tr><td>LocalDomain  <td colspan=\"2\"> %s <td><input name=\"LocalDomain\" value=\"%s\"> <td> path\n",
        localdomain_padding, localdomain_padding);
    fdprintf(ofd, "<tr><td>ReportHook  <td colspan=\"2\"> %s <td><input name=\"ReportHook\" value=\"%s\"> <td> path\n",
        report_hook_hostname, report_hook_hostname);
    fdprintf(ofd, "<tr><td>NumSession  <td align=\"right\"> %8d <td>  <td><input name=\"NumSession\" value=\"%d\"> <td> max %d min %d\n",
        require_sess_length, require_sess_length, N_SESS_MAX, 128);
    fdprintf(ofd, "<tr><td>NumHandler  <td align=\"right\"> %8d <td>  <td><input name=\"NumHandler\" value=\"%d\"> <td> max %d min %d\n",
        require_n_handler, require_n_handler, N_HANDLER_MAX, 1);
    fdprintf(ofd, "<tr><td>SessionChunkLength  <td align=\"right\"> %8d <td>  <td><input name=\"SessionChunkLength\" value=\"%d\"> <td> max %d min %d\n",
        require_chunk_size, require_chunk_size, SESS_CHUNK_SIZE, 1);
    fdprintf(ofd, "<tr><td>CacheDir  <td colspan=\"2\"> %s <td><input name=\"CacheDir\" value=\"%s\"> <td> path\n",
        cache_dir_prefix, cache_dir_prefix);
    fdprintf(ofd, "<tr><td>RadixH2  <td align=\"right\"> %8d <td>  <td><input name=\"RadixH2\" value=\"%d\"> <td> max %d min %d\n",
        require_radix_h2, require_radix_h2, 293, 3);
    fdprintf(ofd, "<tr><td>RadixH3  <td align=\"right\"> %8d <td>  <td><input name=\"RadixH3\" value=\"%d\"> <td> max %d min %d\n",
        require_radix_h3, require_radix_h3, 293, 3);
    fdprintf(ofd, "<tr><td>DirSize  <td align=\"right\"> %8d <td> K (x 1024) <td><input name=\"DirSize\" value=\"%d\"> <td> max %d min %d\n",
        require_dir_size/(1024), require_dir_size/(1024), 1048576, 1);
    fdprintf(ofd, "<tr><td>SwapTokenVal  <td align=\"right\"> %8d <td>  <td><input name=\"SwapTokenVal\" value=\"%d\"> <td> max %d min %d\n",
        require_swap_token_val, require_swap_token_val, 1073741824, 1);
    fdprintf(ofd, "<tr><td>SwapTokenInterval  <td align=\"right\"> %8d <td>  <td><input name=\"SwapTokenInterval\" value=\"%d\"> <td> max %d min %d\n",
        require_swap_token_interval, require_swap_token_interval, 7200, 1);
    fdprintf(ofd, "<tr><td>DiskSweepInterval  <td align=\"right\"> %8d <td>  <td><input name=\"DiskSweepInterval\" value=\"%d\"> <td> max %d min %d\n",
        require_disk_sweep_interval, require_disk_sweep_interval, 3600000, 1);
    fdprintf(ofd, "<tr><td>NumDiskSweeper  <td align=\"right\"> %8d <td>  <td><input name=\"NumDiskSweeper\" value=\"%d\"> <td> max %d min %d\n",
        require_n_dsweeper, require_n_dsweeper, 32, 1);
    fdprintf(ofd, "<tr><td>NumMemorySweeper  <td align=\"right\"> %8d <td>  <td><input name=\"NumMemorySweeper\" value=\"%d\"> <td> max %d min %d\n",
        require_n_msweeper, require_n_msweeper, 32, 1);
    fdprintf(ofd, "<tr><td>TotalMemorySize  <td align=\"right\"> %8d <td> M (x 1048576) <td><input name=\"TotalMemorySize\" value=\"%d\"> <td> max %d min %d\n",
        require_total_memory_size/(1048576), require_total_memory_size/(1048576), 4096, 8);
    fdprintf(ofd, "<tr><td>NumBucket  <td align=\"right\"> %8d <td>  <td><input name=\"NumBucket\" value=\"%d\"> <td> max %d min %d\n",
        require_num_bucket, require_num_bucket, 10240, 4);
    fdprintf(ofd, "<tr><td>LevelBucketSkipList  <td align=\"right\"> %8d <td>  <td><input name=\"LevelBucketSkipList\" value=\"%d\"> <td> max %d min %d\n",
        require_level_bucket_skiplist, require_level_bucket_skiplist, 64, 2);
    fdprintf(ofd, "<tr><td>SizeFrag  <td align=\"right\"> %8d <td>  <td><input name=\"SizeFrag\" value=\"%d\"> <td> max %d min %d\n",
        require_size_frag, require_size_frag, 1024000, 4);
    fdprintf(ofd, "<tr><td>RatioFragLength  <td align=\"right\"> %8d <td>  <td><input name=\"RatioFragLength\" value=\"%d\"> <td> max %d min %d\n",
        require_ratio_frag_length, require_ratio_frag_length, 128, 2);
    fdprintf(ofd, "<tr><td>NumPrefetchReferedPages  <td align=\"right\"> %8d <td>  <td><input name=\"NumPrefetchReferedPages\" value=\"%d\"> <td> max %d min %d\n",
        num_prefetch_referedpages, num_prefetch_referedpages, 64, 0);
    fdprintf(ofd, "<tr><td>NumPrefetchIncludedObjects  <td align=\"right\"> %8d <td>  <td><input name=\"NumPrefetchIncludedObjects\" value=\"%d\"> <td> max %d min %d\n",
        num_prefetch_includedobjs, num_prefetch_includedobjs, 128, 0);
    fdprintf(ofd,
"\n\
\n\
</table>\n\
\n");

    fdprintf(ofd,
"<input name=\"send\" type=\"submit\" value=\"APPLY\">\n\
<input name=\"clear\" type=\"reset\" value=\"CLEAR\">\n");


    fdprintf(ofd, "</form>\n");

}


void
ShowProgramStatus(FILE *outport)
{
    fprintf(outport, "KOTETU\n");

    fflush(outport);
}

void
ShowUsage()
{
    FixGlobals();   /* check in temporary */
    ShowProgramStatus(stdout);
    ShowGlobals(stdout);
}

#if 0
int
SetProcessComment(char *comment)
{
    char line[STRING_SIZE];

    sprintf(line, "wcol : %s", comment);
    return SetProcessState(line);
}
#endif


/*
 * ReadCommandlineOption
 */
void ReadCommandlineOptions(int argc, char **argv)
{
    int flag;

    extern char *optarg;
    extern int  optind;

    while((flag = getopt(argc, argv,
        "hvf:l:tPp:R:n:H:C:c:D:M:U:r:i:")) != -1) {

        switch(flag) {
        case 'h':
            ShowUsage();
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'v':
            ShowProgramStatus(stdout);
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'f':
            strcpy(config_filename, optarg);
            ReadConfigFile(config_filename);
            break;
        case 'l':
            log_filename = strdup(optarg);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'P':
            prefetch_flag = 1 - prefetch_flag;
            break;
        case 'p':
            service_portno = (int) atoi(optarg);
            break;
        case 'R':
            report_hook_hostname = strdup(optarg);
            break;
        case 'n':
            require_sess_length = (int) atoi(optarg);
            break;
        case 'H':
            require_n_handler = (int) atoi(optarg);
            break;
        case 'C':
            require_chunk_size = (int) atoi(optarg);
            break;
        case 'c':
            cache_dir_prefix = strdup(optarg);
            break;
        case 'D':
            require_n_dsweeper = (int) atoi(optarg);
            break;
        case 'M':
            require_n_msweeper = (int) atoi(optarg);
            break;
        case 'U':
            require_total_memory_size = (off_t) atoi(optarg) * 1048576;
            break;
        case 'r':
            num_prefetch_referedpages = (int) atoi(optarg);
            break;
        case 'i':
            num_prefetch_includedobjs = (int) atoi(optarg);
            break;
        }
    }
}



