/*
 * code-name 'KOTETU'
 *
 * proxy_main.c : main body for proxy server.
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1999, 2000
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#include <sys/poll.h>
#endif
#ifdef HAVE_SELECT_H
#include <sys/select.h>
#endif
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <strings.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <assert.h>
#include <errno.h>

/* private headers */

#include "range.h"
#include "msg.h"
#include "url.h"

#include "socket.h"
#include "sigs.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "misc.h"
#include "http.h"
#include "xsk.h"
#include "sess.h"
#include "squeue.h"
#include "parse.h"

#include "reason.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DEFAULT_CONFIG_ENVNAME
#define DEFAULT_CONFIG_ENVNAME "KOTETU"
#endif

#ifndef DEFAULT_CONFIGFILE
#define DEFAULT_CONFIGFILE      "/etc/kotetu.conf"
#endif
#ifndef DEFAULT_LOGFILE
#define DEFAULT_LOGFILE         "kotetu.log"
#endif

#ifndef MONITOR_INTERVAL
#define MONITOR_INTERVAL    (600)
#endif

#ifndef MALLOC
#define MALLOC(x)   malloc(x)
#endif
#ifndef FREE
#define FREE(x)     free(x)
#endif


/* dummy */

#define cache_show_blockarray(x,y)      ((void)0)
#define cache_list_entry(x)             ((void)0)
int nline;
int nblock;
int linesize;



pthread_mutex_t stat_lock = PTHREAD_MUTEX_INITIALIZER;
time_t boot_date = 0;
time_t service_start_date = 0;
int    session_queue_report_flag = 0;

volatile long long request_sirial_count = 0;
volatile int receive_request_count = 0;
volatile int process_request_count = 0;
volatile int issue_prefetch_count  = 0;
volatile int report_request_count  = 0;
volatile int reject_request_count  = 0;
volatile int ignore_request_count  = 0;

volatile int send_reply_count    = 0;
volatile int hit_reply_count     = 0;
volatile int miss_reply_count    = 0;
volatile int error_reply_count   = 0;

volatile int hash_colision      = 0;
volatile int compaction_count   = 0;
volatile int expire_count       = 0;

volatile int skip_cache_count   = 0;
volatile int alloc_tmp_buf_count = 0;
volatile int alloc_sess_buf_count = 0;

extern volatile int swapload_count;
extern volatile int swapin_count;
extern volatile int swapout_count;
extern volatile int swapload_success;
extern volatile int swapin_success;
extern volatile int swapout_success;

extern volatile int swapout_token;
extern volatile int swapout_token_init;
extern volatile int swapout_token_interval;

extern volatile int sweeper_seq;
extern volatile int sweeper_count;

int read_reply_total   = 0;
int read_reply_success = 0;
int read_reply_fail    = 0;

int cache_newover_count = 0;
int cache_newfree_count = 0;
int cache_newmin_count = 0;



volatile off_t intra_incoming = (off_t)0; /* requests from clients */
volatile off_t extra_outgoing = (off_t)0; /* requests to super class servers */
volatile off_t extra_incoming = (off_t)0; /* replys from super class servers */
volatile off_t intra_outgoing = (off_t)0; /* replys to clients */



sim_stat_rec rcode_stat;
long    rcode_stat_count[1000];

sim_stat_rec len_perread_stat;
static long server_buffer_stat_count[K(32)];

sim_stat_rec object_size_stat;
static long object_size_stat_count[K(1)];

sim_stat_rec url_first_hash_stat;
static long url_first_hash_stat_count[K(2)];
sim_stat_rec url_second_hash_stat;
static long url_second_hash_stat_count[K(2)];
sim_stat_rec url_third_hash_stat;
static long url_third_hash_stat_count[K(10)];
sim_stat_rec url_fourth_hash_stat;
static long url_fourth_hash_stat_count[K(10)];

sim_stat_rec object_alloc_stat;
static long object_alloc_stat_count[K(10)];


pthread_t th_listener;
pthread_mutex_t accept_lock = PTHREAD_MUTEX_INITIALIZER;

pthread_t th_handlers[N_HANDLER_MAX];
pthread_t th_monitor;

pthread_t th_signalhandler;

sigset_t handler_sig_mask;

int require_sess_length;

int think_queue_length;
sess_queue_t X_think_queue;
sess_queue_t *think_queue;
#if 0
sess_queue_t X_send_queue;
sess_queue_t *send_queue;
sess_queue_t X_fetch_queue;
sess_queue_t *fetch_queue;

sess_queue_t X_prompt_queue;
sess_queue_t *prompt_queue;
#endif

sess_queue_t X_wait_queue;
sess_queue_t *wait_queue;

sess_queue_t X_msweep_queue;
sess_queue_t *msweep_queue;


#ifdef CHECK_QUEUE

#define push_sess(q,c) \
	RawTrace("push sess# %d, %p to %p : %d/%d : %s %5d\n",\
		c->sid, c, q, q->count, q->cmax, __FILE__ , __LINE__), sq_push(q,c) 
#define pull_sess(q) \
	RawTrace("pull from %p : %s %5d\n", q, __FILE__ , __LINE__), sq_pull(q,c) 

#else
#define push_sess(q,c) sq_push(q,c)
#define pull_sess(q) sq_pull(q)
#endif
#define pull_sess_nonblock(q) sq_pull_nonblock(q)



sim_hist_rec comp_hist;
static long comp_hist_area[128];
sim_hist_rec expire_hist;
static long expire_hist_area[128];


/*
 * externals
 */

extern int nhot;

extern int radix_h1, radix_h2, radix_h3;
extern int *dir_sig;

#if 0
#define push_sess(q,c) sq_timedpush(q,c,100000) /* 100 msec */
#endif


int InitSession(int);


/* automatic */
pthread_mutex_t config_lock = PTHREAD_MUTEX_INITIALIZER;

char config_filename[L_PATH];
char *log_filename;
int trace_flag;
int monitor_flag;
int prefetch_flag=1;
int service_portno;
int require_n_handler;
int require_n_dsweeper;
int require_n_msweeper;
int require_chunk_size;
off_t require_total_memory_size;

int require_num_bucket;
int require_level_bucket_skiplist;
int require_size_frag;
int require_ratio_frag_length;

int num_prefetch_referedpages;
int num_prefetch_includedobjs;

#ifndef MAX_FRAG_PER_OBJ
#define MAX_FRAG_PER_OBJ    (256)
#endif
#ifndef MAX_FRAG_PER_WRITE
#define MAX_FRAG_PER_WRITE  (32)
#endif
int max_frag_per_obj    = MAX_FRAG_PER_OBJ;
int max_frag_per_write  = MAX_FRAG_PER_WRITE;

int last_keep_depth;
int cache_normal_idle_th_base   = 3600;
int cache_normal_idle_th        = 3600;
int cache_cracked_idle_th_base  =  600;
int cache_cracked_idle_th       =  600;

char *localdomain_padding;
char *report_hook_hostname;

/* static */
pid_t    wcol_pid;
char     *progname;


extern off_t sess_memory_size;
off_t cache_memory_size;


extern int sq_check_count();

int
check_sq(char *msg)
{
	Trace("# %s: sq_count(free) %d, diff %d, think Q %d/%d, wait Q %d/%d, msweep Q %d/%d\n",
		msg,
		sq_check_count(),
		sq_check_count()
			- (think_queue->count + wait_queue->count + msweep_queue->count),
		think_queue->count, think_queue->cmax,
		wait_queue->count, wait_queue->cmax,
		msweep_queue->count, msweep_queue->cmax);
	return 0;
}

int
CHKTIME(time_t thetime)
{
    int diff;
    diff = time(NULL) - thetime;
    if(diff<0) diff = -diff;
    if(diff>86400) {
        return 0;
    }
    else {
        return 1;
    }
}

void
dump_bkinfo()
{
	int i;
	bucket_t *b;

	for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
		Trace(" %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
			i,
			b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
			b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
	}
}

int
check_bkinfo()
{
	int i;
	bucket_t *b;
	int r, c;

	c = 0;
	for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
		r = 0;
		if(b->entry_freelist==NULL && b->entry_use <  b->entry_hwater) {
			r++;
		}
		if(b->frag_freelist==NULL && b->frag_use <  b->frag_hwater) {
			r++;
		}

		if(r) {
		  Trace(" be %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
			i,
			b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
			b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
			c++;
		}
	}

	if(c==0) 
		return 1;
	else
		return 0;
}

pthread_mutex_t wait_check_lock;

inline int check_timeout(sess_t *sessp, time_t now);

int
check_friend_lock(sess_t *this)
{
	sess_t *cu;
	int n;
	int i;
	int chk;
	int w, d;
	time_t now;
	int idle;

	if(this==NULL || this->cache_entry==NULL) {
		return -1;
	}
Trace("check_friend_lock:\n");
Trace("  target sess# %d, oid# %d\n",
	this->sid, ((skcache_entry_t*)this->cache_entry)->oid);


	MUTEX_LOCK(&wait_check_lock);

	now = time(NULL);
	
	n = wait_queue->count*2; /* XXX */
	i = 0;
	w = 0;
	d = 0;
	while(i<n) {
		cu = pull_sess_nonblock(wait_queue);
		if(cu==NULL) {
			break;
		}
		idle = now - cu->last_action;

#if 0
  if(cu->cache_entry) {
 	Trace("  %3d: sess# %d, %p %s : entry %p, oid# %d, ref %d, owner %p, idle %d\n",
	i,
	cu->sid, cu, MSG_STATE_[cu->state],
	cu->cache_entry,
	((skcache_entry_t*)cu->cache_entry)->oid,
	((skcache_entry_t*)cu->cache_entry)->ref,
	((skcache_entry_t*)cu->cache_entry)->owner, idle
	);
  }
#endif

        if(check_timeout(cu, now)) {
            cu->state = STATE_QUERY_ADDR;
            cu->cachable = SESS_UNCACHABLE;

            chk = push_sess(think_queue, cu);
            if(chk) {
Trace("timeout waited friend sess# %d from wait Q, fail to push think Q\n",
                    cu->sid);
                d++;

				DoSendMessage(cu);
				DoCloseSess(cu);
				DoReleaseSess(cu);
				cu = NULL;

            }
            else {
Trace("timeout waited friend sess# %d from wait Q, push think Q\n",
                    cu->sid);
                w++;
            }
        }
        else
		if(cu->cache_entry &&
			((skcache_entry_t*)cu->cache_entry)->oid == 
				((skcache_entry_t*)this->cache_entry)->oid) {

			/* re-push to think Q */
			chk = push_sess(think_queue, cu);
			if(chk) {
Trace("pick up waited friend sess# %d from wait Q, fail to push think Q\n",
					cu->sid);
				d++;

				DoSendMessage(cu);
				DoCloseSess(cu);
				DoReleaseSess(cu);
				cu = NULL;

			}
			else {
Trace("pick up waited friend sess# %d from wait Q, push think Q bk# %d,\n",
					cu->sid, cu->fhash);
				w++;
			}
			cu = NULL;
		}
		else {
			chk = push_sess(wait_queue, cu);
			if(chk) {
				DoSendMessage(cu);
				DoCloseSess(cu);
				DoReleaseSess(cu);

			}
			cu = NULL;
		}
		i++;
	}

	MUTEX_UNLOCK(&wait_check_lock);

	Trace("check_friend_lock: sess# %d, oid# %d, n %d, w %d, d %d\n",
		this->sid, ((skcache_entry_t*)this->cache_entry)->oid,
		n, w, d);

	return w;
}

/*
 * release_buf - release allocated area
 *
 */
int
release_buf(sess_t *this)
{
    int ret;
    int chk;
    bucket_t *bk;
    skcache_entry_t *centry;


#ifdef TRACE_CACHE
    Trace("release_buf: sess# %d, '%s' [%d:%d] %s\n",
        this->sid, this->loc, this->fhash, this->shash,
        MSG_STATE_[this->state]);

#if 0
    cache_showblockaddr(this->fhash, 0);
#endif

#endif

    bk = &skcache_bucket_base[this->fhash];

    LOCK_BK(bk);
 assert(check_bkentry(bk, "rb0"));


    centry = this->cache_entry;
    if(centry) {
        centry->ref--;

#if 0
        if(centry->owner==this->sid) {
            centry->owner = 0;
        }
#endif
		if(centry->owner==this) {
            centry->owner = NULL;
		}

		if(centry->ref>0) {
#ifdef USE_WQUEUE
			check_friend_lock(this);
#endif
		}

		this->cache_entry = NULL;
    }

    ret = 0;

 assert(check_bkentry(bk, "rb1"));
	UNLOCK_BK(bk);
	bk = NULL;

out:

    return ret;
}



int
DoAccept(int listen_fd)
{
    struct sockaddr_in from;
    int len;
    int tmpfd;


    len = sizeof(from);

    MUTEX_LOCK(&accept_lock);
    tmpfd = accept(listen_fd, (struct sockaddr*)&from, &len);
    MUTEX_UNLOCK(&accept_lock);

    if(tmpfd<0) {
#if 1
        Error("accept error (%d)", errno);
#endif
        tmpfd = -1;
        goto done;
    }

    MUTEX_LOCK(&stat_lock);
    receive_request_count++;
    request_sirial_count++;
    MUTEX_UNLOCK(&stat_lock);

done:

    return tmpfd;
}


int
DoReadReq(sess_t *this)
{
    int nread;
    int c;
    int method;
    char url[L_URL];
    char version[L_URL];
    void *cpos;
    int chk;
    int err;
    int ims;

#if 0
    Trace("DoReadReq : sess# %d, cfd %d\n", this->sid, this->cfd);
#endif

#if 0
dump_stringsegment("before", this->reqbuf, this->reqlen);
#endif

try_read:
    nread = read(this->cfd,
                &this->reqbuf[this->reqlen], L_REQBUF-this->reqlen);
    err = errno;
    if(nread<0) {
        if(err==EINTR) {
            goto try_read;
        }
        Error("sess# %d, read() fail (%d)", this->sid, err);
        DoHalfClientCloseSess(this);

        this->reason = R_FAIL_CLIENT_READ;
        this->state = STATE_MESSAGE;

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        return -1;
    }
    if(nread==0) {
        Error("sess# %d, read() empty", this->sid);
#if 0
        Error("sess# %d, th %x, fd# %d, empty client request (%d)",
            this->sid, pthread_self(), this->cfd, err);
#endif
        DoHalfClientCloseSess(this);

        this->reason = R_FAIL_CLIENT_READ_EMPTY;
        this->state = STATE_MESSAGE;

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        return -1;
    }

    this->reqlen += nread;
    this->cread += nread;
    this->loc[0] = '\0';
#if 0
    this->last_action = time(NULL);
#endif

#if 0
dump_stringsegment("o.req.", this->reqbuf, this->reqlen);
#endif

    c = HTTP_find_header_term(this->reqbuf, this->reqlen);

#if 0
Trace("DoReadReq: sess %p, read %d, header term %d\n", this->sid, nread, c);
#endif

    if(c>0) {
#if 0
        dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

#if 0
 Trace(" 1 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

        version[0] = '\0';
        ims = -1;

        c = HTTP_parse_request_header(this->reqbuf,
                &method, url, version,
                &ims, &this->cachable, &this->reload);

        if(c) {
            Error("ignore request from fd #%d", this->cfd);
#if 1
            dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

            if(c==-2) {
                this->reason = R_TOOLONG_URL;
            }
            else {
                this->reason = R_IGNORE_URL;
            }

            this->state = STATE_MESSAGE;
            return -1;
        }

        chk = -1;
#if 1
#if 0
        Trace("sess# %d, version '%s'\n", this->sid, version);
#endif
        if(version[0]) {
            chk =
             HTTP_parse_version(version, &this->cver_major, &this->cver_minor);
        }
        if(chk<0) {
            this->cver_major = 0;
            this->cver_minor = 9;
        }
#if 0
        Trace("sess# %d, chk %d, client HTTP ver %d.%d\n",
            this->sid, chk, this->cver_major, this->cver_minor);
#endif
#endif

/*
Trace("sess# %d, client HTTP ver %d.%d\n",
    this->sid, this->cver_major, this->cver_minor);
*/

#ifdef TRACE_SESS
        Trace("sess# %d, <client> method <%s>,URL '%s' version <%s>\n",
            this->sid, MSG_M_[method], url, version);
#endif

        this->method = method;

        strcpy(this->loc, url);
#if 0
        this->fhash = first_hash(url);
        this->shash = second_hash(url);
        this->thash = third_hash(url);
        this->ffhash = fourth_hash(url);
#endif
		calc_hash(this->loc,
			&this->fhash, &this->shash, &this->thash, &this->ffhash);

#if 0
Trace("rreq: %03d:%03d:%03d:%03d %s\n", 
	this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif

#if 0
        sim_stat_add_value(&url_first_hash_stat,  this->fhash);
        sim_stat_add_value(&url_second_hash_stat, this->shash);
        sim_stat_add_value(&url_third_hash_stat,  this->thash);
#endif

/*****
 ***** check cachable or not
 *****/

        if(istempURL(url)) {
            this->cachable  = SESS_UNCACHABLE;
            this->reload    = SESS_RELOAD;
        }

#if 0
        if(this->cachable == SESS_UNCACHABLE) {
            Trace(" 4 sess# %d, %s\n", this->sid, this->loc);
            /*
            dump_string_chunk("req.", this->reqbuf);
            */
            dump_stringsegment("=req.", this->reqbuf, this->reqlen);
        }
#endif


        if(ims>0) {
            this->ims = ims;
        }

#if 0
Trace("sess# %d, ver HTTP/%d.%d, %c%c, ims %d\n",
        this->sid,
        this->cver_major, this->cver_minor,
        (this->cachable==SESS_CACHABLE ? 'c' : 'N'),
        (this->reload==SESS_RELOAD ? 'R' : '-'),
        this->ims);
#endif

#if 0
 Trace(" 3 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif


done_cachable_check:
#if 0
 Trace(" 4 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

        this->state = STATE_THINK;
    }
    else {
        Error("DoReadReq: HTTP_find_header_term() returns %d", c);
#if 1
        dump_stringsegment("e.req", this->reqbuf, this->reqlen);
#endif
    }
#if 0
 Trace(" 5 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

    return 0;
}


int
DoSetReq(sess_t *this, char *aurl)
{
    int nread;
    int c;
    int method;
    char url[L_URL], version[L_URL];
    char protocol[L_URL], portstr[L_URL];
    void *cpos;
    int istmp;


#if 0
    Trace("DoSetReq : cfd %d\n", this->cfd);
#endif

    this->reqlen = sprintf(this->reqbuf,
            "GET %s HTTP/1.0\r\nAccept: */*\r\n\r\n", aurl);

    this->loc[0] = '\0';

#if 0
    this->last_action = time(NULL);
#endif

    c = HTTP_parse_simple_request(this->reqbuf, &method, url, version);
    if(c) {
        Error("ignore request from fd #%d", this->cfd);
#if 1
        dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

        this->reason = R_IGNORE_URL;
        this->state = STATE_MESSAGE;
        return -1;
    }

#if 0
    Trace("receive request\n");
    Trace("  method  <%s>\n  URL     <%s>\n  version <%s>\n",
        method, url, version);
#endif

    this->method = method;

    strcpy(this->loc, url);
#if 0
    this->fhash = first_hash(url);
    this->shash = second_hash(url);
    this->thash = third_hash(url);
    this->ffhash = fourth_hash(url);
#endif

	calc_hash(this->loc,
		&this->fhash, &this->shash, &this->thash, &this->ffhash);

#if 0
Trace("sreq: %03d:%03d:%03d:%03d %s\n", 
	this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif


#if 0
    sim_stat_add_value(&url_first_hash_stat,  this->fhash);
    sim_stat_add_value(&url_second_hash_stat, this->shash);
    sim_stat_add_value(&url_third_hash_stat, this->thash);
#endif

    this->state = STATE_THINK;

    return 0;
}



int
DoReadReqFirst(sess_t *this)
{
    int ret;
    int b_state;


    b_state = this->state;
    this->state = STATE_RECV_REQ;
    ret = DoReadReq(this);
#if 0
    Trace("DoReadReqFirst: sess# %d, %s -> %s , ret %d\n",
        this->sid, MSG_STATE_[b_state], MSG_STATE_[this->state], ret);
#endif

    return ret;
}


int
DoReadReqRest(sess_t *this)
{
    int ret;
    int b_state;


    b_state = this->state;
    ret = DoReadReq(this);
#if 0
    Trace("DoReadReqRest: sess# %d, %s -> %s , ret %d\n",
        this->sid, MSG_STATE_[b_state], MSG_STATE_[this->state], ret);
#endif

    return ret;
}

int
HostnameNormalize(char *src)
{
    register char *p;
    register int c;
    int d;

    p = src;
    c = 0;
    while(*p) {
        c++;
        if(*p=='.') {
            goto have_dot;
        }
        p++;
    }

not_have_dot:

    MUTEX_LOCK(&config_lock);
    if(c+strlen(localdomain_padding)+1<L_HOST) {
        strcat(p, localdomain_padding);
    }
    MUTEX_UNLOCK(&config_lock);

    return 0;

have_dot:
    return 0;
}


int check_bkentry(bucket_t*,char*);


int
call_msweeper(sess_t *this,  bucket_t *bk, int e_or_f)
{
	int ret;
	int chk;
	char *msg;

#if 1
	Trace(
"bk# %d, %p, push %d, skiplist %4d/%d, entry %4d/%d, frag %4d/%d\n",
		bk->bid, bk, bk->gc_push,
		bk->cup.num, bk->elemnum,
		bk->entry_use, bk->entry_all,
		bk->frag_use,  bk->frag_all);

	if(bk->entry_freelist==NULL) {
		msg = "no";
	}
	else {
		msg = "few";
	}
 assert(check_bkentry(bk, "cs1"));

	Trace("th %08x, %s free entry, push to sweeper %d/%d\n",
		pthread_self(), msg,
		msweep_queue->count, msweep_queue->cmax);
#endif

	ret = 0;

#if 1
	if(bk->gc_push>=0) {
        sess_t *vp;

 assert(check_bkentry(bk, "cs2"));

        vp = (sess_t*) sq_get_free();
		if(vp==NULL) {
			goto no_vs_1;
		}
        vp->last_action = this->last_action;
        vp->fhash = this->fhash;
        vp->shash = vp->thash = vp->ffhash = -1;
#if 0
        vp->mode  = XSK_ENTRY;
#endif
        vp->mode  = e_or_f;

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
 assert(check_bkentry(bk, "cs3"));

#if 0
		/* XXX */
		/* dead lock spot */
		chk = push_sess(msweep_queue, vp);
#else
        chk = sq_timedpush(msweep_queue, vp, 5000);	/* 5msec */
#endif
        if(chk) {
            Error("fail push to msweep_queue <1f> %d/%d",
				msweep_queue->count, msweep_queue->cmax);

			sq_be_free(vp);
			vp = NULL;

			ret = -2;
			goto out;
        }
		else {
            Trace("  push to mseep_queue <1c> %d/%d\n",
				msweep_queue->count, msweep_queue->cmax);
		}

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
 assert(check_bkentry(bk, "cs4"));
no_vs_1:
		(void)1;

    }
#endif

	bk->gc_push++;

#if	1
	if(
		( bk->elem_freelist==NULL
			|| bk->entry_freelist==NULL
			|| bk->frag_freelist==NULL )
	  ||
		bk->gc_push>10
	  )
    {
		int tc;

		Trace("call_msweeper: self compaction, bk# %d, %p, %d\n",
			bk->bid, bk, e_or_f);

		tc = 1;
		chk = skcache_compaction(bk, -79, 1, e_or_f);
		while(chk==0 && tc<2) {
#if 1
			struct timespec w;

			w.tv_sec = 0;
			w.tv_nsec = 10 * 1000000;	/* 10msec */

			nanosleep(&w, NULL);
#endif
			tc++;

			Trace("call_msweeper: retry self compaction, %d, bk# %d, %p\n",
				tc, bk->bid, bk);
			chk = skcache_compaction(bk, -291, 1, XSK_ENTRY);
		}
		if(chk<=0) {
			Error("call_msweeper: fail or nothing skcache_compaction() %d", chk);
			ret = 1;
		}
	}
#endif

out:
	return ret;
}


int
DoThink(sess_t *this)
{
    int ihave;
    int chk;
    char protocol[L_URL], portstr[L_URL];

    bucket_t *bk;
    skcache_entry_t *stored_pos;
    int check_already_have;
    skcache_entry_t *newone;
/*
    skcache_entry_t *tmp;
*/

    int sig_pos;

    struct timeval xbefore, xafter;
	extern int sq_check_count();


#if 0
	Trace("sq_count %d, think Q %d/%d, wait Q %d/%d, msweep Q %d/%d\n",
		sq_check_count(),
		think_queue->count, think_queue->cmax,
		wait_queue->count, wait_queue->cmax,
		msweep_queue->count, msweep_queue->cmax);
#endif


    chk = ParseURL(this->loc, protocol,
            this->servername, portstr, this->serverpath);
    if(portstr[0]=='\0') {
        this->serverport = 80;
    }
    else {
        this->serverport = atoi(portstr);
    }
    chk = HostnameNormalize(this->servername);

#if 0
    Trace("DoThink: '%s' [%04d:%04d:%04d] '%s' %s %s\n",
        this->loc, this->fhash, this->shash, this->thash, this->loc,
        this->cachable==SESS_CACHABLE ? "cachable" : "no-cachable",
        this->reload==SESS_RELOAD ? "reload" : "not-reload");
#endif

    /* this system permit only GET and POST method */
    if(! (this->method == M_GET || this->method == M_POST) ) {
        Trace("ignore method, sess %p (#%d)\n", this, this->sid);

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->reason = R_UNSUPPORT_METHOD;
        this->state = STATE_MESSAGE;
        return 0;
    }

#if 1
    if(strncasecmp(this->loc, "http:", 5)!=0) {
        this->reason = R_UNSUPPORT_PROTOCOL;
        this->state = STATE_MESSAGE;
        return 0;
    }
#endif

    if(strcmp(this->servername, report_hook_hostname)==0)
    {
#if 0
        Trace("report request. sess# %d %p\n", this->sid, this);
#endif

        MUTEX_LOCK(&stat_lock);
        report_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_REPORT;
        this->cachable = SESS_UNCACHABLE;
        this->reload   = SESS_RELOAD;
        return 0;
    }


    sim_stat_add_value(&url_first_hash_stat,  this->fhash);
    sim_stat_add_value(&url_second_hash_stat, this->shash);
    sim_stat_add_value(&url_third_hash_stat,  this->thash);
    sim_stat_add_value(&url_fourth_hash_stat,  this->ffhash);

    switch(this->method) {
    /* GET method */
    case M_GET:

get_first:

        bk = &skcache_bucket_base[this->fhash];
		LOCK_BK(bk);

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
 assert(check_bkentry(bk, "bs2"));

#if 0
	Trace("think: * sess# %d, bk# %d, %p\n",	
		this->sid, bk->bid, bk);
#endif

    if(bk->elem_freelist==NULL
        || bk->entry_freelist==NULL || bk->entry_use > bk->entry_hwater
        || bk->frag_freelist==NULL || bk->frag_use > bk->frag_hwater)
    {
#if 1
		Trace("think: memory check bk# %d,\n", bk->bid);
#endif
		chk = call_msweeper(this, bk, XSK_ENTRY);
		
		if(chk == -2) {
			Trace("think: sess# %d, fail to push sweeper Q, bk# %d,\n",
				this->sid, bk->bid);

			UNLOCK_BK(bk);

			goto get_first;
		}

		if(chk) {
			Trace("think: sess# %d, no space bk# %d,\n",
				this->sid, bk->bid);

            this->state  = STATE_MESSAGE;
            this->reason = R_MEMORY_OVERFLOW;

			goto get_last;
		}
	}


try_get_newone:
 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
        newone = skcache_pull_new_entry(bk);

 assert(check_bkinfo());
 assert(check_bkentry(bk, "bs5"));


        if(newone==NULL) {
            Error("think: sess# %d, no free element in entry free list of bk# %d, %p <2b>",
				this->sid, this->fhash, bk);

            this->state  = STATE_MESSAGE;
            this->reason = R_MEMORY_OVERFLOW;

			goto get_last;
        }

 assert(bk->cup.num==bk->entry_use-1);
 assert(check_bkinfo());

#if 0
		Trace(
"request %p sess# %-7d,      %03d:%03d:%03d:%03d %s\n",
	this, this->sid,
	this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif

#if 0
		Trace("sess# %d, TH magic %02x %02x\n",
			this->sid, newone->magic[0], newone->magic[1]);
#endif

		newone->oid = -1;
        newone->h1 = this->fhash;
        newone->h2 = this->shash;
        newone->h3 = this->thash;
        newone->h4 = this->ffhash;
        strcpy(newone->name, this->loc);

#if 0
		if(newone) {
			Trace(
"newone %p oid# %-7d, ref %d, %03d:%03d:%03d:%03d %s\n",
 	newone, newone->oid, newone->ref,
 	newone->h1, newone->h2, newone->h3, newone->h4,
	newone->name);
		}
		else {
			Trace(
"newone     NULL\n");
		}
#endif


        gettimeofday(&xbefore, NULL);
        stored_pos =
            (skcache_entry_t*) sk_assign(&bk->cup, newone,
                &bk->elem_freelist, skcache_entry_cmp,
                &check_already_have, 0);
        gettimeofday(&xafter, NULL);


#if 0
	Trace("th: sess# %d, stored_pos %p, check_already_have %d\n",
			this->sid, stored_pos, check_already_have);
#endif

 assert(check_bkinfo());

        if(stored_pos==NULL || check_already_have<0) {
            Error("think: cannot assign to entry free list of bk# %d <1b>",
                    this->fhash);

			Trace("think: stored_pos %p, check_already_have %d\n",
				stored_pos, check_already_have);

            this->state  = STATE_MESSAGE;
            this->reason = R_UNKNOW_ERROR;

#if 0
/* XXX */
            break;
#endif
			goto get_last;
        }

		/* stored_pos != NULL */
	
		if(stored_pos->ref<0) {
			Error("dead object **STRANGE** entry %p", stored_pos);
		}
		
		if(stored_pos!=newone) {
			/* HIT */
            /* 'newone' is not stored in bucket, use old one */
            skcache_push_idle_entry(bk, newone);
            newone = NULL;
#if 0
				Trace(
"hit    %p oid# %d, ref %d, m %p, %03d:%03d:%03d:%03d %s\n",
 	stored_pos, stored_pos->oid, stored_pos->ref,
	stored_pos->m_recorded,
 	stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
	stored_pos->name);
#endif
		}
		else {
			/* MISS */
            /* 'newone' is stored in bucket */
            newone = NULL;
#if 0
				Trace(
"miss   %p oid# %d, ref %d, m %p, %03d:%03d:%03d:%03d %s\n",
 	stored_pos, stored_pos->oid, stored_pos->ref,
	stored_pos->m_recorded,
 	stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
	stored_pos->name);
#endif
		}



 assert(bk->cup.num==bk->entry_use);

		/* not found in memory
		 * the object is allocated new area
		 */
		if(check_already_have==0 && stored_pos->m_recorded==NULL) {
			int xb, xa;
			int sig;

			sig_pos = stored_pos->h1*radix_h2*radix_h3
				+ stored_pos->h2*radix_h3 + stored_pos->h3;

			MUTEX_LOCK(&stat_lock);
			sig = dir_sig[sig_pos];
			MUTEX_UNLOCK(&stat_lock);

#if 0
			Trace("sig_pos %6d dir_sig[] %8x & h4 %8x -> %8x\n",
				sig_pos, sig, stored_pos->h4,
				sig & stored_pos->h4);
#endif
#if 0
			sig = -1;
#endif

			/*  if sig is negative. it means unknown directory.
				you should access the directory.
				if masked bit is same bits of target,
				the directory have or had the object in several seconds.
				you should access the directory, also.
			 */
            if(sig<0 ||
				(sig>0 && 
					(sig & (stored_pos->h4+1)) == (stored_pos->h4+1) ) ) {

                xb = bk->entry_use;

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
 assert(check_bkentry(bk, "bs8"));

                chk = skcache_entry_swapload(bk, stored_pos);
                if(chk==0) {
#if 0
                    Trace("sess# %d, swapload is success %p\n",
                        this->sid, stored_pos);
#endif

#if 1
                    if(stored_pos->oid<0) {
                    Trace("negative swaped entry %p oid# %d, ref %d <%d>\n",
                            stored_pos, stored_pos->oid, stored_pos->ref,
                            check_already_have);
                    }
#endif
                    check_already_have=1;
                }
                else {
#if 0
                    Trace("sess# %d, swapload is fail %p\n",
                        this->sid, stored_pos);
#endif
                }
 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
 assert(check_bkentry(bk, "bs9"));

                xa = bk->entry_use;

#if 0
Trace("= %d %d : %d -> %d /%d\n", chk, xa-xb, xb, xa, bk->entry_all);
#endif

            }

		}


        this->cache_entry = stored_pos;
        stored_pos->ref++;
        stored_pos->hot = xafter.tv_sec;
 assert(CHKTIME(stored_pos->hot));
 assert(check_bkentry(bk, "bs6"));

#if 0
Trace("DoThink: sess# %d, cache_entry %p, owner %p, ref %d\n",
        this->sid, stored_pos, stored_pos->owner, stored_pos->ref);
#endif
#if 0
Trace("DoThink: sess# %d, entry %p, oid# %d, %s(%d), owner %p, ref %d\n",
        this->sid, stored_pos, stored_pos->oid,
        (check_already_have==1) ? "USED" :
            ((check_already_have== -1) ? "ERROR" : "NEW"),
        check_already_have,
        stored_pos->owner, stored_pos->ref);
#endif


 assert(check_bkinfo());
 assert(check_bkentry(bk, "bs7"));


#if 1
        if(check_already_have==0) {
        }
        else {
            /* error check */
            if(stored_pos->oid<0) {
                Trace("negative memory entry %p oid# %d, ref %d <%d>\n",
                    stored_pos, stored_pos->oid, stored_pos->ref,
                    check_already_have);
            }
        }
#endif


  if(this->reload!=SESS_RELOAD && this->cachable==SESS_CACHABLE) {

            if(check_already_have==0) {
#if 0
Trace("  *UNKNOWN*\n");
#endif


                /* it is new object, initialize attributes */
#if 0
                stored_pos->owner       = this->sid;
#endif
                stored_pos->owner       = this;
                stored_pos->oid         = this->sid;
                stored_pos->issue_time  = this->last_action;

#if 0
                strcpy(stored_pos->name, this->loc);
#endif

                stored_pos->totallength = (off_t) 0L;
                stored_pos->m_recorded = NULL;
                stored_pos->d_recorded = -1;


#if 0
Trace("  stored_pos        %p m_recorded %p\n",
        stored_pos, stored_pos->m_recorded);
Trace("  this->cache_entry %p m_recorded %p\n",
        this-> cache_entry, ((skcache_entry_t*)this->cache_entry)->m_recorded);
#endif

                MUTEX_LOCK(&stat_lock);
                miss_reply_count++;
                MUTEX_UNLOCK(&stat_lock);

                this->state = STATE_QUERY_ADDR;

            }
            else {
#if 0
Trace("  *KNOWN*\n");
#endif
#if 0
				Trace("known object: request sess# %d, %d:%d:%d:%d %s\n\
found %p oid# %d, ref %d, %d:%d:%d:%d %s\n",
 	this->sid,
	this->fhash, this->shash, this->thash, this->ffhash, this->loc,
 	stored_pos, stored_pos->oid, stored_pos->ref,
 	stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
	stored_pos->name);

#endif


                if(stored_pos->ref>1) {
                    this->state = STATE_WAIT_FRIEND;
#if 1
                    Trace("sess# %d, wait friend session with %p %p\n",
                        this->sid, this->cache_entry, stored_pos);

        Trace("stored %p oid# %4d ref %3d, %4d %4d %4d %4d %s\n",
            stored_pos, stored_pos->oid, stored_pos->ref,
            stored_pos->h1, stored_pos->h2,
            stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

                }
                else
                if(stored_pos->owner==this) {
#ifdef TRACE_SESS
                    Trace("sess# %d, cachable, not have [%d:%d]\n",
                        this->sid, this->fhash, this->shash);
#endif

                    MUTEX_LOCK(&stat_lock);
                    miss_reply_count++;
                    MUTEX_UNLOCK(&stat_lock);

                    this->state = STATE_QUERY_ADDR;


                }
                else {

                    if(stored_pos->cracked) {
                        /* cracked object
                         * don't warry other session
                         */
#ifdef TRACE_SESS
                        Trace("sess# %d, cachable, have and cracked [%d:%d]\n",
                            this->sid, this->fhash, this->shash);
#endif
                        MUTEX_LOCK(&stat_lock);
                        miss_reply_count++;
                        MUTEX_UNLOCK(&stat_lock);

                        this->state = STATE_QUERY_ADDR;

                    }
                    else
                    if(stored_pos->owner==NULL) {
                        /* the object is not included other sessions
                         * this session can use the object
                         */
#ifdef TRACE_SESS
                        Trace("sess# %d, cachable, have [%d:%d], cold\n",
                            this->sid, this->fhash, this->shash);
#endif
#if 0
    cache_showblockaddr(this->fhash, 1);
#endif
#if 0
                        if(stored_pos->expire>0) {
                        Trace("sess# %d, expire %d, now %d\n",
                            this->sid, stored_pos->expire, xafter.tv_sec);
                        }
#endif
                        if(stored_pos->expire>0 &&
                            xafter.tv_sec > stored_pos->expire) {

                            MUTEX_LOCK(&stat_lock);
                            miss_reply_count++;
                            MUTEX_UNLOCK(&stat_lock);

                            this->state = STATE_QUERY_ADDR;

                        }
                        else {
                            MUTEX_LOCK(&stat_lock);
                            hit_reply_count++;
                            MUTEX_UNLOCK(&stat_lock);

                            this->state = STATE_SEND_CACHE;
                        }
                    }
                    else {
                        /* the object is included another session
                         * this session have to wait complete of that
                         */
#ifdef TRACE_SESS
                        Trace("sess# %d, cachable, have [%d:%d], hot\n",
                            this->sid, this->fhash, this->shash);
#endif
#if 1
                        Trace("sess# %d, cachable, have [%d:%d], hot\n",
                            this->sid, this->fhash, this->shash);
#endif

                        /* not hit and not miss */

                        this->state = STATE_WAIT_FRIEND;
#if 0
                        this->reason2 = stored_pos->owner;
#endif
#if 0
                        if(this->mode==MODE_CLIENT) {
                        Trace("sess# %d, to waiting owner %p, client\n",
                            this->sid, stored_pos->owner);
                        }
                        else {
                        Trace("sess# %d, to waiting owner %d, self\n",
                            this->sid, stored_pos->owner);
                        }
#endif
                    }
                }
            }
  }
  else {

#if 0
            Trace("delete old entry 2\n");
            skcache_entry_delete(bk, stored_pos);

            MUTEX_LOCK(&stat_lock);
            miss_reply_count++;
            MUTEX_UNLOCK(&stat_lock);

            this->state = STATE_QUERY_ADDR;
#endif

			Trace("sess# %d, uncachable, make new entry\n", this->sid);


	        if(check_already_have==1) {
				Trace("th: uncahable, have entry %p, oid# %d, ref %d\n",
					stored_pos, stored_pos->oid, stored_pos->ref);

#if 0
				Trace("delete old entry 2\n");
				skcache_entry_delete(bk, stored_pos);
#endif
				if(stored_pos->owner==NULL) {
					stored_pos->cracked = 1;
					MUTEX_LOCK(&stat_lock);
					miss_reply_count++;
					MUTEX_UNLOCK(&stat_lock);

					this->state = STATE_QUERY_ADDR;
				}
				else {
					this->state = STATE_WAIT_FRIEND;
#if 0
					this->reason2 = stored_pos->owner;
#endif
				}
			}
			else {
				Trace("th: uncahable, not have entry %p, oid# %d, ref %d\n",
					stored_pos, stored_pos->oid, stored_pos->ref);

				this->oid = this->sid;

#if 0
                stored_pos->owner       = this->sid;
#endif
                stored_pos->owner       = this;
                stored_pos->oid         = this->sid;
                stored_pos->issue_time  = this->last_action;

				stored_pos->cracked = 1;
				MUTEX_LOCK(&stat_lock);
				miss_reply_count++;
				MUTEX_UNLOCK(&stat_lock);

				this->state = STATE_QUERY_ADDR;
			}


 assert(check_bkentry(bk, "th8.8"));
 assert(check_bkinfo());
 assert(bk->cup.num==bk->entry_use);

  }

		if(this->state == STATE_THINK) {
			Trace("sess# %d, condition leak\n", this->sid);
		}

get_last:

        if(bk) {
 assert(check_bkentry(bk, "th9"));
 assert(check_bkinfo());
			UNLOCK_BK(bk);

            bk = NULL;
        }

#if 0
Trace("sess# %d, end of GET, state %d %s\n", 
	this->sid, this->state, MSG_STATE_[this->state]);
#endif

        break;


    /* POST method */
    case M_POST:
Trace("POST method <%s>\n", this->loc);
        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_QUERY_ADDR;

        break;

    default:
        Error("sess# %d, ignore method %d '%s'",
            this->sid, this->state, MSG_STATE_[this->state]);

        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->reason = R_UNSUPPORT_METHOD;
        this->state = STATE_MESSAGE;

        break;
    }

#if 0
    MUTEX_LOCK(&stat_lock);
    process_request_count++;
    MUTEX_UNLOCK(&stat_lock);
#endif

    return 0;
}


/*
#ifndef TRACE_SESS
#define TRACE_SESS
#endif
*/


int
DoCheckEndFriendSess(sess_t *this)
{
    struct timeval xbefore, xafter;
    int chk;
    bucket_t *bk;
    skcache_entry_t *stored_pos;
    int check_already_have;

    skcache_entry_t tmp;
	int idle;

#ifdef TRACE_SESS
    Trace("DoCheckEndFriendSess: sess# %d, entry 0x%x, '%s' [%d:%d:%d:%d] %s\n",
        this->sid,
		this->cache_entry,
		this->loc,
        this->fhash, this->shash, this->thash, this->ffhash,
        MSG_STATE_[this->state]);
#endif

    bk = &skcache_bucket_base[this->fhash];

	LOCK_BK(bk);

 assert(check_bkentry(bk, "cf1"));

    stored_pos = this->cache_entry;

    /* error */
    if(stored_pos==NULL) {
        Error("sess# %d, not found such entry", this->sid);

		goto do_retrieve;
    }

#ifdef TRACE_SESS
        Trace("stored %p ref %d hash %4d %4d %4d %4d %s\n",
            stored_pos, stored_pos->ref,
            stored_pos->h1, stored_pos->h2,
            stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

    if(stored_pos->h4!=this->ffhash||stored_pos->h3!=this->thash
        ||stored_pos->h2!=this->shash||stored_pos->h1!=this->fhash||
            strcmp(stored_pos->name, this->loc)!=0) {

        Error("checkfriend: sess# %d, entry was changed. may be expired.",
			this->sid);

#ifdef TRACE_SESS
        Trace("stored %p %4d %4d %4d %4d %s\n",
            stored_pos,
            stored_pos->h1, stored_pos->h2,
            stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

		goto do_retrieve;
    }

#if 0
 assert(CHKTIME(stored_pos->hot));
    stored_pos->hot = time(NULL);
#endif

    if(stored_pos->owner==this) {
#ifdef TRACE_SESS
        Trace("sess# %d, the line is mine. *strange*\n", this->sid);
#endif
		goto do_retrieve;
    }
    else {
        if(stored_pos->cracked) {
#ifdef TRACE_SESS
            Trace("sess# %d, line was cracked\n", this->sid);
#endif
			goto do_retrieve;
        }
        else
#if 0
        if(stored_pos->ref==1 && stored_pos->owner==NULL)
#endif
        if(stored_pos->owner==NULL)
		{
#ifdef TRACE_SESS
            Trace("sess# %d, line is free\n", this->sid);
#endif
            stored_pos->owner = this;

			goto do_send;
        }
        else {
			idle = this->last_action - stored_pos->hot;
			if(idle < 0 && idle > 3*WFRIEND_TIMEOUT/4) {
#ifdef TRACE_SESS
				Trace("sess# %d, give up wait (%d sec)\n",
					this->sid, idle);
#endif
				goto do_retrieve;
			}
			else {
#ifdef TRACE_SESS
				Trace("sess# %d, oid# %d is still hold by sess %p\n",
					this->sid, stored_pos->oid, stored_pos->owner);
#endif
				/* not hit and not miss */
				goto still_wait;
			}
        }
    }

#ifdef TRACE_SESS
	Trace("reach to end\n");
#endif
    goto skip;

do_retrieve:
#ifdef TRACE_SESS
	Trace("do retrieve\n");
#endif
	if(stored_pos) {
#if 1
        stored_pos = NULL;
#else
        this->cache_entry = NULL;
#endif
	}

	this->cachable = SESS_UNCACHABLE;

    MUTEX_LOCK(&stat_lock);
    miss_reply_count++;
    MUTEX_UNLOCK(&stat_lock);

    this->state = STATE_QUERY_ADDR;
    goto skip;

do_send:
#ifdef TRACE_SESS
	Trace("do send\n");
#endif
    MUTEX_LOCK(&stat_lock);
    hit_reply_count++;
    MUTEX_UNLOCK(&stat_lock);

    this->state = STATE_SEND_CACHE;
    goto skip;

still_wait:
	/* not hit and not miss */
#ifdef TRACE_SESS
	Trace("still wait\n");
#endif

    this->state = STATE_WAIT_FRIEND;
    goto skip;


skip:

#ifdef TRACE_SESS
    if(this->state != STATE_WAIT_FRIEND) {
        Trace("DoCheckEndFriendSess: sess# %d, change to %s\n",
			this->sid, MSG_STATE_[this->state]);
    }
#endif
#if 1
        Trace("DoCheckEndFriendSess: sess# %d, change to %s\n",
			this->sid, MSG_STATE_[this->state]);
#endif

    if(bk) {
 assert(check_bkentry(bk, "cf9"));

        UNLOCK_BK(bk);
        bk = NULL;
    }

out:

    return 0;
}


/*
#ifdef TRACE_SESS
#undef TRACE_SESS
#endif
*/


int
DoFindServer(sess_t *this)
{
    int chk;
    int ret;


#if 0
    Trace("DoFindServer: URL '%s'\n\t host '%s', port %d, path '%s'\n",
        this->loc, this->servername, this->serverport, this->serverpath);
#endif
    if(this->servername==NULL || this->servername[0]=='\0') {
        Error("sess# %d, empty hostname is specified", this->sid);
        return -1;
    }


#if 0
    hcache_list();
#endif

    chk = hcache_seek_addr_incache(this->servername, &(this->serveraddr));
    if(chk) {
        /* miss */
#if 0
        Trace("miss in hostname cache, '%s' (%d)\n",
            this->servername, chk);
        Trace("send query to resolver\n");
#endif
        chk = hcache_send_query_addr_direct(this->servername,
                &(this->serveraddr), &this->rfd);
#if 0
        Trace("    ret %d\n", chk);
#endif
        if(chk) {
            Error("fail hcache_send_query_addr_direct() (%d)\n", chk);
            this->reason = R_RESOLVER_DOWN;
            this->state = STATE_MESSAGE;
            goto out;
        }

        this->state = STATE_WAIT_ADDR;
        ret = 0;
    }
    else {
        /* hit */
#if 0
        Trace("hit in hostname cache, '%s'\n",
            this->servername);
#endif
        this->serveraddr.sin_family = AF_INET;

        this->state = STATE_CONN_SERV;
        ret = 0;
    }

out:
    return ret;
}


int
DoServerConnect(sess_t *this)
{
    int chk;

#if 0
    Trace("DoServerConnect: server '%s' [%s], port %d",
        this->servername, inet_ntoa(this->serveraddr.sin_addr),
        this->serverport);
#endif

#if 0
    this->last_action = time(NULL);
#endif

    /* asyncronus connect */
    this->sfd = TCP_aconnect_addr(&this->serveraddr, this->serverport);
    if(this->sfd<0) {
        if(this->sfd == KOTETU_SOCK_SOCKETERROR) {
            this->reason = R_FAIL_SERVER_SOCKET;
#if 0
            exit(437);
#endif
        }
        else {
            Error("DoServerConnect: sess# %d, reason %d",
                this->sid, this->sfd);
            this->reason = R_FAIL_SERVER_CONNECT;
        }
        this->state = STATE_MESSAGE;
        return 0;
    }

    this->state = STATE_WAIT_CONN_SERV;

    return 0;
}

int
DoWaitServerConnect(sess_t *this)
{
    volatile int errno_backup;
    int err, len;
    int chk;


#if 0
    Trace("DoWaitServerConnect: server '%s' [%s], port %d\n",
        this->servername, inet_ntoa(this->serveraddr.sin_addr),
        this->serverport);
#endif

    chk = iswritable_via_poll(this->sfd, 1, 0);
    if(chk>0) {
        this->state = STATE_SEND_REQUEST;
    }

    return 0;
}

int
DoSendRequest(sess_t *this)
{
    char req[L_BUF];
    int len, rqlen;
    register char *p, *q, *r;
    int  nw;
    char line[STRING_SIZE];
    char *cpos;
	int c;


    /*
     * several server requires HOST field, independly protocol versions.
     * then, the program put HOST field.
     */
    if(this->method==M_POST) {
        sprintf(req, "POST %s HTTP/1.0\r\nHost: %s\r\n",
            this->serverpath, this->servername);
#if 0
    dump_stringchunk("orignal", this->reqbuf);
#endif
#if 0
    dump_stringsegment("orignal", this->reqbuf, this->reqlen);
    dump_stringchunk("first", req);
#endif
    }
    else {
        sprintf(req, "GET %s HTTP/1.0\r\nHost: %s:%d\r\n",
            this->serverpath, this->servername, this->serverport);
    }

#if 0
    dump_stringsegment("r.req.", this->reqbuf, this->reqlen);
#endif

    rqlen = HTTP_find_header_term(this->reqbuf, this->reqlen);
    if(rqlen>=L_REQBUF-1) {
        Error("too long requst (meybe rqlen is %d)", rqlen);
        dump_stringsegment("the request", this->reqbuf, this->reqlen);

        this->state = R_TOOLONG_REQ;
        this->state = STATE_MESSAGE;
        return -1;
    }

    p = this->reqbuf;
    r = this->reqbuf + rqlen;

    while(*p && *p!='\n') {
        p++;
    }
    p++;

    /* skip first line, because it was already copied */
    q = req;
    while(*q) {
        q++;
    }

    /* copy rest of header */
    cpos = p;
    while((cpos = sgets(p, cpos, line, STRING_SIZE))!=NULL) {
        if(line[0]=='\0') {
            break;
        }
        if((line[0]=='H' || line[0]=='h')
            && strncasecmp(line, "Host:", 5)==0) {
            continue;
        }
        if((line[0]=='P' || line[0]=='p')
            && strncasecmp(line, "Proxy-Connection:", 17)==0) {
            continue;
        }
        strcat(req, line);

        if(line[0]=='\r'||line[0]=='\n') {
            break;
        }
    }


    len = strlen(req);

#if 0
    dump_stringchunk("cooked request", req);
#endif

#if 1
    if(len>=L_REQBUF-1) {
        Error("too long requst (current %d, original %d)", len, this->reqlen);
        dump_stringchunk("the requst", req);

        this->state = R_TOOLONG_REQ;
        this->state = STATE_MESSAGE;
        return -1;
    }
#endif

    if(this->method==M_POST) {
        int c;

        c = this->reqlen - (cpos-&this->reqbuf[0]);

#if 0
        dump_stringchunk("cooked", req);
#endif
        if(c) {
#if 0
            dump_stringsegment("rest  ", cpos, c);
#endif
            p = cpos;
            q = req + len;
            while(c>0) {
                *q++ = *p++;
                c--;
                len++;
            }
#if 0
        dump_stringsegment("cooked", req, len);
#endif
        }
    }

#if 0
  if(this->mode == MODE_SELF ) {
    Trace("sess# %d, request final form, length %d\n", this->sid, len);
    dump_stringsegment("s.req.", req, len);
  }
#endif
#if 0
    Trace("sess# %d, request final form, mode %d, length %d\n",
        this->sid, this->mode, len);
    dump_stringsegment("s.req.", req, len);
#endif

	c = 0;
try_write:
	c++;
    nw = write(this->sfd, req, len);
    if(nw<0) {
        if(errno==EINTR) {
            goto try_write;
        }

        if(errno==EPIPE) {
            this->reason = R_LOST_SERVER_CONNECT;
#if 0
            Error("sess# %d, server pipe (fd# %d) error (%d)",
                this->sid, this->sfd, errno);
#endif
        }
#if 0
		else
		if(errno==EAGAIN) {
		}
#endif
        else {
            this->reason = R_FAIL_SERVER_WRITE;

            Error("cannot write request to server sess %p fd #%d (%d)",
                this, this->sfd, errno);
        }

        this->state = STATE_MESSAGE;

        return -1;
    }
    if(nw==0) {
        this->state = STATE_CLOSE;
        return -1;
    }
#if 0
    Trace("write request %d byte(s)\n", nw);
#endif

    this->swrite += (off_t) nw;


#if 0
    this->last_action = time(NULL);
#endif


    if(this->method == M_GET) {
        this->state = STATE_RECV_RES;
    }
    else
    if(this->method == M_POST) {
        this->state = STATE_RELAY_DIRECT;
    }
    else {
    }

    return 0;
}



/*
 * for POST method
 */
int
DoRelayDirect(sess_t *this)
{
    int rest;
    int use_tmpbuf;
    char tmpbuf[L_BUF];
    int l;
    struct iovec iovec[2];
    int ciovec;
    int blen;
    int nr;
    int nw;
    int chk;
    int err;

    int dir;
    int cr, cw, sr, sw;

#if 0
    Trace("DoRelayDirect: self %#x, sess# %d, hash %d:%d\n",
        pthread_self(), this->sid, this->fhash, this->shash);
#endif



    dir = 0;
    cr = cw = sr = sw = (off_t)0L;

    /*
     * Since connection starts by client, client has high priority.
     */

client_side:
    if(this->cfd<0) {
        goto server_side;
    }
    chk = isreadable_via_poll(this->cfd, 0, 0);
    if(chk) {
        dir = 1; /* client */
    }
    else {
        goto server_side;
    }

try_c_read:
    nr = read(this->cfd, tmpbuf, L_BUF);
    if(nr<0) {
        if(errno==EINTR)
            goto try_c_read;

        this->reason = R_FAIL_CLIENT_READ;
        this->state = STATE_MESSAGE;
        goto out;
    }
    else
    if(nr==0) {
        DoHalfClientCloseSess(this);
        goto server_side;
    }


    cr += nr;

try_s_write:
    nw = write(this->sfd, tmpbuf, nr);
    if(nw<0) {
        if(errno==EINTR)
            goto try_s_write;

        this->reason = R_FAIL_SERVER_WRITE;
        this->state = STATE_MESSAGE;
        goto out;
    }

    sw += nw;


server_side:
    if(this->sfd<0) {
        goto disconn_check;
    }
    chk = isreadable_via_poll(this->sfd, 0, 0);
    if(chk) {
        dir += 2; /* server */
    }
    else {
        goto disconn_check;
    }

try_s_read:
    nr = read(this->sfd, tmpbuf, L_BUF);
    if(nr<0) {
        if(errno==EINTR)
            goto try_s_read;

        this->reason = R_FAIL_SERVER_READ;
        this->state = STATE_MESSAGE;
        goto out;
    }
    else
    if(nr==0) {
        DoHalfServerCloseSess(this);
        goto disconn_check;
    }

    sr += nr;

try_c_write:
    nw = write(this->cfd, tmpbuf, nr);
    if(nw<0) {
        if(errno==EINTR)
            goto try_c_write;

        this->reason = R_FAIL_CLIENT_WRITE;
        this->state = STATE_MESSAGE;
        goto out;
    }

    cw += nw;


disconn_check:

    if(this->cfd<0 || this->sfd<0) {
        this->state = STATE_CLOSE;
    }

out:
    if(cr>0) this->cread  += cr;
    if(cw>0) this->cwrite += cw;
    if(sr>0) this->sread  += sr;
    if(sw>0) this->swrite += sw;


#if 0
    Trace("sess# %d, dir %x, cfd,sfd=%d,%d r/w=c %d/%d, s %d/%d %s\n",
        this->sid, dir, this->cfd, this->sfd,
        cr, cw, sr, sw, MSG_STATE_[this->state]);
#endif

    return 0;
}



/*
 *
 */

int
DoServerReply(sess_t *this)
{
    int rest;
    int use_tmpbuf;
    char tmpbuf[L_BUF];
    int l;
    struct iovec iovec[2];
    int ciovec;
    int blen;
    int nr;
    int nw;
    int chk;
    int err;
    struct timeval xbefore, xafter;

    int hlen;
    int xvmajor, xvminor;
    time_t xdate, xlmd, xexpire;
    off_t xclen;
    int xrcode;
    int xcachable;
    int xreload;

    bucket_t *bk;
    skcache_entry_t *centry;
    int check_already_have;
    skcache_entry_t tmp;

    skcache_frag_t *flist;

    skcache_frag_t *last;
    skcache_frag_t *cur;
    off_t frag_sum;
    int u;

	int c;


/*
#ifdef TRACE_CACHE
    Trace("DoServerReply: self %#x, sess# %d, hash %d:%d\n",
        pthread_self(), this->sid, this->fhash, this->shash);
#endif
*/

    /***
     *** alloc buffer
     ***/
    bk = NULL;
    use_tmpbuf = 0;
    last = NULL;
    cur = NULL;
    blen = 0;

    if(this->cachable==SESS_UNCACHABLE) {
#if 0
        Trace("sess# %d, uncachable\n", this->sid);
#endif
        goto alloc_tmpbuf;
    }

    bk = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);

 assert(check_bkentry(bk, "sr0"));

    centry = (skcache_entry_t*) this->cache_entry;
    if(centry==NULL) {
#if 0
        Trace("sess# %d, centry is NULL\n", this->sid);
#endif
        goto alloc_tmpbuf;
    }
    if(centry->cracked) {
#if 0
        Trace("sess# %d, centry is cracked\n", this->sid);
#endif
        goto alloc_tmpbuf;
    }

#if 0
	Trace("sess# %d, SR magic %02x %02x\n",
		this->sid, centry->magic[0], centry->magic[1]);
#endif


    /* seek last position of object frags in this object entry */
    /* retrieving object must be upon memory */
#if 0
{
    last = NULL;
    frag_sum = 0L;
    u = 0;
    cur = centry->m_recorded;
    while(cur) {
        frag_sum += cur->len;
        last = cur;
		Trace(" %4d: %p last %p, len %5ld, sum %9ld\n",
			u, cur, last, (long)cur->len, (long)frag_sum);

        if(cur->next==cur) {
            Error("sess# %d, %dth frag %p is looped itself, %ld byte(s)",
                this->sid, u, cur, frag_sum);
			break;
        }

        cur = cur->next;

        u++;
        if(u>=max_frag_per_obj) {
            Error("sess# %d, too large, %d frag(s) max %d %ldbyte(s)",
                this->sid, u, max_frag_per_obj, frag_sum);
			break;
        }
    }
	Trace("sr: sess# %d, u %d, cur %p, last %p\n",
		this->sid, u, cur, last);
}
#endif
    last = NULL;
    frag_sum = 0L;
    u = 0;
    cur = centry->m_recorded;
    while(cur) {
        frag_sum += cur->len;
        last = cur;

        if(cur->next==cur) {
            Error("sess# %d, %dth frag %p is looped itself, %ld byte(s)",
                this->sid, u, cur, frag_sum);
            centry->cracked = 1;
            goto alloc_tmpbuf;
        }

        cur = cur->next;

        u++;
        if(u>=max_frag_per_obj) {
#if 0
            Error("sess# %d, too large, %d frag(s) max %d %ldbyte(s)",
                this->sid, u, max_frag_per_obj, frag_sum);
#endif
            centry->cracked = 1;
            goto alloc_tmpbuf;
        }
    }

#if 0
	Trace("sr: sess# %d, u %d, cur %p, last %p\n",
		this->sid, u, cur, last);
#endif

 assert(check_bkentry(bk, "sr3"));

    if(bk->elem_freelist==NULL
        || bk->entry_freelist==NULL || bk->entry_use > bk->entry_hwater
        || bk->frag_freelist==NULL || bk->frag_use > bk->frag_hwater)
    {
#if 1
		Trace("servereply: memory check bk# %d,\n", bk->bid);
#endif
		chk = call_msweeper(this, bk, XSK_FRAG);

#if 0
		if(chk==-2) {
			Trace("think: sess# %d, fail to push sweeper Q, bk# %d,\n",
				this->sid, bk->bid);

			UNLOCK_BK(bk);

			goto get_first;
		}
#endif

		if(chk) {
			Trace("sendcache: sess# %d, no space bk# %d,\n",
				this->sid, bk->bid);

#if 0
            this->state  = STATE_MESSAGE;
            this->reason = R_MEMORY_OVERFLOW;
#endif

			centry->cracked = 1;
			goto alloc_tmpbuf;
		}
	}

    /* get new frag */
try_get_frag:
 assert(bk->cup.num==bk->entry_use);
    cur = skcache_pull_new_frag(bk);
 assert(bk->cup.num==bk->entry_use);

 assert(check_bkentry(bk, "sr4"));

    if(cur==NULL) {
        Error("ServerReply: sess# %d, cannot get new frag on bk# %d %p",
            this->sid, bk->bid, bk);
	
		centry->cracked = 1;
        goto alloc_tmpbuf;
    }


#if 0
Trace("  new frag from free-list %p <next %p>\n", cur, cur->next);
#endif

    if(last) {
        if(last==cur) {
            Error("PANIC !!! bk %p, last %p, cur %p, frag %d/%d %p",
                bk, last, cur,
                bk->frag_use, bk->frag_all, bk->frag_freelist);

            skcache_dump_entry_frag("strange", centry);

            centry->cracked = 1;
            goto alloc_tmpbuf;
        }
        last->next = cur;
    }
    else {
        centry->m_recorded = cur;
    }

#if 0
    skcache_dump_entry_frag("alloc", centry);
#endif

    blen = 1;
    iovec[0].iov_base = cur->ent;
    iovec[0].iov_len  = cur->size;

    MUTEX_LOCK(&stat_lock);
    alloc_sess_buf_count++;
    MUTEX_UNLOCK(&stat_lock);

    goto alloc_done;



alloc_tmpbuf:
	Trace("sess# %d, alloc tmpbuf\n", this->sid);
#ifndef NDEBUG 
  	if(bk) {
 assert(check_bkentry(bk, "sr7"));
	}
#endif

    if(centry) {
#if 0
        Trace("sess# %d, delete old entry 3\n", this->sid);
        skcache_entry_delete(bk, centry);
#endif
        centry = NULL;
#if 0
        this->cache_entry = NULL;
#endif
    }

    blen = 1;
    iovec[0].iov_base = tmpbuf;
    iovec[0].iov_len  = sizeof(tmpbuf);

    MUTEX_LOCK(&stat_lock);
    alloc_tmp_buf_count++;
    MUTEX_UNLOCK(&stat_lock);

    goto alloc_done;


alloc_done:
#ifndef NDEBUG 
 	if(bk) {
 assert(check_bkentry(bk, "sr8"));
	}
#endif

    if(blen==1) {
        rest = iovec[0].iov_len;
    }
    else
    if(blen==2) {
        rest = iovec[0].iov_len + iovec[1].iov_len;
    }
    else {
        rest = -1;
    }

#if 0
    Trace("read:\n");
    Trace("  blen %d\n", blen);
    Trace("  iovec[0] base %p, len %d\n",
        iovec[0].iov_base, iovec[0].iov_len);
    if(blen>=2) {
        Trace("  iovec[1] base %p, len %d\n",
        iovec[1].iov_base, iovec[1].iov_len);
    }
    Trace("  rest %d\n", rest);
#endif

    /***
     *** read socket
     ***/
    MUTEX_LOCK(&stat_lock);
    read_reply_total++;
    MUTEX_UNLOCK(&stat_lock);

try_read:
    errno = 0;
    nr = readv(this->sfd, iovec, blen);
    err = errno;
#ifdef TRACE_CACHE
    if(use_tmpbuf) {
        Trace("  sess# %d, fd# %d, read %d/%d (temporaly area)\n",
            this->sid, this->sfd, nr, rest);
    }
    else {
        Trace("  sess# %d, fd# %d, read %d/%d (cache area)\n",
            this->sid, this->sfd, nr, rest);
    }
#endif
    if(nr<0) {
        if(errno==EINTR) {
            goto try_read;
        }
        /* for OSF/1 */
        if(errno==0) {
            goto try_read;
        }

        MUTEX_LOCK(&stat_lock);
        read_reply_fail++;
        MUTEX_UNLOCK(&stat_lock);

        if(errno==EAGAIN) {
            goto out;
        }
        if(errno==ECONNRESET) {
            goto lost_connect;
        }
        Error("cannot read server reply nr %d (%d)", nr, err);

        goto fail_read;
    }

    MUTEX_LOCK(&stat_lock);
    read_reply_success++;
    MUTEX_UNLOCK(&stat_lock);

    if(nr==0) {
#if 0
        skcache_dump_entry_frag("before", centry);
#endif

        /* restore
         *  But if totallength is 0 (means empty),
         *  We have to keep this one as mark of empty.
         */
        if(cur && centry && centry->totallength > 0) {
            skcache_push_idle_frag(bk, cur);
            cur = NULL;

            blen = 0;

            if(last) {
                last->next = NULL;
            }
            else {
#if 0
Trace("sess# %d, first read is empty\n", this->sid);
#endif
                centry->m_recorded = NULL;
                centry->d_recorded = -1;
            }

        }

#if 0
        skcache_dump_entry_frag("after", centry);
        Trace("===\n");
#endif

        this->reason = -1;
        goto reach_end;
    }

#if 0
dump_stringsegment("rep", iovec[0].iov_base, MIN(nr, 128));
#endif

    this->sread += (off_t) nr;

    sim_stat_add_value(&len_perread_stat, (long)nr);

#if 0
    if(this->pos <= (off_t)0 && this->sver_major < 0)
#endif
    if(this->pos == (off_t)0 )
    {

#if 0
dump_stringsegment("first", iovec[0].iov_base, MIN(nr, 300));
#endif

        hlen = HTTP_find_header_term(iovec[0].iov_base,
                MIN(nr, iovec[0].iov_len));
#if 0
Trace("sess# %d, r header term %4d / %5d\n",
        this->sid, hlen, MIN(nr, iovec[0].iov_len));
#endif
        if(hlen>0) {
#if 1
#if 0
            if(hlen>320) {
Trace("sess# %d, r * header term %4d / %5d\n",
        this->sid, hlen, MIN(nr, iovec[0].iov_len));
/*
                dump_stringsegment("res.", iovec[0].iov_base, hlen);
*/
            }
            else
#endif
            if(hlen>=iovec[0].iov_len) {
Trace("sess# %d, r $ header term %4d / %5d\n",
        this->sid, hlen, MIN(nr, iovec[0].iov_len));
/*
                dump_stringsegment("broken",
                    iovec[0].iov_base, iovec[0].iov_len);
*/
            }
#endif

/*
            this->pos   = (off_t) hlen;
*/
            this->hlen  = hlen;
            xcachable   = this->cachable;
            xreload     = this->reload;

            chk = HTTP_parse_response_header(iovec[0].iov_base, hlen,
                    &xvmajor, &xvminor, &xrcode,
                    &xdate, &xlmd, &xexpire, &xclen,
                    &xcachable, &xreload);

            if(!chk) {
#if 0
Trace("%c(@", 0x1b);
Trace("%c$B%c(B", 0x1b, 0x1b);
#endif
#if 0
Trace("sess# %d, <hlen %d><chk %d> ver %d.%d, rcode %d, lmd %d, expire %d, clen %d\n",
    this->sid, hlen, chk,
    xvmajor, xvminor, xrcode, xlmd, xexpire, xclen);
#endif
                this->sver_major    = xvmajor;
                this->sver_minor    = xvminor;
                this->rcode         = xrcode;
                this->lmd           = xlmd;
                this->expire        = xexpire;
                this->clen          = xclen;

#if 1
				if(centry) {
					centry->lmd 	= xlmd;
					centry->expire 	= xexpire;
				}
#endif


                if(xcachable==SESS_UNCACHABLE) {
#if 0
    Trace(" 74= sess# %d, %s\n", this->sid, this->loc);
    dump_stringsegment("req.", this->reqbuf, this->reqlen);
#endif
                    this->publish = SESS_PRIVATE;


                    if(centry) {
                        centry->cracked = 1;
                    }

                }
                else {
                    this->publish = SESS_PUBLIC;
                }

            }
            else {
                this->sver_major    = 0;
                this->sver_minor    = 0;
                this->rcode         = -1;
                this->lmd           = -1;
                this->expire        = -1;
                this->clen          = -1;
                this->publish       = SESS_PRIVATE;
            }

            this->pos += (off_t) this->sread;
        }
        else {
#if 0
Trace("sess# %d, hlen %d, nr %d, buf-len %d\n",
        this->sid, hlen, nr, iovec[0].iov_len);

dump_stringsegment("req.", this->reqbuf, this->reqlen);
#endif
        }

    }
    else {
        this->pos += (off_t) nr;
    }

#if 0
    if(this->rcode<0) {
        Trace("sess# %d, rcode %d\n", this->sid, this->rcode);
    }
#endif


    if(this->len<0) {
        this->len = (off_t) nr;
    }
    else {
        this->len += (off_t) nr;
    }

    if(centry) {
        if(centry->totallength<=0)
            centry->totallength = nr;
        else
            centry->totallength += nr;

        if(blen==1) {
            if(cur) {
                cur->len += nr;
            }
            else {
                Error("no cur frag to record length");
            }
        }
        else {
            Error("ignore blen %d", blen);
        }
#if 0
        else
        if(blen==2) {
            if(nr <= iovec[0].iov_len) {
                tlpos->cntl.length += nr;
            }
            else {
                tlpos->cntl.length += iovec[0].iov_len;
                nlpos->cntl.length = (nr - iovec[0].iov_len);
            }
        }
#endif
    }

#if 0
    else {
        Trace("does not have centry to save, sess# %d\n", this->sid);
    }
#endif

#if 0
    this->last_action = time(NULL);
#endif
    if(centry) {
        centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));
    }

    if(this->cfd>=0) {
        if(blen==2) {
            if(nr<=iovec[0].iov_len) {
                blen = 1;
                iovec[0].iov_len = nr;
            }
            else {
                if(nr<=iovec[0].iov_len + iovec[1].iov_len) {
                    iovec[1].iov_len = nr - iovec[0].iov_len;
                }
                else {
        Error("length of read <%d> is large than buffer <%d(%d+%d)>\n",
            nr, iovec[0].iov_len + iovec[1].iov_len);
                }
            }
        }
        else
        if(blen==1) {
            if(nr<iovec[0].iov_len) {
                iovec[0].iov_len = nr;
            }
        }
#if 0
Trace("write:\n");
Trace("  blen %d\n", blen);
Trace("  iovec[0] base %p, len %d\n", iovec[0].iov_base, iovec[0].iov_len);
if(blen>=2)
Trace("  iovec[1] base %p, len %d\n", iovec[1].iov_base, iovec[1].iov_len);
#endif

try_write:
        nw = writev(this->cfd, iovec, blen);

        if(nw<0) {
            if(errno==EINTR) {
                goto try_write;
            }
            if(errno==EPIPE) {
                this->reason = -2;
                DoHalfClientCloseSess(this);
                goto reach_end;
            }

            Error("sr: sess# %d, cannot write to client fd# %d (%d)",
                this->sid, this->cfd, errno);
#if 0
            this->state = STATE_CLOSE;
#endif
            this->reason = R_FAIL_CLIENT_WRITE;
            DoHalfClientCloseSess(this);

            goto fail_write;
        }
        if(nw<nr) {
            Error("sr: wrote size (%d) is smaller than read size (%d)", nw, nr);
            goto fail_write;
        }
        if(nw>nr) {
            Error("sr: wrote size (%d) is larger than read size (%d)", nw, nr);
        }
		if(nw==0) {
            Error("sr: wrote size (%d) is zero blen %d", nw, blen);
		}

/*
        this->pos += (off_t) nw;
*/
        this->cwrite += (off_t) nw;

#if 0
Trace("sess# %d, nw %ld, cwrite %ld\n", (long)nw, (long)this->cwrite);
#endif

    }


#if 0
    if(this->pos != this->len) {
    Trace("sess# %d, pos %ld, len %ld\n",
        this->sid, (long)this->pos, (long)this->len);
    }
#endif
#ifndef NDEBUG
	if(bk) {
 assert(check_bkentry(bk, "sr9"));
	}
#endif

    goto out;

/*
 * error recovery
 */

fail_memory:
    Error("sr: fail memory arrengment or memory trouble, sess# %d",
        this->sid);
    this->reason = R_LOST_BLOCK;

    goto fail;

lost_connect:
    Error("sr: lost conection establishment to server with some reason, sess# %d",
        this->sid);
    this->reason = R_LOST_SERVER_CONNECT;

    goto fail;

fail_connect:
    Error("sr: fail conection establishment to server with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_SERVER_CONNECT;

    goto fail;

fail_read:
    Error("sr: fail reading from server with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_SERVER_READ;

    goto fail;

fail_write:
    Error("sr: fail writing to client with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_CLIENT_WRITE;

    goto fail;

fail:
    this->state = STATE_MESSAGE;

    goto out;




reach_end:
    this->state = STATE_CLOSE;
#ifdef TRACE_CACHE
    Trace("Ok,sess# %d [%d:%d]                 read %12ld / %4ldK / %4ldM\n",
        this->sid, this->fhash, this->shash,
        (long)this->len,
        (long)(this->len/1024), (long)(this->len/(1024*1024)) );
#endif

    sim_stat_add_value(&object_size_stat, (long)(this->len/1024) );

    if(this->len==0) {
        Log("; sess# %d, empty ?", this->sid);
    }

#if 0
	Trace("sr: sess# %d, %d %d diff %d, cl %ld, hl %ld, sr %ld, cw %ld, pos %ld\n",
		this->sid, 
		(this->sread==this->cwrite && this->cwrite==this->pos),
		(this->clen+this->hlen + 1 == this->sread),
		(long)(this->clen + this->hlen + 1 - this->sread),
		(long)this->clen,
		(long)this->hlen,
		(long)this->sread, (long)this->cwrite, (long)this->pos);
#endif

    if(centry) {
        /* discard cached object */
        if(this->rcode == 304) {
            centry->cracked = 1;
        }

#if 0
        Trace("sess# %d, length server said/recoreded %12ld/%12ld\n",
            this->sid, (long)this->clen, (long)centry->totallength);
#endif

        centry->owner = NULL;
    }

    goto out;

out:
    if(bk) {
 assert(check_bkentry(bk, "sra"));

        UNLOCK_BK(bk);
        bk = NULL;
    }
	else {
Trace("bk is empty\n");
	}

    if(centry) {
#if 0
        centry->hot = xafter.tv_sec;
#endif
        centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));
    }
	else {
Trace("centry is empty\n");
	}


#ifdef TRACE_CACHE
Trace(" @\n\n");
#endif

    return 0;
}





int
DoSendCache(sess_t *this)
{
    int chk;
    int sum;
    int nw;
    int l;
    struct iovec iov[MAX_FRAG_PER_OBJ];
    int ciov;
    time_t now;
    int hlen;
    int xvmajor, xvminor, xrcode;
    time_t xdate, xlmd, xexpire;
    off_t xclen;
    int xcachable, xreload;
    bucket_t *bk;
    skcache_entry_t *centry;
    skcache_frag_t *cur;
    skcache_frag_t *first;


#if 0
    struct timeval reftime;
#endif

#if 0
    Trace("DoSendCache: [%d:%d] '%s'\n",
    this->fhash, this->shash, this->loc);
#endif


    bk  = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);
 assert(check_bkentry(bk, "sc1"));

    now = time(NULL);

    centry = this->cache_entry;
    if(centry==NULL) {
#if 0
        Error("DoSendCache: hit sess# %d [%d:%d:%d:%d]. however, not found entry <%s,%s>",
            this->sid, this->fhash, this->shash, this->thash, this->ffhash,
            this->cachable==SESS_CACHABLE ? "cachable" : "NOTCACHABLE",
            this->reload==SESS_RELOAD ? "RELOAD" : "noreload");
#endif
#if 0
        Trace("  target='%s'\n", this->loc);
        cache_showblockaddr(this->fhash, 0);
#endif

#if 0
        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        hit_reply_count--;
        skip_cache_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_QUERY_ADDR;

        goto out;
#endif

		goto tobe_miss;	
    }
#if 0
	Trace("stored %p oid# %d, ref %d, [%d:%d:%d:%d] %s\n",
		centry, centry->oid, centry->ref,
		centry->h1, centry->h2, centry->h3, centry->h4, centry->name);
#endif

	if(centry->ref<0) {
        Error("sendcache: hit sess# %d [%d:%d:%d:%d], it was dead. may be expired.",
            this->sid, this->fhash, this->shash, this->thash, this->ffhash);

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
	}

#if 0
                Trace("comp centry %x vs this\n\
%d/%d %d/%d %d/%d %d/%d\n  %s\n  %s\n",
	centry,
    centry->h1, this->fhash, 
	centry->h2, this->shash,
	centry->h3, this->thash,
	centry->h4, this->ffhash, 
    centry->name, this->loc);
#endif


    if(centry->h4!=this->ffhash||centry->h3!=this->thash
        ||centry->h2!=this->shash||centry->h1!=this->fhash||
            strcmp(centry->name, this->loc)!=0) {

        Error("sendcache: hit sess# %d [%d:%d:%d:%d], entry was changed. may be expired.",
            this->sid, this->fhash, this->shash, this->thash, this->ffhash);

#if 1
                Trace("comp centry %x vs this\n\
%d/%d %d/%d %d/%d %d/%d\n  %s\n  %s\n",
	centry,
    centry->h1, this->fhash, 
	centry->h2, this->shash,
	centry->h3, this->thash,
	centry->h4, this->ffhash, 
    centry->name, this->loc);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif

		goto tobe_miss;
    }



 assert(bk->cup.num==bk->entry_use);
 assert(check_bkentry(bk, "sc3"));

    if(centry->m_recorded == NULL) {

#if 0
        Trace("centry %d, memory is empty, disk is %d\n",
			centry->d_recorded, -1);
#endif
        if(centry->d_recorded>=0 && bk->frag_freelist) {
#if 0
            Trace("DoSendCache: hit sess# %d that object is stored in disk\n",
                this->sid);
#endif

 assert(check_bkentry(bk, "sc4b"));
            chk = skcache_entry_swapin(bk, centry);
 assert(check_bkentry(bk, "sc4a"));
#if 0
            Trace("DoSendCache: sess# %d, that object is stored in disk, %d\n",
                this->sid, chk);
#endif
            if(chk) {
                Trace("sess# %d, fail to swapin oid# %d from disk <%d>\n",
                    this->sid, centry->d_recorded, chk);
                first = NULL;
            }
            else {
                first = centry->m_recorded;
#if 0
                centry->oid = this->sid;
#endif
            }
        }
        else {
            first = NULL;
        }
    }
    else {
        first = centry->m_recorded;
    }

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkentry(bk, "sc5"));

    /* not found */
    if(first==NULL) {
#if 0
        Error("DoSendCache: hit sess# %d [%d:%d:%d]. however, not found fragment <%s,%s>",
            this->sid, this->fhash, this->shash, this->thash,
            this->cachable==SESS_CACHABLE ? "cachable" : "NOTCACHABLE",
            this->reload==SESS_RELOAD ? "RELOAD" : "noreload");
#endif
#if 0
        Trace("  target='%s'\n", this->loc);
        cache_showblockaddr(this->fhash, 0);
#endif

#if 0
        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        hit_reply_count--;
        skip_cache_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_QUERY_ADDR;

        goto out;
#endif

		goto tobe_miss;
    }

#if 0
    Trace("DoSendCache: sess# %d, len %d, pos %d\n",
        this->sid, this->len, this->pos);
#endif

#if 0
    Trace("  object# %d <%p>, '%s'\n\
  create time %d, hot %d, lmd %d, expire %d\n\
  totallength %d\n",
        centry->oid, centry, centry->name,
        centry->issue_time, centry->hot, centry->lmd, centry->expire,
        centry->totallength);
#endif

    if(centry->oid<0) {
        Error("sendcache: sess# %d, stored oid is negative oid# %d",
            this->sid, centry->oid);

#if 1
    Trace("  object# %d <%p>, '%s'\n\
  create time %d, hot %d, lmd %d, expire %d\n\
  totallength %d\n",
        centry->oid, centry, centry->name,
        centry->issue_time, centry->hot, centry->lmd, centry->expire,
        centry->totallength);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }

    /*
     * length check
     */

    if(centry->totallength<=0) {
#if 0
        Trace("sess# %d, length<=0 (length %d, total %d, clen %ld)\n",
            this->sid,
            hlpos->cntl.length, hlpos->cntl.totallength, (long)this->clen);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }
    if(centry->totallength<=0) {
#if 0
        Trace("sess# %d, total<=0 (length %d, total %d, clen %ld)\n",
            this->sid,
            hlpos->cntl.length, hlpos->cntl.totallength, (long)this->clen);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }

	this->oid = centry->oid;



  if((off_t)this->pos==(off_t)0L) {


#if 0
    Trace("sc: sess# %d, first->ent %p, len %d/%d, this->pos %ld\n",
        this->sid, first->ent, first->len, first->size, (long)this->pos);
#endif

    hlen = HTTP_find_header_term( (char*)(first->ent + this->pos), first->len);
    if(hlen<=0) {
#if 1
        Trace("sc: sess# %d, oid# %d, not found header %d/%d\n",
            this->sid, this->oid, hlen, first->len);
#endif

    	dump_stringsegment("sc:header", (char*)(first->ent + this->pos), first->len);

		centry->cracked = 1;
#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }
#if 0
    Trace("sc: hlen %d\n", hlen);
#endif

    chk = HTTP_parse_response_header(
            (char*)(first->ent + this->pos), hlen,
            &xvmajor, &xvminor, &xrcode,
            &xdate, &xlmd, &xexpire, &xclen,
            &xcachable, &xreload);

    if(chk) {
#if 1
        Trace("sc: sess# %d, ignore header\n", this->sid);
#endif
    	dump_stringsegment("sc:header", (char*)(first->ent + this->pos), hlen);

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }
    else {
        this->clen = xclen;
    }
#if 0
    Trace("sc: version %d.%d, rcode %d\n", xvmajor, xvminor, xrcode);
    Trace("sc: lmd %d, expire %d, clen %ld\n", xlmd, xexpire, (long)xclen);
#endif

    if(xrcode<0) {
#if 0
        Trace("sess# %d, rcode is ignore (%d)\n",
            this->sid, xrcode);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }
    this->rcode = xrcode;

#if 0
{
	off_t sum;
	int a;

	sum = (off_t)0L;
	cur = first;
	a = 0;

	while(cur) {
		sum += cur->len;

		if(cur==cur->next) {
			Error("sc; frag looped, broken data was send");
			break;
		}
		cur = cur->next;
		a++;
	}
	Trace("sc: sess# %d, actual size %ld, the number of frag %d\n",	
		this->sid, (long)sum, a);
}
#endif

    if(this->clen>0) {
#if 0
		Trace("sc: sess# %d, oid# %d, f total-hlen-1==clen (%ld - %d - 1 = %ld vs %ld)\n",
       		this->sid, this->oid, (long)centry->totallength, hlen,
           	(long)(centry->totallength - hlen -1), (long)this->clen);
#endif

        if((centry->totallength-hlen-1)!=this->clen) {
#if 1
			Trace("sc: sess# %d, oid# %d, fignoresize\n",
       			this->sid, this->oid);
#endif
			Error("sc: sess# %d stored object size miss match from header",
				this->sid);

			centry->cracked = 1;
#if 0
            this->state = STATE_QUERY_ADDR;

            goto out;
#endif
			goto tobe_miss;
        }
        else {
        }
    }


    this->len = centry->totallength;



    /* cracked check */

    if(centry->cracked) {
#if 0
        Trace("sess# %d, cached object is cracked (cracked=1).  Try retrieval once more.\n", this->sid);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }


    /* expire check */

    if(centry->expire>0 && this->last_action > centry->expire) {
#if 0
        Trace("sess# %d, cached object will be expired (expire %d)\n",
            this->sid, centry->expire);
#endif

#if 0
        this->state = STATE_QUERY_ADDR;

        goto out;
#endif
		goto tobe_miss;
    }



    /*
     *
     */
    if(this->ims>0) {
#if 0
        Trace("DoSendCache: sess# %d, ims %d, lmd %d, expire %d\n",
                this->sid, this->ims, centry->lmd, centry->expire);
#endif
    }


    /*
     *
     */
    centry->count++;
    this->count = centry->count;
#if 0
    hlpos->cntl.hot = reftime.tv_sec;
 assert(check_dup(bcr)==0);
#endif
    centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));

    this->ishit = SESS_HIT;
    this->oid = centry->oid;


	if(bk) {
 assert(check_bkentry(bk, "sc7"));
	}

 }


    if(this->pos>=centry->totallength) {
#if 1
        Trace("sc: sess# %d, all content of object was send pos/len=%d/%d\n",
            this->sid, this->pos, this->len);
#endif
        goto done;
    }



    /*
     * skip to last of sending
     */
	if(this->pos>(off_t)0) 
	{
		off_t sum;
		int a;

#if 0
Trace("sess# %d, this->pos %ld, totallength %ld // seek end\n",
	this->sid, (long)this->pos, (long)centry->totallength);
#endif

#if 0
		sum = (off_t)0L;
		cur = first;
		a = 0;

		while(cur) {
			sum += cur->len;
			Trace(" %4d: %c %p 0x%08x %5ld -> %9ld v.s. %9ld\n",
				a, (sum<this->pos) ? '<' : '-',
				cur, cur->next,
				(long)cur->len, (long)sum, (long) this->pos);

			if(cur==cur->next) {
				Error("sc; frag looped, broken data was send");
				break;
			}
			cur = cur->next;
			a++;
		}
#endif


		sum = (off_t)0L;
		cur = first;
		a = 0;

		do {
			sum += cur->len;

			if(cur==cur->next) {
				Error("sc; frag looped, broken data was send");
				centry->cracked = 1;

				goto done;
			}

			cur = cur->next;
			a++;
		} while(cur && sum<this->pos);

		if(cur==NULL) {
			Error("sess #d, maybe reach to end", this->sid);
			goto done;
		}
#if 0
		Trace("sess# %d, cur %p\n", this->sid, cur);
#endif
	}
	else {
		cur = first;
	}


	if(bk) {
 assert(check_bkentry(bk, "sc6"));
	}

    /*
     *
     */

    sum = 0;
/*
    cur = first;
*/

    do
    {
        l = 0;
        do {
            iov[l].iov_base = cur->ent;
            iov[l].iov_len = cur->len;
            if(cur->len>0) {
                sum += cur->len;
            }

            l++;

            cur = cur->next;
        } while(cur && l < max_frag_per_write && cur->len>0 );
        ciov = l;

        if(this->cfd<0) {
            break;
        }

try_write:
#if 0
Trace("sess# %d, iov %d/%d\n", this->sid, ciov, max_frag_per_write);
#endif
        nw = writev(this->cfd, iov, ciov);
        if(nw<0) {
            if(errno==EINTR) {
                goto try_write;
            }
            if(errno==EPIPE) {
                DoHalfClientCloseSess(this);
                goto done;
            }
            Error("sess# %d, cannot write to client fd# %d, ciov %d, (%d)",
            this->sid, this->cfd, ciov, errno);
#if 1
            this->reason = R_FAIL_CLIENT_WRITE;
            this->state = STATE_CLOSE;
#endif
#if 0
            this->reason = R_FAIL_CLIENT_WRITE;
            DoHalfClientCloseSess(this);
#endif
            goto out;
        }

        this->cwrite += (off_t) nw;
        this->pos += nw;

    } while(sum<max_frag_per_write && cur!=NULL);

#if 1
	this->last_action = time(NULL);
#endif

#if 0
    Trace("sc: sess# %d, oid# %d, tl %ld cw %ld, pos %ld\n",
     	this->sid, this->oid,
		(long)centry->totallength, (long)this->cwrite, (long)this->pos);
#endif
#if 0
    Trace("sc: sess# %d, oid# %d, r cw %ld hlen %d -1 = %ld vs clen %ld\n",
     	this->sid, this->oid,
		(long)this->cwrite, hlen, (long)(this->cwrite - hlen -1),
		(long)this->clen);
    if((this->cwrite - hlen - 1)!=this->clen) {
		Trace("sc: sess# %d, does not reach to end\n", this->sid);
	}
	else {
		Trace("sc: sess# %d, reach to end\n", this->sid);
	}
#endif

    if(this->pos>=centry->totallength) {
        goto done;
    }

still:
	/* do nothing */
#if 0
	Trace("sc: sess# %d, continued.\n", this->sid);
#endif
	goto out;

done:
    this->state = STATE_CLOSE;
    if(centry) {
#if 0
        centry->hot = 0;
 assert(CHKTIME(centry->hot));
#endif

    }
	goto out;


tobe_miss:

	/* when enter this routine, the program increment hit counter.
	 * if miss, the program has to decrement that and increment miss counter.
	 */
	MUTEX_LOCK(&stat_lock);
	miss_reply_count++;
	hit_reply_count--;
	skip_cache_count++;
	MUTEX_UNLOCK(&stat_lock);

	this->state = STATE_QUERY_ADDR;

	goto out;

out:
#if 0
    if(this->state==STATE_QUERY_ADDR) {
        Trace("  change to STATE_QUERY_ADDR\n");
    }
#endif
    if(bk) {
 assert(check_bkentry(bk, "sc9"));
        UNLOCK_BK(bk);
    }



    return 0;
}





/* support only IPv4 */
static
int
store_scrambled_name(char *dst, struct sockaddr_in *addr, int sc_key)
{
    register char *p, *q;
    register int i;
    register int c;

    q = (char*) &addr->sin_addr;

    c = 0xc634d3a5;
    c |= sc_key;
    for(i=3;i>=0;i--) {
        p = &q[i];
        c += *p;
        c = c << 3 + 0x01;

#if 0
        Trace(" #%d %3u %2x -> %12u %12x\n",
            i, (unsigned int)*p, (unsigned int)*p,
            (unsigned int)c, (unsigned int)c);
#endif
    }

    sprintf(dst, "c%u", c);

    return 0;
}

static
int
store_direct_name(char *dst, struct sockaddr_in *addr)
{
    register char *p, *q;
    register int i;
    register int c;

    p = (char*) &addr->sin_addr;
    q = dst;

    for(i=0;i<4;i++) {
        if(*p>100)
            *q++ = '0' + (*p / 100);
        if(*p>10)
            *q++ = '0' + ((*p % 100)/10);
        *q++ = '0' + (*p % 10);
        *q++ = '.';
        *q = '\0';

#if 0
        Trace(" #%d %3u %2x -> '%s'\n",
            i, (unsigned int)*p, (unsigned int)*p, dst);
#endif

        p++;
    }

    return 0;
}


static
int client_scramble_key;

void*
Listener(void *called_option)
{
    int chk;
    struct sockaddr_in from;
    int len;
    int cfd;
    int find;
    int p, c;
    int op;
    int listen_fd;
    sess_t *sessp;
    int t_req, t_rjq;
    int t_nhot;
    int err;
    int try_c;
    struct timeval tmp_accept_time;
    char tmp_clientname[L_HOST];
    unsigned char *caddr;
    off_t nr, nw;


    listen_fd = *(int*)called_option;
    len = sizeof(from);

    while(1) {
try_accept:
#ifdef DO_ACCEPT_LOCK
        chk = MUTEX_LOCK(&accept_lock);
        if(chk) {
            Error("Listener: fail lock (%d)", chk);
            if(chk==EBUSY) {
                Trace(" ---  busy\n");
                goto try_accept;
            }

            goto try_accept;
        }
#endif

        err = errno = -1;
        cfd = accept(listen_fd, (struct sockaddr*)&from, &len);
        err = errno;
        gettimeofday(&tmp_accept_time, NULL);

#if 0
        Trace("Listener<%p> accept fd# %d (errno %d)\n",
            pthread_self(), cfd, err);
#endif

#ifdef DO_ACCEPT_LOCK

        try_c=0;
try_unlock:
        try_c++;

        chk = MUTEX_UNLOCK(&accept_lock);
        if(chk) {
            Error("Listener: fail unlock (%d) [%d]", chk, try_c);
            goto try_unlock;
        }
#endif

        if(cfd<0) {
            if(err==EINTR)
                goto try_accept;
#if 0
            if(err==ENFILE || err==EMFILE) {
                sleep(5);
                /*
                exit(143);
                */
            }
#endif

            Error("accept error (%d)", err);

            goto next;
        }


        MUTEX_LOCK(&stat_lock);
        receive_request_count++;
        request_sirial_count++;
        t_req = request_sirial_count;
        MUTEX_UNLOCK(&stat_lock);

#if 0
        Trace("th %d, conn. count %6d, cfd %3d\n",
            pthread_self(), t_req, cfd);
#endif


        /*
         * store peer (client or lower proxy) name
         */
#ifdef LOG_DIRECTCLIENTIPADDRESS
        store_direct_name(tmp_clientname, &from);
#else
        store_scrambled_name(tmp_clientname, &from, client_scramble_key);
#endif

#if 0
        Trace("peer <%s>\n", tmp_clientname);
#endif


        /*
         * check permition
         */
        caddr = (unsigned char*) &from.sin_addr.s_addr;
        if(!Acceptable(caddr)) {
            goto not_permitted;
        }


found_free:

        sessp = (sess_t*)sq_get_free();
        if(sessp==NULL) {
            Error("not enough memory for session");
            goto not_found_free;
        }


        /*
         * NOTE:
         *   # of hot session will be discrease.
         *   however, it is not increase. because increasing is only
         *   happen in this routines.
         */

        MUTEX_LOCK(&sess_lock);
        nhot++;
        MUTEX_UNLOCK(&sess_lock);



        chk = pthread_mutex_init(&(sessp->busy), NULL);

        sessp->sid    = t_req;
        sessp->psid   = -1;
        sessp->mode   = MODE_CLIENT;
        sessp->level  = 0;
        sessp->state  = STATE_ACCEPT;

        sessp->accept_time = tmp_accept_time;
        sessp->last_action = tmp_accept_time.tv_sec;

        sessp->reported = 0;
        sessp->reason = -1;

        sessp->pos    =  0;
        sessp->len    = -1;
        sessp->rcode  = -1;
        sessp->count  = 0;
        sessp->lmd    = -1;
        sessp->expire = -1;

        sessp->cver_major = -1;
        sessp->cver_minor = -1;
        sessp->sver_major = -1;
        sessp->sver_minor = -1;

        sessp->cachable = SESS_CACHABLE;
        sessp->reload   = SESS_NORELOAD;
        sessp->ishit    = SESS_MISS;
        sessp->publish  = SESS_PUBLIC;

        sessp->cache_entry = NULL;

        sessp->oid    = -1;

        sessp->cfd    = cfd;
        sessp->sfd    = -1;
        sessp->rfd    = -1;
        sessp->rcount = 3;

        strcpy(sessp->clientname, tmp_clientname);

        memset(sessp->loc, 0, sizeof(sessp->loc));
        sessp->method = M_UNKNOW;

        memset(sessp->reqbuf, 0, sizeof(sessp->reqbuf));
        sessp->reqlen = 0;

        sessp->ims = -1;

#ifdef TRACE_SESS
        Trace("sess# %d, new, addr %p\n", sessp->sid, sessp);
#endif

        sessp->state  = STATE_RECV_REQ;


        chk = push_sess(think_queue, sessp);
        if(chk) {
            Error("Listener: sess# %d, fail push to think Q", sessp->sid);
            sessp->reason = R_MEMORY_OVERFLOW;

            DoSendMessage(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;

            goto next;
        }

#if 0
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

        goto next;



not_found_free:
#if 0
        Error("not found idle session cell");
#endif
        MUTEX_LOCK(&stat_lock);
        reject_request_count++;
        t_rjq = reject_request_count;
        MUTEX_UNLOCK(&stat_lock);
#if 1
        if(t_rjq%100==0) {
            LogLine(": ; session list overflow");
        }
#endif


        nw = HTTP_SendIgnoreMessage(cfd, 501,
            "service denided", "please wait few minutus");
        shutdown(cfd, 2);
        close(cfd);
        cfd = -1;

        MUTEX_LOCK(&stat_lock);
        intra_outgoing += (off_t) nw;
        MUTEX_UNLOCK(&stat_lock);

        goto next;



not_permitted:
#if 0
        Error("not permitted host");
#endif
        MUTEX_LOCK(&stat_lock);
        reject_request_count++;
        t_rjq = reject_request_count;
        MUTEX_UNLOCK(&stat_lock);
#if 1
        LogLine(": ; request from not permitted host <%s>");
#endif

        nw = HTTP_SendIgnoreMessage(cfd, 403,
            "access denided", "you are not permitted this server");
        shutdown(cfd, 2);
        close(cfd);
        cfd = -1;

        MUTEX_LOCK(&stat_lock);
        intra_outgoing += (off_t) nw;
        MUTEX_UNLOCK(&stat_lock);

        goto next;



next:
        chk = 0;

    }

}



int
DoParseAnchor(sess_t *this)
{
    int chk;
    int rest;
    int nw;
    int l;

    bucket_t *bk;
    skcache_entry_t *centry;
    skcache_frag_t *first, *cur;

    int t_req=1234; /* dummy */

    int nc;
    int ntag;
    int ihave;

    int howprefetch;

    tags_t  *tags;
    char    **urllist;
    int     nofref;

    sess_t *psessp;
    char **aurl;
    char nurl[L_URL];

    int fh, sh, th, ff;

#ifdef TRACE_THREADS
    Trace("DoParseAnchor: sid %d, [%d:%d]\n",
        this->sid, this->fhash, this->shash);
#endif



    tags = NULL;

    if(this->cachable==SESS_UNCACHABLE) {
#ifdef TRACE_PREFETCH
        Trace("DoParseAnchor: sess# %d is not cachable\n", this->sid);
#endif
        goto out_nothing;
    }
#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor: sess# %d, level %d, URL '%s'\n",
        this->sid, this->level, this->loc);
#endif

    /* allow level 0 or 1 */
    if(this->level<0 || this->level>2) {
#if 0
        Trace("sess# %d %p, strange level %d\n",
            this->sid, this, this->level);
#endif
        goto out;
    }

    /*
     *
     */
    bk = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);
#if 0
	/* XXX */
    if(chk) {
        Trace("DoParseAnchor: fail lock block #%d [%d:%d:%d]\n",
            this->fhash, this->fhash, this->shash, this->thash);
        goto out_pending;
    }
#endif

    centry = (skcache_entry_t*) this->cache_entry;

    /* found object */
    if(centry==NULL) {
        Trace("sess# %d, no entry in block# %d [%d:%d:%d]\n",
            this->sid, this->fhash, this->fhash, this->shash, this->thash);
#if 0
        cache_showblockaddr(this->fhash, 0);
#endif

        UNLOCK_BK(bk);
        bk = NULL;
        goto not_found;
    }

    first = centry->m_recorded;
/* XXX */
#if 0
    if(centry->d_recorded>=0) {
    }
#endif
    if(first==NULL) {
        Trace("sess# %d, no frag in block# %d [%d:%d]\n",
            this->sid, this->fhash, this->fhash, this->shash);
#if 0
        cache_showblockaddr(this->fhash, 0);
#endif

		UNLOCK_BK(bk);
        bk = NULL;
        goto not_found;
    }


    /***
     ***
     ***/

    tags = HTML_NewTags();

    if(this->level==0) {
#ifdef TRACE_PREFETCH
        Trace("parse refered: '%s' %db\n",
            this->loc, first->length);
#endif
        chk = MUTEX_LOCK(&config_lock);
        howprefetch = num_prefetch_referedpages;
        chk = MUTEX_UNLOCK(&config_lock);

        ntag = HTML_ParseTags_FromMemory(tags, PARSE_REFERED,
                first->ent, first->len);
#ifdef TRACE_PREFETCH
        Trace("parse refered: #%d %ldb %d\n",
            this->sid, first->length, ntag);
#endif
    }
    else
    if(this->level==1) {
#ifdef TRACE_PREFETCH
        Trace("parse included: '%s' %db\n",
            this->loc, first->length);
#endif
        chk = MUTEX_LOCK(&config_lock);
        howprefetch = num_prefetch_includedobjs;
        chk = MUTEX_UNLOCK(&config_lock);

        ntag = HTML_ParseTags_FromMemory(tags, PARSE_INCLUDED,
                first->ent, first->len);

#ifdef TRACE_PREFETCH
        Trace("parse included: #%d %db %d\n",
            this->sid, first->length, ntag);
#endif
    }


#if 0
    dump_stringchunk("entity", (char*)&line_pos->body);
    Trace("there are %d tag(s).\n", ntag);
#endif

	UNLOCK_BK(bk);
    bk = NULL;
    centry = NULL;
    first = NULL;




    if(ntag<=0) {
        FREE(tags);
        tags = NULL;
        goto out;
    }

    if(tags==NULL) {
        goto out;
    }
    else {
        if(tags->count==0) {
            FREE(tags);
            tags = NULL;
            goto out;
        }
    }

    urllist = HTML_ConvURLList(tags, &nofref);
    HTML_DestroyTags_withoutvalue(tags);
    tags = NULL;



    /*
     *
     */

#if 0
    /*
    HTML_ShowIndependURLList(this->loc, urllist);
    */
    HTML_ShowURLList(urllist);
#endif

#ifdef TRACE_PREFETCH
    if(urllist) {
        Trace("DoParseAnchor<%d>: decide prefetch targets about\n  '%s'\n",
            pthread_self(), this->loc);
    }
    else {
        Trace("DoParseAnchor<%d>: prefetch target list is empty\n");
    }
#endif

    nc = 0;
    aurl = urllist;
    while(*aurl!=NULL && nc<howprefetch) {
#if 0
        Trace("  newone '%s'\n", *aurl);
#endif


        URL_JoinStr(nurl, L_URL, this->loc, *aurl);
        if(nurl[0]=='\0') {
            goto next;
        }
#ifdef TRACE_PREFETCH
        Trace("  full URL '%s'\n", nurl);
#endif

#if 0
        fh = first_hash(nurl);
        sh = second_hash(nurl);
        th = third_hash(nurl);
        ff = fourth_hash(nurl);
#endif

/*
		calc_hash(this->loc, &fh, &sh, &th, &ff);
*/
		calc_hash(nurl, &fh, &sh, &th, &ff);
#if 0
Trace("pan: %03d:%03d:%03d:%03d %s\n", 
		fh, sh, th, ff, this->loc);
#endif

#if 0
        Trace("  check object existance in cache\n");
#endif
#if 0
        ihave = cache_have(fh, sh, nurl, 1, 0);
#endif
        ihave = skcache_have(fh, sh, th, ff, nurl, 1, 0);

#if 0
        Trace("  ihave %d\n", ihave);
#endif
        if(ihave>0) {
            goto next;
        }
        /* URL is not empty and it is not hold in cache */
        else
        if(ihave==0) {
            psessp = (sess_t*)sq_get_free();
            if(psessp==NULL) {
                Error("not enough memory for new prefetching session");
                goto skip;
            }

            chk = pthread_mutex_init(&(psessp->busy), NULL);

            MUTEX_LOCK(&sess_lock);
            nhot++;
            MUTEX_UNLOCK(&sess_lock);


#if 1
            MUTEX_LOCK(&stat_lock);
            request_sirial_count++;
            t_req = request_sirial_count;

            issue_prefetch_count++;

            MUTEX_UNLOCK(&stat_lock);
#endif


            /***
             ***
             ***/

            psessp->sid    = t_req;
            psessp->psid   = this->sid;
            psessp->mode   = MODE_SELF;
            psessp->level  = this->level+1;
            psessp->state  = STATE_THINK;
            psessp->last_action = this->last_action;

            gettimeofday(&(psessp->accept_time), NULL);

            psessp->reported = 0;
            psessp->reason  = -1;

            psessp->pos     =  0;
            psessp->len     = -1;
            psessp->rcode   = -1;
            psessp->count   = 0;
            psessp->lmd     = -1;
            psessp->expire  = -1;

            psessp->cachable = SESS_CACHABLE;
            psessp->reload   = SESS_NORELOAD;
            psessp->ishit    = SESS_MISS;
            psessp->publish  = SESS_PUBLIC;

            psessp->oid     = -1;

            psessp->cfd     = -1;
            psessp->sfd     = -1;
            psessp->rfd     = -1;
            psessp->rcount  = 2;


            strcpy(psessp->clientname, "self");

            memset(psessp->loc , 0, sizeof(psessp->loc));
            memset(psessp->reqbuf, 0, sizeof(psessp->reqbuf));
            psessp->reqlen = 0;

            psessp->cver_major = 0;
            psessp->cver_minor = 0;
            psessp->sver_major = -1;
            psessp->sver_minor = -1;
            psessp->ims = -1;

            DoSetReq(psessp, nurl);

#if 0
            dump_stringchunk("generated request", psessp->reqbuf);
#endif

#ifdef TRACE_PREFETCH
            Trace("nurl = %4d-%02d %p (#%d) '%s'\n",
                this->sid, nc, psessp, psessp->sid, psessp->loc);
#endif


            chk = DoThink(psessp);
            if(chk==0 && psessp->state==STATE_QUERY_ADDR) {
                DoFindServer(psessp);
                if(psessp->state==STATE_CONN_SERV) {
#if 0
                    Trace("connect prefetch target '%s'\n", psessp->loc);
#endif
                    DoServerConnect(psessp);
                }

#ifdef TRACE_PREFETCH
                Trace("sess# %d, <self>   method <%s>,URL '%s' version ?\n",
                    psessp->sid, MSG_M_[psessp->method], psessp->loc);
#else
#ifdef TRACE_SESS
                Trace("sess# %d, <self>   method <%s>,URL '%s' version ?\n",
                    psessp->sid, MSG_M_[psessp->method], psessp->loc);
#endif
#endif

#if 0
                Trace("     push to think Q as sess# %d %s\n",
                    psessp->sid, MSG_STATE_[psessp->state]);
#endif
                if(psessp) {
                    chk = push_sess(think_queue, psessp);
                    if(chk) {
                        Error("DoParseAnchor: sess# %d, fail push to think Q",
                            psessp->sid);
						sq_be_free(psessp);
						psessp = NULL;
                    }
                    psessp = NULL;

#if 0
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

                }
            }
            else {
                DoReleaseSess(psessp);
                psessp = NULL;
            }

#if 0
            Trace("  ..done\n");
#endif

skip:
            (void)1;    /* dummy for label */
#if 0
            Trace("  skip\n");
#endif

        }
        else {
            /*
            Trace("DoParseAnchor: fail cache_have(), return %d\n", ihave);
            */
        }
next:
#if 0
        Trace("  next\n");
#endif

        aurl++;
        nc++;
#if 0
        Trace("  ? next\n");
#endif
    }

#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor<%d>: loop end\n");
#endif

    HTML_FreeURLList(urllist);
    FREE(urllist);
    urllist = NULL;

    tags = NULL;


#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor<%d>: block end\n");
#endif

    goto out;

done:

not_found:
#if 0
    /*
    Error("DoParseAnchor: cache was not found #%d [%d:%d] when hit",
        this->sid, this->fhash, this->shash);
    */
    Error("DoParseAnchor: sess#%d [%d:%d:%d] does not have cache entry",
        this->sid, this->fhash, this->shash, this->thash);
#endif
    goto out;


out:

out_nothing:

    return 0;


out_pending:

    return 1;

}




/*
 * Handler
 *
 */

void*
Handler(sess_t *sessp, time_t now)
{
    int chk;
    int cv;
    int pending;

	int m_sid;

    if(sessp==NULL)
        return NULL;


	m_sid = sessp->sid;


#ifdef TRACE_SCHEDULE
    Trace("Handler<%p> start, sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
#endif

#if 0
    if(sessp->rcode<0) {
        Trace("sess# %7d %14s %3d\n",
            sessp->sid, MSG_STATE_[sessp->state], sessp->rcode);
    }
#endif

    sessp->last_action = now;
	sessp->tic++;


    switch(sessp->state) {
    /*
     * like manager
     */
    case STATE_ACCEPT:
    case STATE_RECV_REQ:

        if(sessp->mode==MODE_CLIENT && sessp->cfd>=0) {
#if 0
            Trace("client request (%d) and client fd (%d) is normal\n",
                    sessp->mode, sessp->cfd);
#endif

            if(sessp->reqlen<4) {
                chk = DoReadReqFirst(sessp);
            }
            else
            if(sessp->reqlen > L_REQBUF) {
                Error("request is too large");
                sessp->reason = R_TOOLONG_URL;
                goto ignore;
            }
            else {
                chk = DoReadReqRest(sessp);
            }

            if(sessp->state==STATE_MESSAGE)
                goto ignore;

            if(chk) {
#if 1
                Error("handler: fail client request parsing");
#endif
                goto fail;
            }
        }
        else {
#if 0
            Trace("self request (%d) or client fd (%d) is too small\n",
                    sessp->mode, sessp->cfd);
#endif
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 5= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        if(sessp->state!=STATE_THINK) {
            break;
        }
#if 0
		Trace("falling THINK sess# %d,\n", sessp->sid);
#endif

    case STATE_THINK:
#if 0
		Trace("sess# %d, call think()\n", sessp->sid);
#endif
        chk = DoThink(sessp);
        if(chk) {
            if(chk==-1) {
                sessp->reason = R_SOME_FATAL_ERROR;
                sessp->state = STATE_MESSAGE;
                goto ignore;
            }

            Error("Handler: fail session dicision");
            goto fail;
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 6= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        switch(sessp->state) {
        case STATE_REPORT:
            DoReport(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;
            goto out;

        case STATE_WAIT_FRIEND:

#ifdef USE_WQUEUE
			/* THINK -> WAIT_FRIEND */
/*
			Trace("TH->WF, sess# %d, enter wait Q %d/%d\n",
				sessp->sid, wait_queue->count, wait_queue->cmax);
*/
			chk = push_sess(wait_queue, sessp);
			if(chk) {
				goto conn;
			}
			sessp = NULL;

			goto out;
#else
			Trace("TH->WF\n");
#endif
            break;

        case STATE_QUERY_ADDR:
            DoFindServer(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 21= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
            if(sessp->state==STATE_CONN_SERV) {
                DoServerConnect(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 22= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
            }

            break;

        case STATE_SEND_CACHE:
            DoSendCache(sessp);

            if(sessp->state==STATE_QUERY_ADDR) {
#if 0
Trace("sess# %d, change to fetch\n", sessp->sid);
#endif
                goto start_fetch;
            }
            break;

        case STATE_MESSAGE:
            goto ignore;

 /*NOTREACHED*/
            break;

        case STATE_DONE:
        case STATE_CLOSE:
            goto done;

#if 1
		/* XXX */
        case STATE_THINK:
/*
			Trace("TH -> TH sess# %d, \n", sessp->sid);
*/
			goto conn;
#endif

        default:
            Error("Handler: sess# %d, strange state (%d) after DoThink() ",
                sessp->sid, sessp->state);
            goto fail;
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 7= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;


    case STATE_CLOSE:
    case STATE_DONE:
        goto done;

 /*NOTREACHED*/
        break;

    case STATE_WAIT_FRIEND:
        DoCheckEndFriendSess(sessp);

#ifdef USE_WQUEUE
		if(sessp->state==STATE_WAIT_FRIEND) {
Trace("WF->WF, re-push sess# %d, to wait Q\n", sessp->sid);


			chk = push_sess(wait_queue, sessp);
			if(chk) {
				goto conn;
			}

			goto out;
		}
		else {
Trace("WF -> %s sess# %p,\n", MSG_STATE_[sessp->state], sessp->sid);
		}
#endif

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 93= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        break;

    case STATE_SEND_CACHE:
        DoSendCache(sessp);

        if(sessp->state==STATE_QUERY_ADDR) {
#if 0
Trace("sess# %d, change to fetch\n", sessp->sid);
#endif
            goto start_fetch;
        }

        break;

    case STATE_QUERY_ADDR:
start_fetch: (void)1;

        DoFindServer(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 11= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        if(sessp->state==STATE_CONN_SERV) {
            DoServerConnect(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 12= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        }

        break;

    case STATE_WAIT_ADDR:

        chk = hcache_recv_response_addr_direct(sessp->servername,
            &(sessp->serveraddr), sessp->rfd);
        sessp->rfd = -1;
        if(chk) {
            Error("fail hcache_recv_response_addr_direct() to '%s', ret %d <%d>",
                sessp->servername, chk, sessp->rcount);

            sessp->rcount--;
            if(sessp->rcount>0) {
                sessp->state = STATE_QUERY_ADDR;
                goto start_fetch;
            }

            sessp->reason = R_RESOLVER_ERROR;
            sessp->state = STATE_MESSAGE;
        }
        else {
#ifdef TRACE_HCACHE
            Trace("  got address via hcache_recv_response_addr_direct()\n");
#endif
            chk = hcache_set_addr_incache(sessp->servername, &(sessp->serveraddr));

#if 0
            hcache_list();
#endif

            sessp->state = STATE_CONN_SERV;
        }

        if(sessp->state!=STATE_CONN_SERV) {
            goto ignore;
        }

    case STATE_CONN_SERV:

        DoServerConnect(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 10= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;

#if 0
    case STATE_WAIT_CONN_SERV:

        DoWaitServerConnect(sessp);
        break;
#endif

    case STATE_WAIT_CONN_SERV:
    case STATE_SEND_REQUEST:

        DoSendRequest(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 13= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        break;

    case STATE_RECV_RES:

        DoServerReply(sessp);

/*
		Trace("RECV_RES -> %s, sess# %d, \n",
			MSG_STATE_[sessp->state], sessp->sid);
*/
		
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 9= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;

    case STATE_RELAY_DIRECT:

        DoRelayDirect(sessp);
        break;


    default:
        Error("Handler: strange state (%d) ", sessp->state);
        goto fail;

    } /* end of state switch */


    if(sessp!=NULL &&
        sessp->state==STATE_MESSAGE) {
        goto ignore;
    }


done:

    if(sessp!=NULL &&
        (sessp->state==STATE_CLOSE || sessp->state==STATE_DONE)) {

#if 0
	Trace("Handler: done, sess# %d, re-push to think Q\n", sessp->sid);
#endif



        release_buf(sessp);
        DoCloseSess(sessp);

        MUTEX_LOCK(&stat_lock);
        process_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        chk = MUTEX_LOCK(&config_lock);
        cv = prefetch_flag;
        chk = MUTEX_UNLOCK(&config_lock);

#ifdef TRACE_PREFETCH
        if(!cv) {
            Trace("Handler: prefetch is disabled\n");
        }
        else
#endif

        if(cv &&
            (sessp->level<2 && sessp->rcode>=200 && sessp->rcode<=399) ) {
/*
        sessp->sread > 8
        sessp->swrite > 10
 */

            pending = DoParseAnchor(sessp);
            if(pending) {
                Trace("Handler: pending sess# %d\n", sessp->sid);
            }
        }
#ifdef TRACE_PREFETCH
        else {
            Trace("Handler: sess# %d, skip parse object content [level %d, rcode %d]\n",
                sessp->sid, sessp->level, sessp->rcode);
        }
#endif

        DoReleaseSess(sessp);
        sessp = NULL;

        goto out;
    }


    goto conn;

ignore:
#if 0
	Trace("Handler: ignore, sess# %d, close.\n", sessp->sid);
#endif


    DoSendMessage(sessp);
    sessp->state = STATE_CLOSE;

    DoCloseSess(sessp);
    DoReleaseSess(sessp);
    sessp = NULL;

    goto out;

fail:
#if 0
	Trace("Handler: fail, sess# %d, close.\n", sessp->sid);
#endif


    DoCloseSess(sessp);
    DoReleaseSess(sessp);
    sessp = NULL;


    goto out;

conn:
    if(sessp!=NULL) {

#if 0
#ifdef TRACE_SCHEDULE
    Trace("Handler<%p> end,   sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
#endif
#endif


#if 0
        Trace("### A\n");
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

#if 0
		Trace("Handler: conn, sess# %d, re-push to think Q\n", sessp->sid);
#endif


        chk = push_sess(think_queue, sessp);
        if(chk) {
            Error("Handler: sess# %d, fail re-push to think Q", sessp->sid);

			sessp->reason = R_MEMORY_OVERFLOW;
            DoSendMessage(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;
        }
		else {
        	sessp = NULL;
		}

#if 0
        Trace("### B\n");
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

    }

    goto out;

out:
    chk = 0;

#if 0
    Trace("### C\n");
    MUTEX_LOCK(&think_queue->lock);
    fflush(stderr);
    sq_show(stdout, think_queue);
    fflush(stdout);
    MUTEX_UNLOCK(&think_queue->lock);
#endif


#ifdef TRACE_SCHEDULE
	if(sessp) {
    Trace("Handler<%p> end,   sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
	}
	else {
    Trace("Handler<%p> end,   lost, sess# %d, \n",
		pthread_self(), m_sid);
	}

    Trace("Handler<%p> nhot %d, think queue %d/%d, wait_queue %d/%d\n",
		pthread_self(), nhot, think_queue->count, think_queue->cmax,
		wait_queue->count, wait_queue->cmax);
#endif

    return NULL;
}

inline
int
check_timeout(sess_t *sessp, time_t now)
{
    int ret;
    int idle, elap;

    ret = 0;

    idle = now - sessp->last_action;
    elap = now - sessp->accept_time.tv_sec;

    switch(sessp->state) {
    case STATE_ACCEPT:
    case STATE_RECV_REQ:
#if 0
    Trace("check_timeout: sess# %d, now %d, idle %d, elap %d, reqlen %d\n",
        sessp->sid, now, idle, elap, sessp->reqlen);
#endif
        if(sessp->last_action == sessp->accept_time.tv_sec) {

            if(sessp->reqlen>0) {
#if 0
    Trace("    Is it empty request ?\n");
#endif
            }

            if(idle > RECEPT_TIMEOUT*3) {
                sessp->reason = R_CLIENT_TIMEOUT;
                ret = 1;
            }
        }
        else
        if(idle > RECEPT_TIMEOUT) {
            sessp->reason = R_CLIENT_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_FRIEND:
        if(idle > WFRIEND_TIMEOUT) {
            sessp->reason = R_WFRIEND_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_CONN_SERV:
    case STATE_RECV_RES:
        if(idle > FETCH_TIMEOUT) {
            sessp->reason = R_SERVER_TIMEOUT;
            ret = 1;
        }
        break;
    default:
        if(idle > ELSE_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
        break;
    }

    if(ret==0) {
        if(elap > SESSION_TOTAL_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
    }

#if 0
if(ret!=0) {
    Trace("sess# %d, elap/idle %d/%d -> reason %d, ret %d\n",
        sessp->sid, elap, idle, sessp->reason, ret);
}
#endif

    return ret;
}



#if 0
#define CHECK {\
    int k;\
    Trace("w_sess_count %d, w_fd_count %d\n", w_sess_count, w_fd_count);\
    if(w_sess_count>0) {\
        Trace("index sess     sid#  state           |cfd   sfd\n");\
        for(k=0;k<w_sess_count;k++) {\
            if(sess_chunk[k]) \
                Trace("%5d %8x %5d %-15s |%5d %5d \n",\
                    k, sess_chunk[k],\
                    sess_chunk[k]->sid, MSG_STATE_[sess_chunk[k]->state],\
                    sess_chunk[k]->cfd, sess_chunk[k]->sfd\
                    );\
            else\
                Trace("%5d empty\n", k);\
        }\
    }\
    else {\
        Trace("***no sessions\n");\
    }\
    if(w_fd_count>0) {\
        Trace("\nindex fd    |sess   sid# state\n");\
        for(k=0;k<w_fd_count;k++) {\
            if(bound[k]) \
                Trace("%5d %5d |%8x %5d %-15s\n",\
                    k, fdrec[k].fd,\
                    bound[k], bound[k]->sid, MSG_STATE_[bound[k]->state]);\
            else\
                Trace("%5d %5d empty\n", k, fdrec[k].fd);\
        }\
    }\
    else {\
        Trace("***no fds\n");\
    }\
    Trace("= = \n");\
}

#else
#define CHECK   ((void)0)
#endif

static
int
sess_comp(const void *ap, const void *bp)
{
    sess_t *a = *(sess_t**)ap;
    sess_t *b = *(sess_t**)bp;

    if(a->fhash==b->fhash) {
        return a->shash - b->shash;
    }
    else {
        return a->fhash - b->fhash;
    }
}



void*
Slave(void *dummy)
{
    time_t now;
    int w_sess_max;
    int w_sess_count;
    int w_fd_count;
    int timeout_count;
    sess_t *sess_chunk[SESS_CHUNK_SIZE];
/*
    sess_t *wsess_chunk[SESS_CHUNK_SIZE];
*/
    struct pollfd fdrec[SESS_CHUNK_SIZE*2];
    sess_t *bound[SESS_CHUNK_SIZE*2];
    int i, j;
    int chk;
    int try;
    int touched;
    int same;

	int focus;
	int c;

	int miss_capability=4;

	int sq_show_w_lock(FILE*, sess_queue_t*);



#ifdef TRACE_SCHEDULE
    Trace("Slave: thread %8x lunched\n",
        pthread_self());
#endif


    w_sess_max = require_chunk_size;

    /* forever loop */
    while(1) {
        timeout_count = 0;
        w_sess_count = 0;
        w_fd_count = 0;

        for(i=0;i<SESS_CHUNK_SIZE;i++) {
            sess_chunk[i] = NULL;
            bound[i] = NULL;
        }

#if 0
        Trace("\nSlave: call sq_pullchunk with %d\n", w_sess_max);
#endif
#if 0
		sq_show_w_lock(stderr, think_queue);
		sq_show_w_lock(stderr, wait_queue);
		sq_show_w_lock(stderr, msweep_queue);
#endif


#if 0
		Trace("\n\n  call chunk nhot %d, queue %d+%d 0\n",
			nhot, think_queue->count, wait_queue->count);
#endif

        w_sess_count = sq_pullchunk(think_queue, sess_chunk, w_sess_max);

/*
Trace("w_sess_count %d\n", w_sess_count);
*/
        if(w_sess_count<=0) {
/*
Trace("empty\n");
*/
            goto next;
        }
        now = time(NULL);


#if 0
		if(nhot == w_sess_count + think_queue->count + wait_queue->count) {
			Trace("maybe correct %d vs %d (%d+%d+%d) 1\n",
				nhot, w_sess_count+think_queue->count+wait_queue->count,
				w_sess_count, think_queue->count, wait_queue->count);
		}
		else {
			Trace("maybe *NOT* correct %d vs %d (%d+%d+%d) 1\n",
				nhot, w_sess_count+think_queue->count+wait_queue->count,
				w_sess_count, think_queue->count, wait_queue->count);
		}
#endif


        /*
         *
         */
#if 0
        Trace("\nSlave: pid# %d, th# %d, got %d/%d session(s)\n",
			getpid(), pthread_self(), w_sess_count, w_sess_max);
        for(i=0;i<w_sess_count;i++) {
			Trace("# %4d: %p sess# %d, %s\n",
				i, sess_chunk[i], sess_chunk[i]->sid, 
                    MSG_STATE_[sess_chunk[i]->state]);
		}
#endif

        /*
         *
         */
        for(i=0;i<w_sess_count;i++) {

#if 0
			sess_chunk[i]->tic++;
#endif

            if(sess_chunk[i]->last_action>0
                && check_timeout(sess_chunk[i], now)) {

#if 1
                /* XXX */
                if(sess_chunk[i]->reason==R_WFRIEND_TIMEOUT) {
Trace("sess# %d, switch wait FRIEND to retrieve\n", sess_chunk[i]->sid);
Log("sess# %d, switch wait FRIEND to retrieve", sess_chunk[i]->sid);
/*
*/

                    sess_chunk[i]->state = STATE_QUERY_ADDR;
                    sess_chunk[i]->cachable = SESS_UNCACHABLE;

/*
                    goto next;
*/
					continue;
                }
#endif

                release_buf(sess_chunk[i]);
                DoTimeoutClose(sess_chunk[i]);
                DoReleaseSess(sess_chunk[i]);
                timeout_count++;

                sess_chunk[i] = NULL;


				continue;
            }

            switch(sess_chunk[i]->state) {
            case STATE_IDLE:
                break;
            case STATE_ACCEPT:
            case STATE_RECV_REQ:
                fdrec[w_fd_count].fd = sess_chunk[i]->cfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_WAIT_ADDR:
                fdrec[w_fd_count].fd = sess_chunk[i]->rfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_WAIT_CONN_SERV:
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLOUT | POLLWRNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_RECV_RES:
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_RELAY_DIRECT:
                fdrec[w_fd_count].fd = sess_chunk[i]->cfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
#if 0
            case STATE_THINK:
                fdrec[w_fd_count].fd = sess_chunk[i]->cfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
#endif
            default:
#ifdef TRACE_SCHEDULE
                Trace("Slave: thread %x, sess# %d, mode %c, state strange %s\n",
                    pthread_self(),
                    sess_chunk[i]->sid,
                    sess_chunk[i]->mode==MODE_CLIENT ? 'c': 's',
                    MSG_STATE_[sess_chunk[i]->state]);
#endif
                break;
            }
        }

#ifdef TRACE_SCHEDULE
        Trace("th %d, w_sess_count/max %d/%d w_fd_count %d\n",
			pthread_self(), w_sess_count, w_sess_max, w_fd_count);

		/*
        Trace("\nbefore ***\n");
        CHECK;
		*/
#endif


        if(w_fd_count<1) {
/*
Trace("skip tp dispatch by sess\n");
*/
            goto dispatch_by_sess;
        }
/*
Trace("call poll()\n");
*/

		c = 0;
try_poll:
		c++;
        chk = poll(fdrec, w_fd_count, 1000); /* 1000msec = 1sec */

#if 0
	/* this debug routine support one thread */
 {
Trace("result of poll() %d/%d, (errno %d)\n", chk, w_fd_count, errno);
	if(chk>0) {
		if(nhot == w_fd_count + think_queue->count + wait_queue->count) {
			Trace("maybe correct %d vs %d (%d+%d+%d) 2\n",
				nhot, w_fd_count+think_queue->count+wait_queue->count,
				w_fd_count, think_queue->count, wait_queue->count);
		}
		else {
			Trace("maybe *NOT* correct %d vs %d (%d+%d+%d) 2\n",
				nhot, w_fd_count+think_queue->count+wait_queue->count,
				w_fd_count, think_queue->count, wait_queue->count);
			if(nhot -
				(w_fd_count + think_queue->count + wait_queue->count)> 16) {
				Trace("\b\a\b\a\n");

				miss_capability--;
				if(miss_capability<=0) {
					exit(73);
				}
			}
		}
	}
 }
#endif

#ifdef TRACE_SCHEDULE
        if(timeout_count) {
            Trace("Slave: <%08x> query/receive/timeout %4d/%4d/%4d\n",
                pthread_self(), w_fd_count, chk, timeout_count);
        }
#endif

   		if(chk<=0) {
   		  if(chk<0) {
            if(errno==EINTR && c<10) {
                goto try_poll;
            }
            Error("fail poll() %d (%d)", c, errno);
			if(errno==ENOMEM) {
				struct timespec tp;
				tp.tv_sec = 0;
				tp.tv_nsec = 50 * 1000000;	/* 50 msec */
				nanosleep(&tp, NULL);
			}
		  }

#if 0
            for(i=0;i<w_sess_count;i++) {
                if(sess_chunk[i]) {
                    chk = push_sess(think_queue, sess_chunk[i]);

					if(chk) {
							sess_chunk[i]->reason = R_MEMORY_OVERFLOW;
							DoSendMessage(sess_chunk[i]);
							DoCloseSess(sess_chunk[i]);
							DoReleaseSess(sess_chunk[i]);
							sess_chunk[i] = NULL;
					}

                }
            }

            goto next;
#endif
			for(i=0;i<w_fd_count;i++) {
				if(bound[i]) {
					chk = push_sess(think_queue, bound[i]);
					if(chk) {
						bound[i]->reason = R_MEMORY_OVERFLOW;

						DoSendMessage(bound[i]);
						DoCloseSess(bound[i]);
						release_buf(bound[i]);
						DoReleaseSess(bound[i]);
					}
					for(j=0;j<w_sess_count;j++) {
						if(sess_chunk[j]==bound[i]) {
							sess_chunk[j] = NULL;
						}
					}
					bound[i] = NULL;
				}
			}

			goto dispatch_by_sess;
        }


dispatch_by_fd:
        for(i=0;i<w_fd_count;i++) {
            touched = 0;

            /* count other fds is bounded to same session */
            same = 0;
            for(j=i+1;j<w_fd_count;j++) {
                if(bound[j]==bound[i]) {
                    same++;
                }
            }

            if(bound[i]==NULL) {
/*
                goto next;
*/
				continue;
            }

            /* server connect when state WAIT_CONN_SERV
             * otherwise, receive anything from server or shutdown arrival
             */
            if(
                (bound[i]->state==STATE_WAIT_CONN_SERV
                    && fdrec[i].revents & POLLOUT)
                  ||(bound[i]->state!=STATE_WAIT_CONN_SERV
                    && fdrec[i].revents & (POLLHUP | POLLIN))
                ) {
#ifdef TRACE_SCHEDULE
                Trace("  call handler<%p> %d: fd# %d, sess %p, sess# %d, %s\n",
                    pthread_self(), i, fdrec[i].fd,
                    bound[i], bound[i]->sid,
                    MSG_STATE_[bound[i]->state]);
#endif
                Handler(bound[i], now);
                touched = 1;
            }
            else {
                if(same==0) {
                    /* nobody is bounded to same session */
                    chk = push_sess(think_queue, bound[i]);
					if(chk) {
							bound[i]->reason = R_MEMORY_OVERFLOW;
							DoSendMessage(bound[i]);
							DoCloseSess(bound[i]);
							DoReleaseSess(bound[i]);
							bound[i] = NULL;
					}
                }
            }

            /* clear a processed session */
            for(j=0;j<w_sess_count;j++) {
                if(bound[i]==sess_chunk[j]) {
                    sess_chunk[j] = NULL;
                }
            }
            if(touched && same) {
                for(j=i+1;j<w_fd_count;j++) {
                    if(bound[j]==bound[i]) {
                        bound[j] = NULL;
                    }
                }
            }
            bound[i] = NULL;
        }


dispatch_by_sess:


        for(i=0;i<w_sess_count;i++) {
            if(sess_chunk[i]) {
#ifdef TRACE_SCHEDULE
                Trace("  call handler<%p> %d: sess %p, sess# %d, %s\n",
                    pthread_self(), i,
                    sess_chunk[i], sess_chunk[i]->sid,
                    MSG_STATE_[sess_chunk[i]->state]);
#endif
                Handler(sess_chunk[i], now);

                for(j=0;j<w_sess_count;j++) {
                    if(i!=j && sess_chunk[i]==sess_chunk[j]) {
                        sess_chunk[j] = NULL;
                    }
                }
                sess_chunk[i]=NULL;
            }
        }

#if 0
 for(i=0;i<w_sess_count;i++) {
	if(sess_chunk[i]) {
		Trace(" rest %3d: sess# %d,\n", i, sess_chunk[i]);
	}
 }
#endif


next:
        (void)0; /* dummy */


#if 0
        skcache_dir_sweeper();
#endif

    } /* while */

}


void*
Reporter(void *dummy)
{
    int cfd;
    int w;
    int p;

    while(1) {
        AcceptReport_fp(stdout);
        ListSession(stdout);
#if 0
        sim_stat_show_lock(&len_perread_stat, stderr);  /* XXX */
#endif
#if 1
        sim_stat_show_lock(&object_size_stat, stderr);  /* XXX */
        cache_list_entry(stderr); /* XXX */
#endif
        sleep(900);
    }
}


extern volatile int disk_sweep_interval;

void*
DSweeper(void* dummy)
{
	int chk;
    struct timespec join_time;

	/* msec, how many times in 1000msec */
	int interval = disk_sweep_interval;	

	while(1) {
		skcache_dir_sweeper();

    	join_time.tv_sec    = 0;
    	join_time.tv_nsec   = interval * 1000000; 
		chk = nanosleep(&join_time, NULL);

		/*
		Trace("DSweeper: chk %d\n", chk);
		*/
	}
}



/*
 * target buckets are serialize by queue
 */

void*
MSweeper(void* dummy)
{
	sess_t *vp;
	time_t now;
	int c;
	int chk;
	bucket_t *bk;
    int w_sess_max;
    int w_sess_count;
    sess_t *sess_chunk[SESS_CHUNK_SIZE];
	int i, j;
	int ac;
	int tk;
	int w;

    w_sess_max = require_chunk_size;

	c = 0;
	while(1) {
#if 0
		MUTEX_LOCK(&stat_lock);
		tk = swapout_token;
		MUTEX_UNLOCK(&stat_lock);
		if(tk<=0) {
#if 1
			Trace("MSweeper: no token, wait a second\n");
#endif
			sleep(1);
			continue;
		}
#endif

		w = 0;

#if 0
		vp = pull_sess(msweep_queue);
		if(vp==NULL) {
			goto next;
		}
#endif

        w_sess_count = sq_pullchunk(msweep_queue, sess_chunk, w_sess_max);
        if(w_sess_count<=0) {
#if 1
			Trace("MSweeper: no request\n");
#endif
            goto next;
        }

		ac = 0;
		for(i=0;i<w_sess_count;i++) {
			if(sess_chunk[i]) {
				ac++;
				for(j=i+1;j<w_sess_count;j++) {
					if(sess_chunk[j] &&
					  sess_chunk[j]->fhash == sess_chunk[i]->fhash
					  && sess_chunk[j]->mode == sess_chunk[i]->mode) {
						sq_be_free(sess_chunk[j]);
						sess_chunk[j] = NULL;
					}
				}
			}
		}

#if 1
Trace("MSweeper: th %08x actual %4d got %4d/%-5d compaction requests\n",
	pthread_self(), ac, w_sess_count, w_sess_max);
#endif

#if 0
		for(i=0;i<w_sess_count;i++) {
			vp = sess_chunk[i];
			if(vp==NULL) {
				Trace("r %3d NULL\n", i);
			}
			else {
				Trace("r %3d bk# %3d %p %s\n",
					i, vp->fhash, vp,
					vp->mode==XSK_ENTRY ? "ENTRY" : "FRAG ");
			}
		}
#endif

		for(i=0;i<w_sess_count;i++) {
			vp = sess_chunk[i];
			if(vp==NULL) {
				continue;
			}
			if(vp->fhash<0 || vp->fhash >= RADIX_H1_MAX) {
				continue;
			}

			now = time(NULL);
			c++;
			bk = &skcache_bucket_base[vp->fhash];

#if 1
	Trace("MSweep: try lock bk# %d, now %d\n", vp->fhash, now);
#endif

#if 1
			LOCK_BK(bk);
#endif

#if 0
			chk = MUTEX_TRYLOCK(&bk->mutex);
			if(chk) {
				if(chk==EBUSY) {
#if 1
Trace("bk# %-3d %p is busy\n", vp->fhash, bk);
#endif
					chk = push_sess(msweep_queue, vp);
					if(chk) {
						sq_be_free(vp);
						vp = NULL;
					}
					w++;

					bk = NULL;
					goto skip;
				}
				Error("MSweep: cannot lock mutex of bk# %3d, %p, %d (%d)",
					bk, vp->fhash, chk, errno);
				bk = NULL;
				goto done;
			}
#endif
			bk->own = 'S';

#if 1
	Trace("MSweep: done lock bk# %d, now %d\n", vp->fhash, now);
#endif

 assert(check_bkentry(bk, "sw0"));


			/* Until swapout, compaction requests have been occured
			 * Then, after swapout this routine may receive requests.
			 * So, this routine has to check duplicated requests.
			 */

			chk = 0;
			if(vp->mode==XSK_ENTRY && bk->entry_use > bk->entry_hwater) {
				chk = 1;
			}
			if(vp->mode==XSK_FRAG && bk->frag_use > bk->frag_hwater) {
				chk = 1;
			}

#if 0
			if(chk) {
				if(now - bk->last_gc < 10) {
					Trace("*** the bucket GCed before %d sec ***\n",
						now - bk->last_gc);
					chk = 2;
				}
			}
#endif

 assert(check_bkentry(bk, "sw1"));

#if 1
Trace("MSweeper:- %6d:%3d, %4d %4d bk# %-3d %p, entry %4d/%d/%d frag %4d/%d/%d %s %s\n",
			c, i, now - vp->last_action, now - bk->last_gc,
			vp->fhash, bk,
			bk->entry_use, bk->entry_hwater, bk->entry_all,
			bk->frag_use, bk->frag_hwater, bk->frag_all,
			vp->mode==XSK_ENTRY ? "ENTRY" : "FRAG ",
			chk==2 ? "YESYES" : (chk==1 ? "YES":"NO"));
#endif
 assert(check_bkentry(bk, "sw2"));
			if(chk==1) {

				switch(vp->mode) {
				case XSK_ENTRY:	/* entry */
					if(bk->elem_freelist==NULL
							|| bk->entry_use >= bk->entry_hhwater)
					{
				chk = skcache_compaction(bk, -67, 1, XSK_ENTRY);
 assert(check_bkentry(bk, "sw4.3"));
					}
					else {
				chk = skcache_compaction(bk, -42, 0, XSK_ENTRY);
 assert(check_bkentry(bk, "sw4.7"));
						if(bk->elem_freelist==NULL
							|| bk->entry_use >= bk->entry_hwater)
						{
				chk = skcache_compaction(bk, -77, 1, XSK_ENTRY);
 assert(check_bkentry(bk, "sw4.9"));
						}
					}

					break;

				case XSK_FRAG:	/* frag */
				chk = skcache_compaction(bk, -52, 0, XSK_FRAG);
 assert(check_bkentry(bk, "sw5"));
					if(bk->frag_freelist==NULL
							|| bk->frag_use >= bk->frag_hwater) {
						/* to achive large effect,
						 * we have to compaction with
					 	 * entry.  */
			chk = skcache_compaction(bk, -596, 1, XSK_ENTRY);
 assert(check_bkentry(bk, "sw6"));
					}

					break;
				}
			}
			else
			if(chk==2) {
			chk = skcache_compaction(bk, -767, 1, XSK_ENTRY);
 assert(check_bkentry(bk, "sw7"));
			}
			else {
				bk->gc_push = 0;
			}
			


done:
			sq_be_free(vp);
skip:
 assert(check_bkentry(bk, "sw8"));
			if(bk) {
				UNLOCK_BK(bk);
				bk->own = ' ';
				bk = NULL;
			}
		}

next:

		(void)0;
		if(w>0) {
			Trace("sleep one second\n");
			sleep(1);
		}
	}
}









extern sim_stat_rec hres_stat;


extern int skcache_mem_keep_min;
extern int skcache_mem_keep_cur;
extern int skcache_mem_keep_max;
extern int skcache_entry_keep_min;
extern int skcache_entry_keep_cur;
extern int skcache_entry_keep_max;


extern long long swap_count;
extern long long swap_size;

void*
Monitor(void *dummy)
{

    char *fmt =
": time %d %d %s %c\
 arr %.1f/%.1f %ld %d, pro %.1f/%.1f %ld %d, prog %.2f\
 err %.1f/%.1f %ld %d :\
 hit %d, miss %d, hit%% rate %.2f ratio %.2f,\
 sess %d, queue %d %d %d, assign_buf fail %d success %d :\
 keep %d %d,\
 swapload %.1f/%.1f %ld %d success %.1f/%.1f %ld %d,\
 swapin %.1f/%.1f %ld %d success %.1f/%.1f %ld %d,\
 swapout %.1f/%.1f %ld %d token %d/%d success %.1f/%.1f %ld %d,\
 comp %.1f/%.1f %ld %d,\
 sweep %.1f/%.1f %ld %d, capacity %lld %lld"
 ;
    char outline[STRING_SIZE];
    char elap_str[STRING_SIZE];
    struct timeval last_print_t;

    int count;
    int last_keep_depth;
    int keep_check_interval;

    /* time */
    struct timeval last_t, t0, t1, t2, t3, t4, t5, t6;
    int delta_t;
    int t_nhot;
    char mark;

    /* arrival */
    int c_arr, last_arr, delta_arr;
    double r_arr, rmax_arr;

    /* process */
    int c_pro, last_pro, delta_pro;
    double r_pro, rmax_pro;

    int print_count;
    time_t elap;

    int c_rjq;

    /* error */
    int c_err, last_err, delta_err;
    double r_err, rmax_err;

    unsigned int sig, last_sig;

    /* hit/miss */
    int c_hit, last_hit, delta_hit;
    int c_miss, last_miss, delta_miss;
    double hrate, hratio;

    int  c_think, c_send, c_fetch, c_prompt;
    int  s_think, s_send, s_fetch, s_prompt;
	int  c_msweep;
	int  c_wait;

    /* compaction */
    int c_comp, last_comp, delta_comp;
    double r_comp, rmax_comp;

    /* expiration */
    int c_expire, last_expire, delta_expire;
    int c_expire_normal_idle_th, c_expire_cracked_idle_th;
#if 0
    int h_expire[COMP_HIST_WIDTH];
#endif
/*
    int r_expire;
*/
    double r_expire;

    double expire_k;
    int expire_raise_tick;
    int expire_min;

    /* swapin/out */
    int c_swapload, last_swapload, delta_swapload;
    double r_swapload, rmax_swapload;
    int c_swaploadsuc, last_swaploadsuc, delta_swaploadsuc;
    double r_swaploadsuc, rmax_swaploadsuc;

    int c_swapin, last_swapin, delta_swapin;
    double r_swapin, rmax_swapin;
    int c_swapinsuc, last_swapinsuc, delta_swapinsuc;
    double r_swapinsuc, rmax_swapinsuc;

    int c_swapout, last_swapout, delta_swapout;
    double r_swapout, rmax_swapout;
    int c_swapoutsuc, last_swapoutsuc, delta_swapoutsuc;
    double r_swapoutsuc, rmax_swapoutsuc;

    /* sweep */
    int c_sweep, last_sweep, delta_sweep;
    double r_sweep, rmax_sweep;

	long long dir_size_sum;


    count = 0;
    print_count = 0;
    last_keep_depth = 0;
    keep_check_interval = 4;

    last_arr  = 0;
    rmax_arr  = -1.0;
    last_pro  = 0;
    rmax_pro  = -1.0;

    last_err  = 0;
    rmax_err  = -1.0;

    last_hit  = 0;
    last_miss = 0;

    last_swapload       = 0;
    rmax_swapload       = -1.0;
    last_swaploadsuc    = 0;
    rmax_swaploadsuc    = -1.0;
    last_swapin         = 0;
    rmax_swapin         = -1.0;
    last_swapinsuc      = 0;
    rmax_swapinsuc      = -1.0;
    last_swapout        = 0;
    rmax_swapout        = -1.0;
    last_swapoutsuc     = 0;
    rmax_swapoutsuc     = -1.0;

    last_sig = 0xf0f0;

    last_comp   = 0;

    /* EX TH */
    expire_k = 2.0;
    expire_raise_tick = 60;
    expire_min = 60;

    last_expire = 0;

    gettimeofday(&last_t, NULL);
    while(1) {
        /* interval */
        sleep(MONITOR_INTERVAL);


        gettimeofday(&t0, NULL);
        MUTEX_LOCK(&stat_lock);
        elap    = t0.tv_sec - boot_date;
        c_arr   = receive_request_count;
        c_pro   = process_request_count;
        c_err   = error_reply_count;
        c_rjq   = reject_request_count;
        c_comp  = compaction_count;
        c_expire  = expire_count;

        c_hit  = hit_reply_count;
        c_miss = miss_reply_count;

        c_swapload  = swapload_count;
        c_swapin    = swapin_count;
        c_swapout   = swapout_count;
        c_swaploadsuc   = swapload_success;
        c_swapinsuc     = swapin_success;
        c_swapoutsuc    = swapout_success;

        c_sweep     = sweeper_count;

        MUTEX_UNLOCK(&stat_lock);


        MUTEX_LOCK(&sess_lock);
        t_nhot = nhot;
        MUTEX_UNLOCK(&sess_lock);

        MUTEX_LOCK(&think_queue->lock);
        c_think = think_queue->count;
        s_think = think_queue->sig;
        MUTEX_UNLOCK(&think_queue->lock);

        MUTEX_LOCK(&wait_queue->lock);
        c_wait = wait_queue->count;
        MUTEX_UNLOCK(&wait_queue->lock);

        MUTEX_LOCK(&think_queue->lock);
        c_msweep = msweep_queue->count;
        MUTEX_UNLOCK(&think_queue->lock);

#if 0
        MUTEX_LOCK(&send_queue->lock);
        c_send = send_queue->count;
        s_send = send_queue->sig;
        MUTEX_UNLOCK(&send_queue->lock);

        MUTEX_LOCK(&fetch_queue->lock);
        c_fetch = fetch_queue->count;
        s_fetch = fetch_queue->sig;
        MUTEX_UNLOCK(&fetch_queue->lock);

        MUTEX_LOCK(&prompt_queue->lock);
        c_prompt = prompt_queue->count;
        s_prompt = prompt_queue->sig;
        MUTEX_UNLOCK(&prompt_queue->lock);
#else
        c_send   = s_send   = 0;
        c_fetch  = s_fetch  = 0;
        c_prompt = s_prompt = 0;
#endif

        gettimeofday(&t6, NULL);

        sig = c_arr | c_pro | t_nhot | s_think | s_send | s_fetch | s_prompt;

#if 0
        if(last_t==0) {
            delta_t = 1;
        }
        else {
            delta_t = t0 - last_t;
        }
#endif
        delta_t = KOTETU_SUB_TIME(t0,last_t);

        delta_arr = c_arr - last_arr;
        r_arr = ((double)delta_arr*1000)/delta_t;
        if(r_arr>rmax_arr) {
            rmax_arr = r_arr;
        }

        delta_pro = c_pro - last_pro;
        r_pro = ((double)delta_pro*1000)/delta_t;
        if(r_pro>rmax_pro) {
            rmax_pro = r_pro;
        }

        delta_err = c_err - last_err;
        r_err = ((double)delta_err*1000)/delta_t;
        if(r_err>rmax_err) {
            rmax_err = r_err;
        }


        delta_hit = c_hit - last_hit;
        delta_miss = c_miss - last_miss;

        hratio = (hit_reply_count+miss_reply_count>0 ?
                    (100.0*hit_reply_count) /(hit_reply_count+miss_reply_count)
                    : -1);
        hrate  = (delta_hit+delta_miss>0 ?
                    (100.0*delta_hit)/(delta_hit+delta_miss)
                    : -1);

        delta_comp   = c_comp - last_comp;
        r_comp = ((double)delta_comp*1000)/delta_t;
        if(r_comp>rmax_comp) {
            rmax_comp = r_comp;
        }

        delta_expire = c_expire - last_expire;

        delta_swapload = c_swapload - last_swapload;
        r_swapload = ((double)delta_swapload*1000)/delta_t;
        if(r_swapload>rmax_swapload) {
            rmax_swapload = r_swapload;
        }
        delta_swaploadsuc = c_swaploadsuc - last_swaploadsuc;
        r_swaploadsuc = ((double)delta_swaploadsuc*1000)/delta_t;
        if(r_swaploadsuc>rmax_swaploadsuc) {
            rmax_swaploadsuc = r_swaploadsuc;
        }

        delta_swapin = c_swapin - last_swapin;
        r_swapin = ((double)delta_swapin*1000)/delta_t;
        if(r_swapin>rmax_swapin) {
            rmax_swapin = r_swapin;
        }
        delta_swapinsuc = c_swapinsuc - last_swapinsuc;
        r_swapinsuc = ((double)delta_swapinsuc*1000)/delta_t;
        if(r_swapinsuc>rmax_swapinsuc) {
            rmax_swapinsuc = r_swapinsuc;
        }

        delta_swapout = c_swapout - last_swapout;
        r_swapout = ((double)delta_swapout*1000)/delta_t;
        if(r_swapout>rmax_swapout) {
            rmax_swapout = r_swapout;
        }
        delta_swapoutsuc = c_swapoutsuc - last_swapoutsuc;
        r_swapoutsuc = ((double)delta_swapoutsuc*1000)/delta_t;
        if(r_swapoutsuc>rmax_swapoutsuc) {
            rmax_swapoutsuc = r_swapoutsuc;
        }

        delta_sweep = c_sweep - last_sweep;
        r_sweep = ((double)delta_sweep*1000)/delta_t;
        if(r_sweep>rmax_sweep) {
            rmax_sweep = r_sweep;
        }


        MUTEX_LOCK(&config_lock);
#if 0
        c_expire_normal_idle_th   = cache_normal_idle_th;
        c_expire_cracked_idle_th  = cache_cracked_idle_th;
#endif
#if 1
        c_expire_normal_idle_th   = skcache_mem_keep_cur;
        c_expire_cracked_idle_th  = -1;
#endif
        MUTEX_UNLOCK(&config_lock);

#if 0
        if(1)
#endif
        if(t0.tv_sec-last_print_t.tv_sec>=MONITOR_INTERVAL|| sig!=last_sig)
        {
            if(sig!=last_sig) {
                mark = ':';
            }
            else {
#if 0
s_show_freelist();
#endif
                mark = '*';
            }

            conv_sectoshortstr(elap_str, elap);

            MUTEX_LOCK(&stat_lock);
            sprintf(outline, fmt,
                t0.tv_sec,
                elap, elap_str,
                mark,

                r_arr, rmax_arr, (long)c_arr, delta_arr,
                r_pro, rmax_pro, (long)c_pro, delta_pro,
                c_arr>0 ?  ((double)c_pro*100.0)/c_arr : -100.0,
                r_err, rmax_err, (long)c_err, delta_err,

                hit_reply_count, miss_reply_count,
                hrate, hratio,

                t_nhot,
                c_think,
                c_wait,
				c_msweep,
                alloc_tmp_buf_count,
                alloc_sess_buf_count,

                skcache_mem_keep_cur, skcache_entry_keep_cur,
        r_swapload, rmax_swapload, (long)c_swapload, delta_swapload,
        r_swaploadsuc, rmax_swaploadsuc, (long)c_swaploadsuc, delta_swaploadsuc,
        r_swapin, rmax_swapin, (long)c_swapin,  delta_swapin,
        r_swapinsuc, rmax_swapinsuc,   (long)c_swapinsuc,   delta_swapinsuc,
        r_swapout, rmax_swapout, (long)c_swapout, delta_swapout,
		swapout_token, swapout_token_init,
        r_swapoutsuc, rmax_swapoutsuc,  (long)c_swapoutsuc, delta_swapoutsuc,
                r_comp, rmax_comp,  (long)c_comp,   delta_comp,
                r_sweep, rmax_sweep,(long)c_sweep,  delta_sweep,
                swap_count, swap_size
                );

#if 1
Trace("\
traffic: intra_incoming %12lld B %10.2f MB %7.2f GB\n\
         extra_outgoing %12lld B %10.2f MB %7.2f GB\n\
         extra_incoming %12lld B %10.2f MB %7.2f GB\n\
         intra_outgoing %12lld B %10.2f MB %7.2f GB\n",
  (long long)intra_incoming, fdM(intra_incoming), fdG(intra_incoming), 
  (long long)extra_outgoing, fdM(extra_outgoing), fdG(extra_outgoing),
  (long long)extra_incoming, fdM(extra_incoming), fdG(extra_incoming),
  (long long)intra_outgoing, fdM(intra_outgoing), fdG(intra_outgoing)
        );
Trace("swap: count %lld (%.2f M), size %lld B %8.2f MB, %5.2f GB\n",
		swap_count, fdM(swap_count),
		swap_size, fdM(swap_size), fdG(swap_size) );
	
 {
	int i;
	int u,d;
	long long guess_size;
	extern int *dir_sz;
	extern int dir_num;

	u = d = 0;
	dir_size_sum = 0LL;
	for(i=0;i<dir_num;i++) {
		/* up */
		if(dir_sz[i]>=(off_t)0L) {
			u++;
			dir_size_sum += dir_sz[i];
		}
		else {
			d++;
		}
	}

	guess_size = ((long long)dir_size_sum*dir_num)/u;

Trace(
"  counted : dir_num %d, up %d, down %d, size %lld B, %8.2f MB, %5.2f GB\n",
		dir_num, u, d, dir_size_sum, fdM(dir_size_sum), fdG(dir_size_sum) );
Trace(
"  guessed by average : %lld B, %8.2f MB, %8.2f GB\n",
		guess_size, fdM(guess_size), fdG(guess_size));
 }
	

#endif

            MUTEX_UNLOCK(&stat_lock);

            LogLine(outline);
            LogSync();

#ifndef LOG_REPORT_CONSOLE
            Trace("%s\n", outline);
#endif


#if 0
            MUTEX_LOCK(&stat_lock);
            sprintf(outline, ": swap in %d (success %d + %d) out %d (success %d + %d)",
                swapin_count, swapin_success, swapin_count-swapin_success,
                swapout_count, swapout_success, swapout_count-swapout_success);
            MUTEX_UNLOCK(&stat_lock);

            LogLine(outline);
            LogSync();

#ifndef LOG_REPORT_CONSOLE
            Trace("%s\n", outline);
#endif

#endif


#ifdef TRACE_BLOCKS
            check_block();
#endif

            last_sig = sig;
#if 0
            last_print_t = t0;
#endif
            memcpy(&last_print_t, &t0, sizeof(struct timeval));
            print_count++;
        }

        {
            sim_hist_add_value(&comp_hist, (long)delta_comp);
            sim_hist_add_value(&expire_hist, (long)delta_expire);

#if 0
            sim_hist_show(&comp_hist, stdout);
            sim_hist_show(&expire_hist, stdout);
#endif

#if 0
            r_comp = sim_hist_average_d(&comp_hist);
            r_expire = sim_hist_average_d(&expire_hist);
            Trace("  $$ rate comp %.2f, expire %.2f\n", r_comp, r_expire);
#endif
        }

#if 1
 if(print_count%100==0) {

    extern sim_stat_rec swapload_time;
    extern sim_stat_rec swapin_time;
    extern sim_stat_rec swapout_time;
    extern sim_stat_rec checkdir_time;
	extern sim_stat_rec compaction_time;

	/* XXX */
    sim_stat_show_lock(&swapload_time, stderr);  
    sim_stat_show_lock(&swapin_time, stderr); 
    sim_stat_show_lock(&swapout_time, stderr);
    sim_stat_show_lock(&checkdir_time, stderr);
    sim_stat_show_lock(&compaction_time, stderr);
 }
#endif

        /* store last values */
#if 0
        last_t = t0;
#endif
        memcpy(&last_t, &t0, sizeof(struct timeval));
        last_arr    = c_arr;
        last_pro    = c_pro;
        last_err    = c_err;

        last_expire = c_expire;
        last_comp   = c_comp;

        last_hit    = c_hit;
        last_miss   = c_miss;

        last_swapload   = c_swapload;
        last_swapin     = c_swapin;
        last_swapout    = c_swapout;
        last_swaploadsuc    = c_swaploadsuc;
        last_swapinsuc      = c_swapinsuc;
        last_swapoutsuc     = c_swapoutsuc;

        last_sweep      = c_sweep;

/*
        LogSync();
*/
		check_busy_buckets();

        count++;
    }

}




static
int
check_mutex(pthread_mutex_t *mx)
{
    int chk;

    chk = MUTEX_TRYLOCK(mx);
    if(chk) {
        if(chk==EBUSY)
            return 0;
        else
            return -chk;
    }
    MUTEX_UNLOCK(mx);
    return 1;
}


static char *_msg_idle  = "IDLE ";
static char *_msg_busy  = "BUSY ";
static char *_msg_err   = "ERROR";

static
int
report_check_mutex(FILE *fp, char *msg, pthread_mutex_t *mx)
{
    int chk;
    char *rep;

    chk = check_mutex(mx);
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    fprintf(fp, "$$$ %s %s\n", msg, rep);

    return 0;
}

void print_sess(FILE *ofp, void* raw_addr, struct timeval *reftime)
{
    sess_t *self;
    int chk, err;
    char *rep;
    struct timeval reftime_local;
    int elap;

    if(reftime==NULL) {
        gettimeofday(&reftime_local, NULL);
        reftime = &reftime_local;
    }

    self = (sess_t*) raw_addr;

    chk = check_mutex(&(self->busy));
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    elap = reftime->tv_sec - self->accept_time.tv_sec;

    fprintf(ofp, "%3d %s %c %3d %3d <%s>\n",
        self->sid, rep, self->mode==MODE_CLIENT ? 'c': '-', elap, self->fhash, self->loc);
#if 0
    fprintf(ofp, "%3d ", self->reqlen);
    fprintf(ofp, "<%s> ", self->loc);
    fprintf(ofp, "\n");
#endif
}

int
check_buckets()
{
	int i;
	bucket_t *bk;
	int f;

	for(i=0;i<radix_h1;i++) {
    	bk = &skcache_bucket_base[i];
		f = check_mutex(&bk->mutex);	
		Trace("bk# %4d: %d\n", i, f);
	}

	return 1;
}

int
check_busy_buckets()
{
	int i;
	bucket_t *bk;
	int f;
	int c;
	char band[RADIX_H1_MAX];

	c = 0;
	for(i=0;i<radix_h1;i++) {
    	bk = &skcache_bucket_base[i];
		f = check_mutex(&bk->mutex);	
		if(f==0) {
			if(c==0) {
				Trace("busy buckets:\n");
			}
			Trace("busy bk# %-3d: %d %d\n", i, f, bk->last_gc);
			c++;
			if(bk->own=='S') {
				band[i] = bk->own;
			}
			else {
				band[i] = '*';
			}
		}
		else
		if(f<0) {
			band[i] = '?';
		}
		else {
			band[i] = '.';
		}
	}
	band[radix_h1] = '\0';
	if(c>0) {
		Trace("busy %d threads\n", c);
#if 0
	Trace("busy band %d %3d %s\n", time(NULL), c, band);
#endif
	}
#if 1
	Trace("busy band %d %3d %s\n", time(NULL), c, band);
#endif

	return 1;
}



static pthread_mutex_t logging_lock = PTHREAD_MUTEX_INITIALIZER;

RETSIGTYPE
Logging_Queues()
{
    FILE *fp;
    int i;
    struct timeval now;
    int t_nhot;

    MUTEX_LOCK(&logging_lock);

    gettimeofday(&now, NULL);

    Trace("logggin queues\n");

    fp = fopen("queues.log", "w");
    if(fp==NULL) {
        goto out;
    }

    fprintf(fp, "process %d, thread %x\n\n", getpid(), pthread_self());

        fprintf(fp, " monitor           %x\n", th_monitor);
        fprintf(fp, " listener          %x\n", th_listener);
    fprintf(fp, " manager  %d/%d\n", require_n_handler, N_HANDLER_MAX);
    for(i=0;i<require_n_handler;i++) {
        fprintf(fp, "   %3d             %x\n", i, th_handlers[i]);
    }
    fprintf(fp, "\n");

    MUTEX_LOCK(&sess_lock);
    t_nhot = nhot;
    MUTEX_UNLOCK(&sess_lock);

    fprintf(fp, "hot session %d\n\n", t_nhot);

    fprintf(fp, "== think queue\n");
    sq_show_mapfunc(fp, think_queue,    print_sess, &now);
#if 0
    fprintf(fp, "== send queue\n");
    sq_show_mapfunc(fp, send_queue,     print_sess, &now);
    fprintf(fp, "== fetch queue\n");
    sq_show_mapfunc(fp, fetch_queue,    print_sess, &now);
    fprintf(fp, "== prompt queue\n");
    sq_show_mapfunc(fp, prompt_queue,   print_sess, &now);
#endif


    report_check_mutex(fp, "mutex stat   ", &stat_lock);
    report_check_mutex(fp, "mutex sess   ", &sess_lock);
    report_check_mutex(fp, "mutex config ", &config_lock);
    report_check_mutex(fp, "mutex logging", &logging_lock);


done:

    fflush(fp);
    fclose(fp);

out:

    MUTEX_UNLOCK(&logging_lock);
}


/*
 *
 */
RETSIGTYPE
TERM_Trap()
{
    Log("; Catch signal for termination");
    MUTEX_LOCK(&stat_lock);
    catch_termination = 1;
    MUTEX_UNLOCK(&stat_lock);
    exit(127);
}

void*
signal_watcher(void *arg)
{
    int sig;
    int chk;

    while(1) {
        chk = sigwait(&handler_sig_mask, &sig);
        if(chk) {
            Trace("signal_watcher: chk %d\n", chk);
            continue;
        }

        Trace("signal_watcher: get signal %d\n", sig);

        switch(sig) {
        case SIGPIPE:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGUSR2:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGHUP:
            LogSync();
            break;
        case SIGINT:
        case SIGTERM:
            MUTEX_LOCK(&stat_lock);
            catch_termination = 1;
            MUTEX_UNLOCK(&stat_lock);
            Log("END");
            LogSync();
            exit(137);
            break;
        case SIGCHLD:
          {
            int stat;
            pid_t pid;

            do {
try_waitpid:
                pid = waitpid((pid_t)-1, &stat, WNOHANG | WUNTRACED);

                if(pid < 0) {
                    if(errno==EINTR) {
                        goto try_waitpid;
                    }
                    printf("error with errno=%d\n", errno);
                    break;
                }
                if (pid > 0) {
                    printf("child %d terminated\n", pid);
                }
            } while(pid>0);

          }
            break;
        default:
            break;
        }
    }

 /*NOTREACHED*/
    return (void*)NULL;
}



/*
 *
 * if fdcap is -1, this routine will set maximum value from getrlimit()
 */
void
ToBeMaximumFDCapacity(int fdcap)
{
    struct rlimit nfile;
    int last_nofile;

#ifdef RLIMIT_NOFILE
#define X   RLIMIT_NOFILE
#else
#define X   RLIMIT_OFILE
#endif

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for current check was fail (%d)", errno);
    }
    else {
        last_nofile = nfile.rlim_cur;
    }

    if(fdcap<0) {
        nfile.rlim_cur = nfile.rlim_max;
    }
    else {
        nfile.rlim_cur = fdcap;
    }

    if(setrlimit(X, &nfile)<0) {
        Error("settrlimit() fail (%d)", errno);
    }

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for verify was fail (%d)", errno);
    }
    else {
        Trace("ToBeMaximumFDCapacity: current to %d (max %d), last %d, specified %d\n",
            nfile.rlim_cur, nfile.rlim_max,
            last_nofile, fdcap);
    }

#undef X

}


extern volatile off_t dir_size_max;


static
char *compile_magic = "This program was compiled at " __DATE__ ", " __TIME__;
int listen_fd;

int
main(int argc, char **argv)
{
    int i;
    int chk;
    int ll;

    pthread_attr_t th_attr_high, th_attr_mid, th_attr_low;
    pthread_attr_t *p_th_attr_high, *p_th_attr_mid, *p_th_attr_low;
    struct sched_param th_param;
    int th_policy_type;

    /***
     *** Initialize
     ***/

    MUTEX_LOCK(&stat_lock);
    boot_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);

#if 0
#if 1
    TraceOn();
    {
        extern char log_file[];
        strcpy(log_file, "kotetu.log");
        Trace("*** log_file '%s'\n", log_file);
    }
#else
    TraceOff();
#endif
    InitLogger("kotetuserv");
    Log("START");
    LogSync();
#endif

    SetSignalHandler(SIGPIPE, SIG_IGN);
#if 0
    SetSignalHandler(SIGABRT, SIG_DFL);
    SetSignalHandler(SIGSEGV, SIG_IGN);
#endif

#if 0
    sigemptyset(&handler_sig_mask);
    sigaddset(&handler_sig_mask, SIGTERM);
#if 0
    sigaddset(&handler_sig_mask, SIGPIPE);
#endif
    pthread_sigmask(SIG_UNBLOCK, &handler_sig_mask, NULL);
#endif


    /* Configuration */

    SetGlobalDefaults();
    if(getenv(DEFAULT_CONFIG_ENVNAME)!=NULL) {
        strcpy(config_filename, getenv(DEFAULT_CONFIG_ENVNAME));
    }
    if(config_filename[0]) {
        ReadConfigFile(config_filename);
    }
    ReadCommandlineOptions(argc, argv);

    FixGlobals();


    InitLogger("kotetuserv");
    Log("START");
    LogSync();

    if(trace_flag) {
        ShowGlobals(stdout);
    }
    if(trace_flag) {
        TraceOn();
    }
    else {
        TraceOff();
    }

    date_zone_offset_init();

#ifndef NDEBUG
	Trace("not set NDEBUG, so it is debug mode\n\
If you want an actual performance, you should re-compile with -DNDEBUG\n");
#endif
#if 1
	if(getenv("PWD")) {
		Trace("PWD is %s\n", getenv("PWD"));
	}
#endif

#if 1
    Trace("parameters:\n\
    L_URL       %6d\n\
    L_HOST      %6d\n\
    L_PATH      %6d\n\
    L_BUF       %6d\n\
    L_REQBUF    %6d\n", L_URL, L_HOST, L_PATH, L_BUF, L_REQBUF);

    Trace("size of structures:\n\
    skiplist cell   %6d\n\
    skiplist        %6d\n\
    entry           %6d\n\
    frag            %6d\n\
    bucket          %6d\n",
        sizeof(skiplist_cell_t), sizeof(skiplist_t),
        sizeof(skcache_entry_t), sizeof(skcache_frag_t),
        sizeof(bucket_t));

#endif

    ToBeMaximumFDCapacity(-1);

    /* sessions and queues */
#if 0
    InitSession(think_queue_length);
#endif
    think_queue_length = require_sess_length;
    InitSession(require_sess_length+require_sess_length/2+2);

    sq_init(&X_think_queue, think_queue_length);
    think_queue = &X_think_queue;
    sq_init(&X_wait_queue, think_queue_length/4);
    wait_queue = &X_wait_queue;
    sq_init(&X_msweep_queue, think_queue_length/4);
    msweep_queue = &X_msweep_queue;

#if 1
	Trace("queues:\n\
	think			%p\n\
	wait			%p\n\
	memory sweep 	%p\n", 
		think_queue, wait_queue, msweep_queue);
#endif

    /* misc. statistically */
    sim_stat_alloc_new(&len_perread_stat, "length of per read(2)",
        K(32),  server_buffer_stat_count);
    sim_stat_alloc_new(&object_size_stat,   "object size [KB]",
        K(1),   object_size_stat_count);

    sim_stat_alloc_new(&rcode_stat,    "session response code",
        1000,   rcode_stat_count);

    sim_stat_alloc_new(&url_first_hash_stat,    "first hash value",
        K(2),   url_first_hash_stat_count);
    sim_stat_alloc_new(&url_second_hash_stat,   "second hash value",
        K(2),   url_second_hash_stat_count);
    sim_stat_alloc_new(&url_third_hash_stat,   "second hash value",
        K(2),   url_third_hash_stat_count);

    sim_stat_alloc_new(&object_alloc_stat,      "object alloc",
        K(10),  object_alloc_stat_count);

    /* history */
    sim_hist_alloc_new(&comp_hist,              "compaction history",
        30,  comp_hist_area);
    sim_hist_alloc_new(&expire_hist,            "expiration history",
        30,  expire_hist_area);


    /* DNS cache */
    hcache_setup();


#if 0
    /* object cache */
    cache_memory_size = require_total_memory_size - sess_memory_size;
    chk = cache_alloc_area(cache_memory_size,
            require_num_large_block, require_size_large_block,
            require_size_block, require_minimum_size_line);
    if(chk) {
        Error("Sorry, I exit. not enough memory.");
        exit(4);
    }
    if(nblock<0) {
        Error("Sorry, I exit. the number of block is too small (%d).", nblock);
        exit(7);
    }
#endif

#if 1
{
    int bk_num;
    int bk_sk_level, bk_sk_len;
    int frag_size, frag_freelist_len;
    long long U;

    long long skcache_calc_size();
#if 0
    int skcache_calc_param(long long want,
        int *xbk_num, int *xsk_level, int *xsk_len,
        int *xffl_num, int *xffl_size);
#endif
    int skcache_calc_param2(long long want,
        int *xbk_num, int *xsk_level,
        int *xffl_size, double r,
        int *xsk_len, int *xffl_num);
    long long skcache_calc_size(
        int *xbk_num, int *xsk_level, int *xsk_len,
        int *xffl_num, int *xffl_size);

    bk_num = require_num_bucket;
    bk_sk_level = require_level_bucket_skiplist;
    frag_size = require_size_frag;

#if 0
    skcache_calc_param(
        (long long)(require_total_memory_size-sess_memory_size),
        &bk_num, &bk_sk_level, &bk_sk_len, &frag_freelist_len, &frag_size);
#endif
    skcache_calc_param2(
        (long long)(require_total_memory_size-sess_memory_size),
        &bk_num, &bk_sk_level, &frag_size, (double)require_ratio_frag_length,
        &bk_sk_len, &frag_freelist_len);

    U = skcache_calc_size(&bk_num, &bk_sk_level, &bk_sk_len, &frag_freelist_len, &frag_size);

#if 1
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

#if 1
    Trace("@ %d %4d %2d %5d %.2f -> %9d %9d => %9d %9d\n",
        boot_date, bk_num, bk_sk_level, frag_size, (double)require_ratio_frag_length,
        bk_sk_len, frag_freelist_len,
        bk_sk_len * bk_num,  frag_freelist_len* bk_num);
#endif


    U = skcache_init(boot_date-86400*7,
            bk_num, bk_sk_level, bk_sk_len, frag_freelist_len, frag_size);
#if 1
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

    cache_memory_size = U;
}
#endif


#if 1

    Trace("Memory allocation:\n\
\ttotal        %12lld // %5ldM+%06ld\n\
\tsession      %12lld // %5ldM+%06ld %6.2f%%\n\
\tobject cache %12lld // %5ldM+%06ld %6.2f%%\n",
        (long long)require_total_memory_size,
          (long)dM(require_total_memory_size),
          (long)mM(require_total_memory_size),
        (long long)sess_memory_size,
          (long)dM(sess_memory_size), (long)mM(sess_memory_size),
          100.0*sess_memory_size/require_total_memory_size,
        (long long)cache_memory_size,
          (long)dM(cache_memory_size), (long)mM(cache_memory_size),
          100.0*cache_memory_size/require_total_memory_size);

#endif


#if 0
    cache_list_entry(stderr);
#endif

#if 0
    {
        struct timeval now;
        gettimeofday(&now, NULL);
        client_scramble_key = now.tv_usec;
    }
#endif
#if 1
        client_scramble_key = 0xabba;
#endif


    /***
     *** Thread launching
     ***/

#ifdef DONT_USE_THREAD_ATTR
    p_th_attr_high  = NULL;
    p_th_attr_mid   = NULL;
    p_th_attr_low   = NULL;

#else
    chk = pthread_attr_init(&th_attr_high);
    chk = pthread_attr_init(&th_attr_mid);
    chk = pthread_attr_init(&th_attr_low);
    Trace("ret pthread_attr_init() %d (%d)\n", chk, errno);

    chk = pthread_attr_getschedparam(&th_attr_high, &th_param);
    Trace("default (initialized) priority value %d\n",
        th_param.sched_priority);
    th_param.sched_priority += th_param.sched_priority; /* double */
    chk = pthread_attr_setschedparam(&th_attr_high, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_mid, &th_param);
    th_param.sched_priority = th_param.sched_priority; /* thru */
    chk = pthread_attr_setschedparam(&th_attr_mid, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_low, &th_param);
    th_param.sched_priority = th_param.sched_priority/2; /* half */
    chk = pthread_attr_setschedparam(&th_attr_low, &th_param);

    chk = pthread_attr_getschedpolicy(&th_attr_mid, &th_policy_type);
    Trace("ret pthread_attr_getschedpolicy() %d (%d)\n", chk, errno);
    Trace("default value %d\n", th_policy_type);

#if 1
    th_policy_type = SCHED_FIFO;
    chk = pthread_attr_setschedpolicy(&th_attr_high, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_mid, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_low, th_policy_type);
#endif

    p_th_attr_high  = &th_attr_high;
    p_th_attr_mid   = &th_attr_mid;
    p_th_attr_low   = &th_attr_low;
#endif


    if(monitor_flag) {
        chk = pthread_create(&th_monitor, p_th_attr_mid,
                Monitor, NULL);
#ifdef TRACE_THREADS
        Trace("launch - monitor %d\n", chk);
#endif
    }


    /* boot listener, use maximum value of OS */
    listen_fd = TCP_bind_port(service_portno, -1);
    Trace("listen_fd %d, chk %d\n", listen_fd, chk);
    if(listen_fd<0) {
        Error("service port(%d) is already use", service_portno);
        exit(4);
    }
#ifdef DO_ACCEPT_LOCK
    chk = pthread_create(&th_listener, p_th_attr_high,
                Listener, (void*)&listen_fd);
#endif
    chk = pthread_create(&th_listener, p_th_attr_high,
                Listener, (void*)&listen_fd);
#ifdef TRACE_THREADS
    Trace("launch - listenter %d\n", listen_fd, chk);
#endif

    ll = require_n_handler;
#ifdef TRACE_THREADS
    Trace("ll %d\n", ll);
#endif

    /* boot handlers */
    for(i=0;i<require_n_handler;i++) {
        chk = pthread_create(&th_handlers[i], p_th_attr_mid,
                Slave, (void*)i);
    }

#if 0
    /* XXX */
    {
        int th;
        void* Banker(void*);

        chk = pthread_create(&th, p_th_attr_low,
                Banker, (void*)i);
    }
#endif

#if 1
    /* XXX */
    {
        pthread_t th;
		int i;

		for(i=0;i<require_n_dsweeper;i++) {
			chk = pthread_create(&th, p_th_attr_low,
					DSweeper, (void*)i);

		}
    }
#endif

#if 1
    /* XXX */
    {
        pthread_t th;
		int i;

		for(i=0;i<require_n_msweeper;i++) {
			chk = pthread_create(&th, p_th_attr_low,
					MSweeper, (void*)i);
		}
    }
#endif



#ifdef TRACE_THREADS
    Trace("launching time is over\n");
#endif

    MUTEX_LOCK(&stat_lock);
    service_start_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);


    /***
     ***
     ***/
#if 0
    pthread_sigmask(SIG_BLOCK, &handler_sig_mask, NULL);
    pthread_create(&th_signalhandler, p_th_attr_low, signal_watcher, NULL);
    pthread_sigmask(SIG_UNBLOCK, &handler_sig_mask, NULL);
#endif

    atexit(LogSync);


    MUTEX_LOCK(&stat_lock);
    Trace("initalize take %d sec(s)\n", time(NULL) - boot_date);
    MUTEX_UNLOCK(&stat_lock);

    /***
     ***
     ***/
    sleep(0);

    /* for child reaping, this thread should be awake at sometime */
    while(1) {
		struct timespec w;

        MUTEX_LOCK(&stat_lock);
		swapout_token = swapout_token_init;
        MUTEX_UNLOCK(&stat_lock);

		w.tv_sec = swapout_token_interval / 1000;
		w.tv_nsec = (swapout_token_interval % 1000) * 1000000;

		nanosleep(&w, NULL);
    }
}

