/*
 *
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * Original algorithm was described by Pascal like langauage.
 * Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/time.h>
#include <pthread.h>
#include <time.h>

#include <malloc.h>
#include <errno.h>
#include <assert.h>


#include "sk.h"

#ifndef MIN
#define MIN(x,y)    ((x<y) ? x : y)
#endif
#ifndef MAX
#define MAX(x,y)    ((x>y) ? x : y)
#endif

void
sk_rawdump(FILE* fp, skiplist_t *list, char* (fn)(void*))
{
    skiplist_cell_t *cur;
    int c, i;

#if 0
    for(i=0;i<list->level;i++) {
        fprintf(stdout, " h %4d: %p\n", i, list->header->skip[i]);
    }
#endif

    fprintf(fp, "--- list addr %p, num %d, func %p\n",
        list, list->num, fn);

    c = 0;
    cur = list->header->skip[0];
    if(fn)
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p %s // %p\n",
                c, cur, cur->entity, fn(cur->entity), cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
    else
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p // %p\n",
                c, cur, cur->entity, cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
}


void
sk_dump(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

    fprintf(fp, "%-24s %-8s ", "header", "header");
    for(i=0;i<list->level;i++) {
        if(list->header->skip[i]==NULL) {
            fprintf(fp, "%-8s ", "NIL");
        }
        else {
            fprintf(fp, "%p ", list->header->skip[i]);
        }
    }
    fprintf(fp, "\n");

#if 0
    cur = list->header->skip[0];
    while(cur) {
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
        }
        else {
            strcpy(label, "*NONE*");
        }


        fprintf(fp, "%-24s %p ", label, cur);

        for(i=0;i<list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

        cur = cur->skip[0];
    }

    fprintf(fp, "%-24s %-8s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
#endif


}

void
bk_dump(bucket_t *bk, skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    char *pcur;
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int j, k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

#ifndef DEBUG
    return;
#endif

#ifdef DEBUG
    fprintf(fp, "#  first %p last %pd\n#\n",
        bk->cell_first, bk->cell_last);

#if 0
    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %s %p\n",
            "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s ",
            "", "", "");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %s %p\n",
            "", "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s %-14s ",
            "", "", "", "");
    }
    if(list->level<0) {
        fprintf(fp,"*** no level ***\n");
    }
    else {

        cur = list->header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
    fprintf(fp, "\n");

    }
#endif

#if 1
    fprintf(fp, "# cup  header %p: ", bk->cup.header);
        cur = bk->cup.header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
    fprintf(fp, "# cup  term   %p: ", bk->cup.term);
        cur = bk->cup.term;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

#endif


    fprintf(fp, "#\n");

    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %8s",
            "num", "addr", "entity", "skips..");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %8s",
            "num", "addr", "entity", "label", "skips..");
    }
    fprintf(fp, "\n");

    j = -1;
    k = 0;
    pcur = (char*)bk->cell_first;
    while(pcur<=(char*)bk->cell_last) {
        cur=(skiplist_cell_t*)pcur;

        j++;
        if(cur->entity==NULL) {
            goto next;
        }
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
            fprintf(fp, "= %6d %6d %p %p %-14s ",
                j, k, cur, cur->entity, label);
        }
        else {
            fprintf(fp, "= %6d %6d %p %p ",
                j, k, cur, cur->entity);
        }

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "x");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
        k++;
#if 0
        if(k>list->num+3) {
            break;
        }
#endif

next:
        pcur += bk->elemsize;
    }

    fprintf(fp, "\n\n");
    fflush(fp);

#endif
}


static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 4097;
    double v;

    lv = 0;
#if 0
    while( v=((double)(rand()%w))/w, v < p && lv < max) {
        lv++;
    }
#endif
    while(drand48() < p && lv < max) {
        lv++;
    }

    return lv;
}




/*
 * cmp - compare
 */
skiplist_cell_t *
sk_find_listpos(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    skiplist_cell_t *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(cmp(x->skip[i]->entity, newface)<0) {
            x = x->skip[i];
        }
        /* XXX */
        if(diff==0) {
            ret = x->skip[i];
            goto out;
        }
    }
    x = x->skip[0];

    if(cmp(x->entity, newface)==0) {
        ret = x;
    }

out:
    return ret;
}

void*
sk_find(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    void *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(diff = cmp(x->skip[i]->entity, newface), diff<0) {
            x = x->skip[i];
        }
        /* XXX */
        if(diff==0) {
            ret = x->skip[i]->entity;
            goto out;
        }
    }
    x = x->skip[0];

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
    }

out:
    return ret;
}

/*
 * if alreay assigned, not new allocation.
 * cmp - compare
 */
void*
sk_assign(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*),
        int* st, int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    skiplist_cell_t *ret;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
        *st = 1;
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            ret = NULL;
            *st = -1;
            goto out;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        ret = x->entity;
        *st = 0;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

out:
    return ret;
}

int
sk_insert(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (ovw)(void**,void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0 && ovw) {
        ovw(&x->entity, newface);
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            return -1;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

    return 0;
}

/*
 * cmp - compare
 * rem - remove
 */
void*
sk_delete(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (rem)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
#if 0
    skiplist_cell_t *y;
#endif
    int i;
    void *ret;
    int diff;
    int chk;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(diff = cmp(x->skip[i]->entity, newface), diff<0) {
            x = x->skip[i];
            if(x->skip[i]==NULL)
                break;
        }
        update[i] = x;
    }
    x = x->skip[0];

    if(!x) {
        return ret;
    }

    if(cmp(x->entity, newface)==0) {
        if(rem) {
            chk = rem(x->entity);
            if(chk) {
                return ret;
            }
        }

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x) {
                update[i]->skip[i] = x->skip[i];
            }
        }

        while(list->level>0 &&
                list->header->skip[list->level]==list->term) {
            list->level--;
        }

        ret = x->entity;
        x->entity = NULL;
        for(i=1;i<=list->maxlevel;i++) {
            x->skip[i] = NULL;
        }
        x->skip[0] = *flist;
        *flist = x;

        list->num--;
    }

    return ret;
}

/*
 * ack - decide delete or not
 *          that function have to return not zero for delete
 */
int
sk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */

        if(ack==NULL)
            goto delete;

        chk = ack(x->entity, opt);

        if(chk) {
delete:
            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
                y->skip[i] = NULL;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }
no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}

int
Xsk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    do {
        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = x->skip[0];
        if(x==list->term)
            break;

        chk = ack(x->entity);
fprintf(stderr, "x %p, entity %p, chk %d\n", x, x->entity, chk);

fprintf(stderr, "\t");
 for(i=0;i<=list->level;i++) {
    fprintf(stderr, "%p ", update[i]);
 }
fprintf(stderr, "\n");


        if(chk) {

            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }



            ret++;
        }
    } while(x!=list->term);

    return ret;
}

skiplist_t*
sk_new(int nlevel, void* user_term)
{
    skiplist_t *this;
    skiplist_cell_t *n;
    int i, j;

    this = (skiplist_t*) malloc(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }
    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;
    n = this->header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }
    n = this->term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)malloc(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    return this;
}


int
bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->freelist);

    n  = bk->freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    return 0;
}

int
bk_new(bucket_t *bk, int nlevel, int nelement, void* user_term)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;
    char *pool;


    if(bk==NULL) {
        return 0;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    bk->elemnum = nelement;
    bk->elemsize = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(nlevel+1);
    bk->totalsize = bk->elemsize*(nelement+4);
    pool = (char*)malloc(bk->totalsize);
    if(pool==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        return -1;
    }

#if 0
    fprintf(stderr, "nlevel %d, nelement %d\n",
            nlevel, nelement);
    fprintf(stderr, "element size %d, num %d -> total %d\n",
            bk->elemsize, nelement, bk->totalsize);
    fprintf(stderr, "pool  [%p..%p)\n", pool, pool+bk->totalsize);
#endif
#if 0
    for(i=0;i<nelement+2;i++) {
        fprintf(stdout, "%4d: <%p>\n", i, pool+bk->elemsize*i);
    }
    fflush(stdout);
#endif
    if(pthread_mutex_init(&bk->mutex, NULL)<0) {
        Error("fail pthread_mutex_init() (%d)", errno);
    }

    bk->cup.num         = 0;
    bk->cup.maxlevel    = nlevel;
    bk->cup.level       = 0;
    bk->cup.header      = (skiplist_cell_t*)pool;
    bk->cup.term        = (skiplist_cell_t*)(pool + 1*bk->elemsize);
    n = bk->cup.header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }
    n = bk->cup.term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }


#if 0
    fprintf(stderr, "cup  header %p\n", bk->cup.header);
    fprintf(stderr, "cup  term   %p\n", bk->cup.term);
#endif

    bk->freelist = (skiplist_cell_t*)(pool+4*bk->elemsize);
#if 0
    fprintf(stderr, "freelist    %p\n", bk->freelist);
#endif


#ifdef DEBUG
    bk->cell_first = (skiplist_cell_t*)(pool+0*bk->elemsize);
    bk->cell_last  = (skiplist_cell_t*)(pool+nelement*bk->elemsize);

    fprintf(stderr, "  first..last %p..%p\n", bk->cell_first, bk->cell_last);
#endif

    n = bk->freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL) {
            break;
        }
#if 0
        fprintf(stdout, "* %d, %p\n", i, n);
        fflush(stdout);
#endif
        n->entity = NULL;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=1;j<nlevel;j++) {
            n->skip[j] = NULL;
        }
        if(i<nelement-1) {
            n->skip[0] = (skiplist_cell_t*)((char*)n+bk->elemsize);
            n = (skiplist_cell_t*)((char*)n+bk->elemsize);
        }
        else {
            n->skip[0]  = NULL;
        }
    }

#if 0
    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    fprintf(stdout, "free: %p\n", bk->freelist);

    n  = bk->freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL)
            break;

#if 0
        fprintf(stdout, "%4d: <%p> ", i, n);
        for(j=0;j<nlevel;j++) {
            fprintf(stdout, "%d:<%p> %p ",
                j, &(n->skip[j]), n->skip[j]);
        }
        fprintf(stdout, "\n");
#endif
#if 1
        fprintf(stdout, "%4d: <%p> %p\n", i, n, n->skip[0]);
#endif

        n = n->skip[0];
    }

#endif

    return bk->totalsize;
}


