/*
 * Skiplist & its Buckets
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 *  $Id$
 *
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * Original algorithm was described by Pascal like langauage.
 * Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

typedef struct _skiplist_cell_t {
    void*                   entity;
    struct _skiplist_cell_t **skip;
} skiplist_cell_t;

typedef struct _skiplist_t {
    int             maxlevel;
    int             level;
    skiplist_cell_t *header;
    skiplist_cell_t *term;
    int             num;
} skiplist_t;

typedef struct _bucket_t {
    pthread_mutex_t mutex;

    skiplist_t      cup;
    int             elemnum;
    int             elemsize;
    int             totalsize;
    skiplist_cell_t *freelist;
#ifdef DEBUG
    skiplist_cell_t *cell_first;
    skiplist_cell_t *cell_last;
#endif
} bucket_t;
#ifdef DEBUG
    void *entity_first;
    void *entity_last;
#endif

#ifndef SKL_MAXLEVEL
#define SKL_MAXLEVEL    64
#endif

/* skiplist */
void sk_rawdump (FILE *fp ,skiplist_t *list ,char *(fn )(void *));
void sk_dump (skiplist_t *list ,FILE *fp ,char *(fn )(void *));
void *sk_find (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
void *sk_assign (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int *st ,int opt );
int sk_insert (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int (ovw )(void **,void *),int opt );
void *sk_delete (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void*),int (rem )(void *),int opt );
int sk_mapdelete (skiplist_t *list ,skiplist_cell_t **flist ,int (ack )(void *,void*),void *);
skiplist_t *sk_new (int nlevel ,void *user_term );

/* bucket */
void bk_dump (bucket_t *bk ,skiplist_t *list ,FILE *fp ,char *(fn )(void *));
skiplist_cell_t *sk_find_listpos (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
int bk_new (bucket_t *bk ,int nlevel ,int nelement ,void *user_term );
int bucket_show (bucket_t *bk );



