/*
 *
 */
/*
run -s 163 -n 220000 -l 15 -f x.128k >& out.x
run -s 163 -n 220000 -l 12 -f x.128k >& out.x
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>
#include <pthread.h>
#include <malloc.h>
#include <errno.h>
#include <assert.h>

#include "sk.h"




typedef struct {
    int     seq;

    int k1;
    int k2;
    int k3;
    char    *loc;

    int     oid;
    int     ref;

    int     stype;
    off_t   spos;
} dummy_t;

#define KOTETU_SUB_TIME(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )


char *dummy_label(void* x)
{
    static char buf[256];
    dummy_t *pos;

    pos = (dummy_t*) x;

    if(pos) {
        if(pos->k1==INT_MAX || pos->k2==INT_MAX || pos->k3==INT_MAX) {
            sprintf(buf, "MAXV:MAXV:MAXV");
        }
        else {
            sprintf(buf, "%04d:%04d:%04d", pos->k1, pos->k2, pos->k3);
        }
    }
    else {
        sprintf(buf, "EMPTY");
    }

    return buf;
}

char *dummy_detail_label(void* x)
{
    static char buf[256];
    dummy_t *pos;

    pos = (dummy_t*) x;

    if(pos) {
        if(pos->k1==INT_MAX || pos->k2==INT_MAX || pos->k3==INT_MAX) {
            sprintf(buf, "MAXV:MAXV:MAXV");
        }
        else {
            sprintf(buf, "%04d:%04d:%04d-%04d",
                pos->k1, pos->k2, pos->k3, pos->ref);
        }
    }
    else {
        sprintf(buf, "EMPTY");
    }

    return buf;
}

int
X_bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: %p <%p> %p %s\n",
            i, n, n->entity, n->skip[0], dummy_label(n->entity));
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->freelist);

    n  = bk->freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: %p <%p> %p %s\n",
            i, n, n->entity, n->skip[0], dummy_label(n->entity));
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

}



int dummy_cmp(void *px, void *py)
{
    dummy_t *x, *y;
    int a, b;
    int r;

    x = (dummy_t*)px;
    y = (dummy_t*)py;

    if(x==NULL) {
        r = INT_MAX;
    }
    else
    if(y==NULL) {
        r = 1;
    }
    else {
        if(x->k1 != y->k1) {
            r = x->k1 - y->k1;
        }
        else {
            if(x->k2 != y->k2) {
                r = x->k2 - y->k2;
            }
            else {
                if(x->k3 != y->k3) {
                    r = x->k3 - y->k3;
                }
                else {
                    r = 0;
                }
            }
        }
#if 0
    fprintf(stderr, "cmp: %p <%04d:%04d:%04d> vs %p <%04d:%04d:%04d> -> r %d\n",
        x, x->k1, x->k2, x->k3,
        y, y->k1, y->k2, y->k3, r);
#endif
    }
#if 0
    fprintf(stderr, "cmp: x %p y %p -> r %d\n", x, y, r);
#endif

    return r;
}

int dummy_cmp_simple(void *px, void *py)
{
    if(px==NULL) {
        return 1;
    }
    else
    if(py==NULL) {
        return -1;
    }
    else {
        return ((dummy_t*)px)->k1 - ((dummy_t*)py)->k1;
    }
}

int dummy_replace(void *px, void *py)
{
    dummy_t *x, *y;

    x = (dummy_t*)px;
    y = (dummy_t*)py;

    *x = *y;
}

int dummy_free(void *px)
{
    dummy_t *x;

    x = (dummy_t*)px;

#if 0
fprintf(stderr, "target adrress %p\n", x);
#endif
}

void
dummy_list_show(skiplist_t *list, FILE *fp)
{
    skiplist_cell_t* cur;
    dummy_t *d;
    int i;
    char label[256];
    char *p;

    if(list==NULL) {
        return;
    }

    cur = list->header->skip[0];
    i = 0;
    while(cur && cur!=list->term) {
        d = (dummy_t*)cur->entity;
        if(d==NULL) {
            break;
        }
/*
        fprintf(fp, "%4d %04d %04d\n", d->k1, d->k2, d->k3);
*/
/*
        fprintf(fp, " %7d: %4d <%p> <%p>\n", i, d->k1, cur, cur->skip[0]);
*/
/*
        fprintf(fp, "%4d\n", d->k1);
*/
        fprintf(fp, " %7d: %s\n", i, dummy_label(d));

        cur = cur->skip[0];
        i++;
    }

#if 0
fprintf(stderr, "%d item(s)\n", i);
#endif

}

void
dummy_array_show(dummy_t *base, int c)
{
    skiplist_cell_t* cur;
    dummy_t *d;
    int i;

    if(base==NULL) {
        return;
    }

    for(i=0;i<c;i++) {
        d = &base[i];
        if(d==NULL) {
            break;
        }
        fprintf(stdout, "%4d %04d %04d\n", d->k1, d->k2, d->k3);
    }
}




int
read_data(char *filename, dummy_t *elem, int limit)
{
    int c;
    char line[1024];
    FILE *fp;
    dummy_t *foo;

    fp = fopen(filename, "r");
    if(fp==NULL) {
        return -1;
    }

    c = 0;
    while(fgets(line, 1024, fp)!=NULL) {
        foo = &elem[c];

        foo->seq = c;

/*
        foo->k1 = atoi(line);
*/
        foo->k1 = c;
        foo->k2 = foo->k1;
        foo->k3 = foo->k2;
        foo->oid = c;
        foo->ref = 0;

#if 0
        fprintf(stdout, "*** %4d: %04d:%04d:%04d ***\n",
            c, foo->k1, foo->k2, foo->k3);
#endif
        c++;
        if(c>=limit) {
            fprintf(stderr, "overun %d [0..%d)\n", c, limit);
            break;
        }
    }

    fclose(fp);

    return c;
}

int
dummy_ref_inc(void* xp)
{
    dummy_t *p;

    p = (dummy_t*)xp;

/*
    fprintf(stderr, "dummy_ref_inc: %p <%d ref %d>\n",
        p, p->k1, p->k3);
*/

    if(p->k3<0) {
        p->k3 = 1;
    }
    else {
        p->k3++;
    }

    return 0;
}

int
dummy_ref_dec(void* xp)
{
    dummy_t *p;

    p = (dummy_t*)xp;

/*
    fprintf(stderr, "dummy_ref_dec: %p <%d ref %d>\n",
        p, p->k1, p->k3);
*/

    p->k3--;
    if(p->k3>0) {
        return 1;
    }
    return 0;
}


int
dummy_trap(void** xdst, void *xnew)
{
    dummy_t **dst;
    dummy_t *newone;

    dst = (dummy_t**)xdst;
    newone = (dummy_t*)xnew;

    fprintf(stderr, "old %p <%d>, new %p <%d>\n",
        *dst, (*dst)->k1, newone, newone->k1);
}

int
dummy_overwrite(void** xdst, void *xnew)
{
    dummy_t **dst;
    dummy_t *newone;

    dst = (dummy_t**)xdst;
    newone = (dummy_t*)xnew;

    fprintf(stderr, "old %p <%d>, new %p <%d>\n",
        *dst, (*dst)->k1, newone, newone->k1);
    *dst = newone;
}

/*
 * return:
 *      1   not refered
 *      0   some refered
 */
int
dummy_check_notreferd(void *xp)
{
    dummy_t *x = (dummy_t*)xp;
/*
fprintf(stderr, "++ ref.check %p ref %4d\n", x, x->ref);
*/
    if(x->ref<=0) {
        return 1;
    }
    else {
        return 0;
    }
}
int
dummy_check_notreferd_big(void *xp)
{
    dummy_t *x = (dummy_t*)xp;
/*
fprintf(stderr, "++ ref.check %p ref %4d\n", x, x->ref);
*/
    if(x->ref<=1000) {
        return 1;
    }
    else {
        return 0;
    }
}
int
dummy_check_notreferd_small(void *xp)
{
    dummy_t *x = (dummy_t*)xp;
/*
fprintf(stderr, "++ ref.check %p ref %4d\n", x, x->ref);
*/
    if(x->ref<=10) {
        return 1;
    }
    else {
        return 0;
    }
}



dummy_t *entry;

int
main(int argc, char **argv)
{
    int len;
    bucket_t bk;
    skiplist_t *list;
    dummy_t foo, bar;
    dummy_t *junk;
    skiplist_cell_t *cur;
    int chk;
    int c;
    int i, j;
    struct timeval begin, end;
    int levelrange;
    int flag;
    char *filename;
    int seed;
    dummy_t *special_term;
    int flag_print_after_skiplist=0;
    int flag_print_after_qsort=0;
    int flag_print_after_delete=0;
    bucket_t seq;
    int seq_num;

    len = 1000;
    levelrange = SKL_MAXLEVEL-1;
    filename = "x";
    seed = time(NULL) % getpid() % 100;

    while((flag=getopt(argc, argv, "n:l:f:s:SQD"))!=EOF) {
        switch(flag) {
        case 'n':
            len = atoi(optarg);
            break;
        case 'l':
            levelrange = atoi(optarg);
            break;
        case 'f':
            filename = strdup(optarg);
            break;
        case 's':
            seed = atoi(optarg);
            break;
        case 'S':
            flag_print_after_skiplist = 1 - flag_print_after_skiplist;
            break;
        case 'Q':
            flag_print_after_qsort = 1 - flag_print_after_qsort;
            break;
        case 'D':
            flag_print_after_delete = 1 - flag_print_after_delete;
            break;
        }
    }

    fprintf(stderr, "levelrange %d, length %d\n", levelrange, len);
    fprintf(stderr, "filename '%s'\n", filename);
    fprintf(stderr, "seed %u\n", seed);

    srand((unsigned int)seed);
    srand48(seed);

    entry = (dummy_t*)malloc(sizeof(dummy_t)*(len+1));
    fprintf(stderr, "entry [%p..%p]\n",
        (char*)entry, (char*)entry+sizeof(dummy_t)*(len+1));
    if(entry==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        exit(17);
    }

#ifdef DEBUG
/*
    entity_first = &entry[0];
    entity_last  = &entry[len-1];
*/
    entity_first = (void*)(char*)entry;
    entity_last  = (void*)((char*)entry+sizeof(dummy_t)*(len+1));
#endif

    special_term = &entry[0];
    special_term->k1 = INT_MAX;
    special_term->k2 = INT_MAX;
    special_term->k3 = INT_MAX;
    fprintf(stderr, "special term %p\n", special_term);

    bk_new(&bk, levelrange, len, special_term);




    fprintf(stderr, "\nskip list:\n");

gettimeofday(&begin, NULL);
    c = read_data(filename, &entry[1], len);
    if(c>=len || c<0) {
        fprintf(stderr, "cannot read data file '%s'\n", filename);
        exit(4);
    }

gettimeofday(&end, NULL);
#if 1
    fprintf(stderr, "  read %d items // takes %.4f sec, rate %.3f\n",
        c, ((double) KOTETU_SUB_TIME(end, begin))/1000.0,
        (((double)c)/KOTETU_SUB_TIME(end, begin))*1000.0);
#endif

/*
    dummy_array_show(&entry[0], c);
*/

#ifdef DEBUG
    /*
    bk_dump(&bk, &bk.cup, stdout, NULL);
    fprintf(stderr, "* * * * * start\n");
    */
#endif

gettimeofday(&begin, NULL);

    for(i=1;i<c+1;i++) {
#if 0
        fprintf(stderr, " insert <%p> %4d (%4dth) ",
            &entry[i], entry[i].k1, i);
#endif
#if 0
        fprintf(stderr, "  +++ insert <%p> %d (%dth)+++\n",
            &entry[i], entry[i].k1, i);
#endif
        junk = (dummy_t*) sk_assign(&bk.cup, &entry[i], &bk.freelist,
                dummy_cmp,  &chk, i);
#if 0
        fprintf(stderr, "    in %p, out %p, chk %d\n",
                &entry[i], junk, chk);
#endif
        if(junk==NULL || chk < 0) {
            fprintf(stdout, "cannot insert new entity\n");
            bk_dump(&bk, &bk.cup, stdout, NULL);
            assert(chk==0);
            exit(43);
        }

        junk->ref++;
        if(chk>0) {
            /* already assigned */
        }
        else {
            /* new assigned */
        }


#ifdef DEBUG
    /*
    bk_dump(&bk, &bk.cup, stdout, NULL);
    */
#endif

    }

gettimeofday(&end, NULL);
/*
    bk_dump(&bk, &bk.cup, stdout, NULL);
*/

#if 1
    fprintf(stderr, "  insert %d item(s)\n", c);
    fprintf(stderr, "  takes %.4f sec, rate %.3f ( %.3e )\n",
        ((double) KOTETU_SUB_TIME(end, begin))/1000.0,
        (((double)c)/KOTETU_SUB_TIME(end, begin))*1000.0,
        (((double)c)/KOTETU_SUB_TIME(end, begin))*1000.0);
    fprintf(stderr, "  cup num %d level %d\n", bk.cup.num, bk.cup.level);
#endif

    if(flag_print_after_skiplist) {
/*
        dummy_list_show(&bk.cup, stdout);
*/
        sk_rawdump(stderr, &bk.cup,  dummy_detail_label);
    }

/*
    sk_rawdump(stderr, &bk.cup,  dummy_detail_label);
*/
/*
    bk_dump(&bk, &bk.cup, stdout, dummy_detail_label);
*/
    fprintf(stderr, "\ndelete with map:\n");

gettimeofday(&begin, NULL);
/*
    c = sk_mapdelete(&bk.cup, &bk.freelist, dummy_check_notreferd_big, 0);
*/
    c = sk_mapdelete(&bk.cup, &bk.freelist, NULL, 0);
gettimeofday(&end, NULL);

    if(flag_print_after_delete) {
        sk_rawdump(stderr, &bk.cup,  dummy_detail_label);
    }

#if 1
    fprintf(stderr, "  delete %d item(s)\n", c);
    fprintf(stderr, "  takes %.4f sec, rate %.3f ( %.3e )\n",
        ((double) KOTETU_SUB_TIME(end, begin))/1000.0,
        (((double)c)/KOTETU_SUB_TIME(end, begin))*1000.0,
        (((double)c)/KOTETU_SUB_TIME(end, begin))*1000.0);
    fprintf(stderr, "  cup num %d level %d\n", bk.cup.num, bk.cup.level);
#endif

    exit(0);
}

