/*
 * Skiplist & its Buckets
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 *  $Id$
 *
 */

/*
 * skcache.h - object cache procedures for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 *
 *
 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * Original algorithm was described by Pascal like langauage.
 * Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

typedef struct _skiplist_cell_t {
    void*                   entity;
    struct _skiplist_cell_t **skip;
} skiplist_cell_t;

typedef struct _skiplist_t {
    int             maxlevel;
    int             level;
    skiplist_cell_t *header;
    skiplist_cell_t *term;
    int             num;
} skiplist_t;


#ifndef SKL_MAXLEVEL
#define SKL_MAXLEVEL    64
#endif

#ifndef RADIX_H1_MAX
#define RADIX_H1_MAX    (257)
#endif
#ifndef RADIX_H2_MAX
#define RADIX_H2_MAX    (257)
#endif
#ifndef RADIX_H3_MAX
#define RADIX_H3_MAX    (257)
#endif

#ifndef MAGIC_0
#define MAGIC_0     'K'
#endif
#ifndef MAGIC_1
#define MAGIC_1     'P'
#endif

#define XSK_ENTRY	1
#define XSK_FRAG	2


typedef struct _skcache_frag_t {
    int size;
    int len;
    char *ent;
    struct _skcache_frag_t *next;
} skcache_frag_t;

typedef struct _skcache_entry_t {
    char magic[2];

    int h1, h2, h3, h4;             /* quad hash */
    int oid;
    unsigned char name[L_URL];

    int count;
    int ref;
/*
    int owner;
*/
    void *owner;
    int cracked;

    time_t issue_time;
    time_t hot;
    time_t lmd;                     /* last modified */
    time_t expire;                  /* expire */

    off_t totallength;
    skcache_frag_t *m_recorded;
    int d_recorded;

    struct _skcache_entry_t *next;

} skcache_entry_t;

typedef struct _bucket_t {
    pthread_mutex_t mutex;

	int				bid;
	int				own;

    skiplist_t      cup;
    int             elemnum;
    int             elemsize;
    int             totalsize;
    skiplist_cell_t *elem_freelist;

	time_t			last_gc;
	int				gc_push;

    int             inmemory;
    int             indisk;
	int				op[4];

    int             dirty2;
    int             dirty3;

    skcache_entry_t *entry_freelist;
    int             entry_use, entry_all;
    int             entry_hhwater;
    int             entry_hwater;
    int             entry_lwater;

    skcache_frag_t  *frag_freelist;
    int             frag_use, frag_all;
    int             frag_hhwater;
    int             frag_hwater;
    int             frag_lwater;

#ifdef DEBUG
    skiplist_cell_t *cell_first;
    skiplist_cell_t *cell_last;
#endif
} bucket_t;

#ifdef DEBUG
    void *entity_first;
    void *entity_last;
#endif


/* skiplist */
void sk_rawdump (FILE *fp ,skiplist_t *list ,char *(fn )(void *));
void sk_dump (skiplist_t *list ,FILE *fp ,char *(fn )(void *));
void *sk_find (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
void *sk_assign (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int *st ,int opt );
int sk_insert (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void *),int (ovw )(void **,void *),int opt );
void *sk_delete (skiplist_t *list ,void *newface ,skiplist_cell_t **flist ,int (cmp )(void *,void*),int (rem )(void *),int opt );
int sk_mapdelete (skiplist_t *list ,skiplist_cell_t **flist ,int (ack )(void *,void*),void *);
skiplist_t *sk_new (int nlevel ,void *user_term );

/* bucket */
void bk_dump (bucket_t *bk ,skiplist_t *list ,FILE *fp ,char *(fn )(void *));
skiplist_cell_t *sk_find_listpos (skiplist_t *list ,void *newface ,int (cmp )(void *,void *));
int bk_new (bucket_t *bk ,int nlevel ,int nelement ,void *user_term );
int bucket_show (bucket_t *bk );


extern skcache_entry_t *NIL;
extern int skcache_bucket_num;
extern bucket_t *skcache_bucket_base;
extern skcache_entry_t **skcache_entry_freelist_base;
extern skcache_frag_t   **skcache_frag_freelist_base;

int skcache_entry_cmp(void *px, void *py);
char* skcache_label_hash(void *pp);
char* skcache_label_ref(void *pp);
char* skcache_label_short(void *pp);

int skcache_have(int , int , int , int , char *, int , time_t );

long long skcache_init(time_t, int xbucket_num, int xsk_level, int xsk_num,
        int xfl_num, int xfl_size
);


int skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e);
int skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e);

skcache_entry_t* skcache_pull_new_entry(bucket_t *bk);
int skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen);

skcache_frag_t* skcache_pull_new_frag(bucket_t *bk);
int skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);

#ifdef CHECK_BKMUTEX

#define LOCK_BK(x)  if(x) {\
    Trace("bkl do   th %d bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    pthread_mutex_lock(&(x->mutex)); \
    Trace("bkl done th %d bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    }\
    else {\
    Trace("bkl null th %d bk. *strange*\n", pthread_self()); \
    }

#define UNLOCK_BK(x)    if(x) {int r;\
    Trace("bku do   th %d bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    r = pthread_mutex_unlock(&(x->mutex)); \
    Trace("bku done th %d bk# %d, : %s %5d\n", \
         pthread_self(), bk->bid, __FILE__ , __LINE__); \
    }\
    else {\
    Trace("bku null th %d bk. *strange*\n", pthread_self()); \
    }

#else

#define LOCK_BK(x)      pthread_mutex_lock(&(x->mutex))
#define UNLOCK_BK(x)    pthread_mutex_unlock(&(x->mutex))

#endif


